/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */
package de.javagl.swogl;

/**
 * This class contains utility methods for creating 
 * {@link Condition Conditions}
 */
public class Conditions
{
    /**
     * Returns a new {@link Condition} that is always <code>true</code>
     * 
     * @param <T> The type of the objects for the condition
     * @return The new condition.
     */
    public static <T> Condition<T> trueCondition()
    {
        return new Condition<T>()
        {
            @Override
            public boolean is(T t)
            {
                return true;
            }
            
            @Override
            public String toString()
            {
                return "trueCondition";
            }
        };
    }
    
    /**
     * Returns a new {@link Condition} that is always <code>false</code>
     * 
     * @param <T> The type of the objects for the condition
     * @return The new condition.
     */
    public static <T> Condition<T> falseCondition()
    {
        return new Condition<T>()
        {
            @Override
            public boolean is(T t)
            {
                return false;
            }
            
            @Override
            public String toString()
            {
                return "falseCondition";
            }
            
        };
    }
    
    /**
     * Returns a new {@link Condition} that is the negation of the given
     * condition.
     * 
     * @param <T> The type of the objects for the condition
     * @param condition The condition to negate
     * @return The new condition.
     */
    public static <T> Condition<T> not(final Condition<T> condition)
    {
        return new Condition<T>()
        {
            @Override
            public boolean is(T t)
            {
                return !condition.is(t);
            }

            @Override
            public String toString()
            {
                return "notCondition("+condition+")";
            }
            
        };
    }
    
    /**
     * Returns a new {@link Condition} that is the conjunction of the given
     * conditions.
     * 
     * @param <T> The type of the objects for the condition
     * @param c0 The first condition
     * @param c1 The second condition
     * @return The new condition.
     */
    public static <T> Condition<T> and(
        final Condition<T> c0, final Condition<T> c1)
    {
        return new Condition<T>()
        {
            @Override
            public boolean is(T t)
            {
                return c0.is(t) && c1.is(t);
            }
            
            @Override
            public String toString()
            {
                return "andCondition("+c0+", "+c1+")";
            }
        };
    }

    /**
     * Returns a new {@link Condition} that is the disjunction of the given
     * conditions.
     * 
     * @param <T> The type of the objects for the condition
     * @param c0 The first condition
     * @param c1 The second condition
     * @return The new condition.
     */
    public static <T> Condition<T> or(
        final Condition<T> c0, final Condition<T> c1)
    {
        return new Condition<T>()
        {
            @Override
            public boolean is(T t)
            {
                return c0.is(t) || c1.is(t);
            }
            
            @Override
            public String toString()
            {
                return "orCondition("+c0+", "+c1+")";
            }
        };
    }
    
    /**
     * Private constructor to prevent instantiation
     */
    private Conditions()
    {
    }
}
