/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.awt.Rectangle;
import java.util.IdentityHashMap;
import java.util.Map;

/**
 * Utility class for managing dirty rectangles.
 * 
 * @param <T> The type of objects to operate on
 */
class DirtyRectangleManager<T>
{
    /**
     * The map from the objects to their dirty rectangles
     */
    private final Map<T, Rectangle> dirtyRectangles;
    
    /**
     * Creates a new DirtyRectangleManager
     */
    public DirtyRectangleManager()
    {
        dirtyRectangles = new IdentityHashMap<T, Rectangle>();
    }

    /**
     * Add the specified dirty rectangle for the given object.
     * If the object already has a dirty rectangle, the union
     * of the present rectangle and the specified rectangle will be
     * computed and stored.
     * 
     * @param object The object
     * @param x The x-coordinate of the dirty rectangle
     * @param y The y-coordinate of the dirty rectangle
     * @param w The width of the dirty rectangle
     * @param h The height of the dirty rectangle
     */
    void addDirtyRectangle(T object, int x, int y, int w, int h)
    {
        if (object == null || w <= 0 || h <= 0)
        {
            return;
        }
        Rectangle rectangle = dirtyRectangles.get(object);
        if (rectangle == null)
        {
            rectangle = new Rectangle(x,y,w,h);
            dirtyRectangles.put(object, rectangle);
        }
        else
        {
            merge(rectangle, x,y,w,h);
        }
        //System.out.println("DirtyRectangle now "+rectangle+" for "+object);
    }
    
    /**
     * Merge the given rectangle with the specified rectangle,
     * writing the result into the given rectangle.
     * 
     * @param rectangle The rectangle
     * @param x The x-coordinate of the rectangle to add
     * @param y The y-coordinate of the rectangle to add
     * @param w The width of the rectangle to add
     * @param h The height of the rectangle to add
     */
    static void merge(Rectangle rectangle, int x, int y, int w, int h)
    {
        int x0 = Math.min(x, rectangle.x);
        int y0 = Math.min(y, rectangle.y);

        int x1a = x+w;
        int x1b = rectangle.x + rectangle.width;

        int y1a = y+h;
        int y1b = rectangle.y + rectangle.height;

        int x1 = Math.max(x1a, x1b);
        int y1 = Math.max(y1a, y1b);

        rectangle.x = x0;
        rectangle.y = y0;
        rectangle.width = x1 - x0;
        rectangle.height= y1 - y0;
    }
    
    /**
     * Returns a copy of the dirty rectangle of the given object, 
     * or <code>null</code> if there is no dirty rectangle 
     * associated with the given object.
     * 
     * @param object The object
     * @return A copy of the dirty rectangle of the object, 
     * or <code>null</code>
     */
    Rectangle getDirtyRectangle(T object)
    {
        Rectangle result = dirtyRectangles.get(object);
        //System.out.println("DirtyRectangle is  "+result+" for "+object);
        if (result == null)
        {
            return null;
        }
        return new Rectangle(result);
    }
    
    /**
     * Clears the dirty rectangle of the given object
     * 
     * @param object The object
     */
    void clearDirtyRectangle(T object)
    {
        dirtyRectangles.remove(object);
    }
    
    /**
     * Clear all dirty rectangles.
     */
    void clearDirtyRectangles()
    {
        dirtyRectangles.clear();
        //System.out.println("DirtyRectangles cleared");
    }
}