/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

/**
 * This class contains utility methods for creating 
 * {@link Condition Conditions} that may be applied
 * to MouseEvents.
 */
public class MouseEventConditions
{
    /**
     * Returns a new Condition that is always <code>true</code>
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> trueCondition()
    {
        return Conditions.<MouseEvent>trueCondition();
    }
    
    /**
     * Returns a new Condition that is always <code>false</code>
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> falseCondition()
    {
        return Conditions.<MouseEvent>falseCondition();
    }
    
    
    /**
     * Returns whether the given flags are set in the extended
     * modifiers of the event.
     * 
     * @param e The event
     * @param flags The flags
     * @return Whether the flags are set
     */
    private static boolean is(InputEvent e, int flags)
    {
        return (e.getModifiersEx() & flags) == flags;
    }

    /**
     * Returns a new condition that checks whether the the specified
     * button was pressed when the mouse event occurred. 
     * 
     * @param button The button. Must be 1, 2 or 3 
     * @return The new condition
     * @throws IllegalArgumentException if the button is not 1, 2 or 3
     */
    public static Condition<MouseEvent> buttonCondition(final int button)
    {
        int mask = 0;
        switch (button)
        {
            case 1: mask = MouseEvent.BUTTON1_DOWN_MASK; break;
            case 2: mask = MouseEvent.BUTTON2_DOWN_MASK; break;
            case 3: mask = MouseEvent.BUTTON3_DOWN_MASK; break;
            default:
                throw new IllegalArgumentException(
                    "Button must be 1, 2 or 3");
        }
        return flagCondition(mask);
    }
    

    /**
     * Returns a new condition that checks whether the the given 
     * flags are set in the extended modifiers of a mouse event.
     * 
     * @param flags The flags
     * @return The new condition
     */
    public static Condition<MouseEvent> flagCondition(final int flags)
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return MouseEventConditions.is(t, flags);
            }
            
            @Override
            public String toString()
            {
                return "mouseModifiers("+MouseEvent.getModifiersExText(flags)+")";
            }
        };
    }
    
    /**
     * Returns a new condition that checks whether a MouseEvent
     * is a MouseWheelEvent
     * 
     * @return The new condition
     */
    public static Condition<MouseEvent> wheelCondition()
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return (t instanceof MouseWheelEvent);
            }
            
            @Override
            public String toString()
            {
                return "wheelCondition";
            }
            
        };
    }
    
    /**
     * Returns a new condition that checks whether mouse events have
     * happened while the CTRL button was down
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> controlDownCondition()
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return t.isControlDown();
            }
            
            @Override
            public String toString()
            {
                return "controlDownCondition";
            }
        };
    }

    /**
     * Returns a new condition that checks whether mouse events have
     * happened while the ALT button was down
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> altDownCondition()
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return t.isAltDown();
            }
            
            @Override
            public String toString()
            {
                return "altDownCondition";
            }
        };
    }

    /**
     * Returns a new condition that checks whether mouse events have
     * happened while the ALT-GR button was down
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> altGraphDownCondition()
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return t.isAltGraphDown();
            }
            
            @Override
            public String toString()
            {
                return "altGraphDownCondition";
            }
            
        };
    }

    /**
     * Returns a new condition that checks whether mouse events have
     * happened while the SHIFT button was down
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> shiftDownCondition()
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return t.isShiftDown();
            }
            
            @Override
            public String toString()
            {
                return "shiftDownCondition";
            }
        };
    }

    /**
     * Returns a new condition that checks whether mouse events have
     * happened while the META button was down
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> metaDownCondition()
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return t.isMetaDown();
            }
            
            @Override
            public String toString()
            {
                return "metaDownCondition";
            }
            
        };
    }
    
    /**
     * Returns a new condition that checks whether mouse events are
     * popup triggers.
     *  
     * @return The new condition
     */
    public static Condition<MouseEvent> popupTriggerCondition()
    {
        return new Condition<MouseEvent>()
        {
            @Override
            public boolean is(MouseEvent t)
            {
                return t.isPopupTrigger();
            }
            
            @Override
            public String toString()
            {
                return "popupTriggerCondition";
            }
            
        };
    }
    
    /**
     * Private constructor to prevent instantiation.
     */
    private MouseEventConditions()
    {
    }
}
