/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.awt.Rectangle;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.Point4f;
import javax.vecmath.SingularMatrixException;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import de.javagl.swogl.geometry.DefaultRay;
import de.javagl.swogl.geometry.Ray;
import de.javagl.swogl.scene.Camera;
import de.javagl.swogl.scene.CameraUtils;
import de.javagl.swogl.scene.View;
import de.javagl.swogl.scene.ViewUtils;

/**
 * Utility methods for picking
 */
public class PickingUtils
{
    /**
     * Compute a picking {@link Ray} from the given screen coordinates
     * based on the given {@link Camera} and {@link View}.
     * 
     * @param camera The camera
     * @param view The view
     * @param x The x coordinate
     * @param y The y coordinate
     * @return The picking ray.
     */
    public static Ray computePickingRay(Camera camera, View view, int x, int y)
    {
        Matrix4f cameraMatrix = 
            CameraUtils.computeViewMatrix(camera);
        Matrix4f projectionMatrix = 
            ViewUtils.computeProjectionMatrix(camera.getFovY(), view);
        
        Rectangle viewport = view.getViewport();
        return computePickingRay(
            cameraMatrix, projectionMatrix, viewport, x, y);
    }


    /**
     * Compute a picking ray given the view matrix, projection matrix,
     * viewport and position. The view matrix contains the inverse
     * of the matrix containing the rotation and translation for the 
     * camera. The projectionMatrix contains solely the perspective 
     * projection transformation. The viewport describes the rectangle 
     * that the picking occurs in. The x and y coordinates are the 
     * position where the picking ray should start, in screen coordinates.
     * 
     * @param viewMatrix The view matrix
     * @param projectionMatrix The projection matrix
     * @param viewport The viewport
     * @param x The x position
     * @param y The y position
     * @return The picking ray
     */
    public static Ray computePickingRay(
        Matrix4f viewMatrix, Matrix4f projectionMatrix, 
        Rectangle viewport, int x, int y)
    {
        int fy = viewport.height - y;
        Point3f p0 = unProject(new Point3f(x,fy,0), 
            viewMatrix, projectionMatrix, viewport); 
        Point3f p1 = unProject(new Point3f(x,fy,1), 
            viewMatrix, projectionMatrix, viewport); 
    
        Point3f rayOrigin = p0;
        Vector3f rayDirection = new Vector3f();
        rayDirection.sub(p1, p0);
        rayDirection.normalize();
    
        Ray ray = new DefaultRay(rayOrigin, rayDirection);
        //System.out.println("Ray "+ray);
        return ray;
    }

    /**
     * Unprojects the given point (in screen coordinates) based on the
     * given view matrix and projection matrix and the viewport. 
     * 
     * @param point The point
     * @param viewMatrix The view matrix
     * @param projection The projection matrix
     * @param viewport The current viewport
     * @return The unprojected point
     */
    private static Point3f unProject(
        Tuple3f point, Matrix4f viewMatrix, 
        Matrix4f projection, Rectangle viewport)
    {
        Matrix4f matrix = new Matrix4f();
        matrix.mul(projection, viewMatrix);
        try
        {
            matrix.invert();
        }
        catch (SingularMatrixException e)
        {
            return new Point3f();
        }
        
        Point4f temp = new Point4f(point.x, point.y, point.z, 1);
        temp.x = (temp.x - viewport.x) / viewport.width;
        temp.y = (temp.y - viewport.y) / viewport.height;
        temp.x = temp.x * 2 - 1;
        temp.y = temp.y * 2 - 1;
        temp.z = temp.z * 2 - 1;

        matrix.transform(temp);
        Point3f result = new Point3f(temp.x, temp.y, temp.z);
        result.scale(1.0f / temp.w);
        return result;
    }
    

    /**
     * Private constructor to prevent instantiation
     */
    private PickingUtils()
    {
    }
}
