/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.ServiceLoader;
import java.util.logging.Logger;

/**
 * This class is responsible for resolving services that may create 
 * {@link SwoglApplication SwoglApplications}. It will resolve 
 * services that implement the {@link SwoglApplicationFactory}
 * interface, and use the factory to create a new SwoglApplication.
 */
public class SwoglApplications
{
    /**
     * The logger used in this class
     */
    private static final Logger logger = 
        Logger.getLogger(SwoglApplications.class.getName());
    
    /**
     * The ServiceLoader for SwoglApplicationFactories
     */
    private static final ServiceLoader<SwoglApplicationFactory> 
        swoglApplicationFactories = 
            ServiceLoader.load(SwoglApplicationFactory.class);
    
    
    /**
     * Creates a new {@link SwoglApplication}. This method will
     * try to obtain a service that implements the 
     * {@link SwoglApplicationFactory} interface. This service
     * will be used to instantiate a SwoglApplication. The first 
     * SwoglApplication that can successfully be instantiated
     * will be returned.
     * 
     * @return A new SwoglApplication, or <code>null</code> if no
     * service for creating a SwoglApplication could be found.
     */
    public static SwoglApplication createSwoglApplication() 
    {
        List<SwoglApplicationFactory> factories = 
            new ArrayList<SwoglApplicationFactory>();
        for (SwoglApplicationFactory factory : swoglApplicationFactories) 
        {
            logger.config("Found factory "+factory);
            factories.add(factory);
        }
        //System.out.println("Found "+factories);
        
        // Sort the list of factories: Prefer factories that
        // create applications that use shaders. 
        final String preference = "Shader";

        // For testing:
        //final String preference = "SwoglApplicationFactoryJOGL";
        //final String preference = "SwoglApplicationFactoryShaderJOGL";
        //final String preference = "SwoglApplicationFactoryLWJGL";
        //final String preference = "SwoglApplicationFactoryShaderLWJGL";
        //final String preference = "SwoglApplicationFactoryJavaGL";
        
        Collections.sort(factories, new Comparator<SwoglApplicationFactory>()
        {
            @Override
            public int compare(
                SwoglApplicationFactory f0, 
                SwoglApplicationFactory f1)
            {
                String n0 = f0.getClass().getName(); 
                String n1 = f1.getClass().getName(); 
                if ( n0.contains(preference) && !n1.contains(preference))            
                {
                    return -1;
                }
                if (!n0.contains(preference) &&  n1.contains(preference))            
                {
                    return 1;
                }
                return 0;
            }
        });
        
        //System.out.println("Sorted "+factories);
        Map<SwoglApplicationFactory, Throwable> errors = 
            new LinkedHashMap<SwoglApplicationFactory, Throwable>();
        for (SwoglApplicationFactory factory : factories) 
        {
            try
            {
                SwoglApplication swoglApplication = 
                    factory.createSwoglApplication();
                logger.config("Created "+swoglApplication);
                if (swoglApplication != null)
                {
                    return swoglApplication;
                }
                errors.put(factory, null);
            }
            catch (Exception t)
            {
                logger.config(createErrorMessageString(t));
                errors.put(factory, t);
            }
        }
        
        
        for (SwoglApplicationFactory factory : errors.keySet())
        {
            logger.severe("Failed to instantiate SwoglApplication " +
                "using factory "+factory+":");
            Throwable t = errors.get(factory);
            if (t == null)
            {
                logger.severe(
                    "The factory did not create non-null SwoglApplication");
            }
            else
            {
                logger.severe(createErrorMessageString(t));
            }
        }
        return null;
    }
    
    private static String createErrorMessageString(Throwable t)
    {
        StringBuilder sb = new StringBuilder();
        sb.append(t+"\n");
        for (StackTraceElement ste : t.getStackTrace())
        {
            sb.append("\t"+ste+"\n");
        }
        return sb.toString();
    }
    
    
    
    /**
     * Private constructor to prevent instantiation
     */
    private SwoglApplications()
    {
    }
}
