/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import javax.swing.JComponent;
import javax.vecmath.Matrix4f;

import de.javagl.swogl.geometry.Geometry;

/**
 * Interface for a Swogl component. A Swogl component contains a Swing 
 * component, as well as a {@link Texture} and a {@link Geometry}.
 * The Texture contains an image of the Swing component, and will be
 * updated automatically by Swogl. The Geometry contains all information
 * required for rendering the SwoglComponent as a 3D object, textured with
 * the Texture. <br /> 
 * <br />
 * Note to possible implementors: This interface may still be extended
 * by further methods in future versions. Specifically, by methods to
 * maintain a set of Listeners that want to be informed about transform
 * changes.
 */
public interface SwoglComponent
{
    /**
     * Returns the Swing component of this SwoglComponent
     * 
     * @return The Swing component of this SwoglComponent
     */
    JComponent getComponent();
    
    /**
     * Returns the {@link Texture} of this SwoglComponent
     * 
     * @return The {@link Texture} of this SwoglComponent
     */
    Texture getTexture();
    
    /**
     * Returns the {@link Material} of this SwoglComponent
     * 
     * @return The {@link Material} of this SwoglComponent
     */
    Material getMaterial();

    /**
     * Returns the {@link Geometry} of this SwoglComponent
     * 
     * @return The {@link Geometry} of this SwoglComponent
     */
    Geometry getGeometry();
    
    /**
     * Returns a copy of the current transform matrix of this
     * SwoglComponent.
     * 
     * @return A copy of the current transform matrix
     */
    Matrix4f getTransform();
    
    /**
     * Set the transform of this SwoglComponent to be a copy
     * of the given transform.
     * 
     * @param matrix The new transform
     */
    void setTransform(Matrix4f matrix);
    
}
