/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.swing.JComponent;
import javax.swing.JLayeredPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.SingularMatrixException;
import javax.vecmath.Vector3f;

import de.javagl.swogl.geometry.Geometry;
import de.javagl.swogl.geometry.GeometryUtils;
import de.javagl.swogl.geometry.Ray;
import de.javagl.swogl.layout.LayoutManager3D;


/**
 * This class is a container for {@link SwoglComponent}s. It is 
 * responsible for the interaction and rendering of the 
 * SwoglComponents.
 */
public final class SwoglContainer
{
    /**
     * When this flag is enabled, the contents of the SwoglComponents
     * will be shown in the lower half of the SwoglContainer
     */
    private static final boolean DEBUG_SHOW_CONTENTS = false;

    /**
     * When this flag is enabled, information about the picking
     * of SwoglComponents will be printed
     */
    private static final boolean DEBUG_PICKING = false;
    
    /**
     * When this flag is enabled, fine-grained information about
     * the event handling and conversion will be printed.
     */
    private static final boolean DEBUG_EVENTS = false;


    /**
     * A Comparator that compares {@link IntersectionInfo} objects
     * by their distance
     */
    private static final Comparator<IntersectionInfo> 
        INTERSECTION_INFO_COMPARATOR = 
            new Comparator<IntersectionInfo>()
            {
                @Override
                public int compare(
                    IntersectionInfo i0, IntersectionInfo i1)
                {
                    return Float.compare(
                        i0.getDistance(), i1.getDistance());
                }
            };
    
    /**
     * The List of SwoglComponents in this SwoglContainer
     */
    private final List<SwoglComponent> swoglComponents;

    /**
     * Stores the SwoglComponents that have been added as "popup"
     * components. (This is actually used like a Set, but 
     * implemented as an IdentityHashMap, emulating an
     * "IdentityHashSet")
     */
    private final Map<SwoglComponent, Void> popups;
    
    /**
     * The main component of this SwoglContainer
     */
    private final JComponent mainComponent;
    
    /**
     * The component that actually displays the SwoglComponents
     */
    private final Component visibleComponent;

    /**
     * The container for the contents of the individual SwoglComponents.
     * This simply is a Scrollable panel that collects all components, 
     * and which is hidden behind the visible component.
     */
    private final JPanel containerPanel;

    /**
     * Manages the dirty rectangles of the components. The dirty 
     * rectangles will be set by the RepaintManager that is 
     * established in the SwoglCore. These dirty rectangles
     * will then eventually be transferred to the SwoglComponents.
     * The SwoglComponentHandler may then update the textures according
     * to the dirty rectangles of the SwoglComponents.
     */
    private final DirtyRectangleManager<Component> dirtyRectangleManager;

    /**
     * The renderer for the SwoglComponents. Its 
     * {@link SwoglComponentHandler#handleSwoglComponent(SwoglComponent)} and
     * {@link SwoglComponentHandler#releaseSwoglComponent(SwoglComponent)}
     * methods will be called when SwoglComponents are added or removed,
     * respectively.
     */
    private final SwoglComponentHandler swoglComponentHandler;
    
    /**
     * This will passed in to the constructor, and will be
     * used to compute picking rays which will then be
     * intersected with the SwoglComponents to process
     * MouseEvents appropriately.
     */
    private final PickingRayComputer pickingRayComputer;
    
    /**
     * The 3D layout manager for this SwoglContainer.
     */
    private LayoutManager3D layoutManager3D; 

    /**
     * The MouseEventProcessor, responsible for handling the mouse
     * events that happen on this SwoglContainer.
     */
    private MouseEventProcessor mouseEventProcessor = 
        new DefaultMouseEventProcessor(
            Conditions.<MouseEvent>trueCondition(), 
            Conditions.<MouseEvent>trueCondition());
    

    /**
     * Creates a new, empty SwoglContainer for the given SwoglApplication.
     * Finally delegates to {@link #create(Component, SwoglComponentHandler, 
     * PickingRayComputer)}.
     * 
     * @param swoglApplication The SwoglApplication. May not be null.
     * @return The SwoglContainer
     */
    public static SwoglContainer create(SwoglApplication swoglApplication)
    {
        return create(swoglApplication.getRenderComponent(), swoglApplication);
    }

    /**
     * Creates a new, empty SwoglContainer that shows the given 
     * visibleComponent and obtains the SwoglComponentHandler and 
     * PickingRayComputer from the given SwoglSetup. Finally 
     * delegates to {@link #create(Component, SwoglComponentHandler, 
     * PickingRayComputer)}.
     * 
     * @param visibleComponent The visible component. May not be null.
     * @param swoglSetup The SwoglSetup. May not be null.
     * @return The SwoglContainer
     */
    public static SwoglContainer create(
        Component visibleComponent, SwoglSetup swoglSetup)
    {
        return create(visibleComponent, 
            swoglSetup.getSwoglComponentHandler(), 
            swoglSetup.getPickingRayComputer());
    }

    /**
     * Creates a new, empty SwoglContainer that shows the given 
     * visibleComponent.<br />
     * <br />
     * Note: When the visibleComponent is a <i>lightweight</i> component 
     * (e.g. a GLJPanel instead of a GLCanvas) then the visibleComponent
     * may not have any MouseListeners, MouseMotionListeners or 
     * MouseWheelListeners. These listeners should instead be attached 
     * directly to the main component of this SwoglContainer. <br />
     * <br />
     * When a SwoglComponent is added to this SwoglContainer, then the 
     * {@link SwoglComponentHandler#handleSwoglComponent(SwoglComponent)} 
     * method of the given SwoglComponentHandler will be called, and when 
     * it is removed, the 
     * {@link SwoglComponentHandler#releaseSwoglComponent(SwoglComponent)}
     * will be called. The SwoglComponentHandler is thus responsible for 
     * maintaining the rendered state of the SwoglComponents.<br />
     * <br />
     * The given PickingRayComputer will be used to compute the
     * picking ray for the interaction with the SwoglComponents:
     * When a MouseEvent occurs on this SwoglContainer, then the
     * pickingRayComputer will be used to compute the picking ray
     * which will be intersected with the SwoglComponents, and if 
     * a SwoglComponent is hit, then the MouseEvent will be passed
     * to the SwoglComponent according to the Conditions that have
     * been specified using 
     * {@link #setMouseEventDispatchingConditions(Condition, Condition)}.
     * By default, both conditions are <code>true</code>. 
     * 
     * @param visibleComponent The visible component. May not be null.
     * @param swoglComponentHandler The SwoglComponentHandler. May not be null.
     * @param pickingRayComputer The picking ray computer. May not be null.
     * @return The SwoglContainer
     */
    public static SwoglContainer create(Component visibleComponent, 
        SwoglComponentHandler swoglComponentHandler, 
        PickingRayComputer pickingRayComputer)
    {
        return new SwoglContainer(
            visibleComponent, swoglComponentHandler, pickingRayComputer);
    }
    

    /**
     * Creates a new, empty SwoglContainer with the given parameters. See 
     * {@link #create(Component, SwoglComponentHandler, PickingRayComputer)}
     * for usage details.<br />
     * 
     * @param visibleComponent The visible component. May not be null.
     * @param swoglComponentHandler The SwoglComponentHandler. May not be null.
     * @param pickingRayComputer The picking ray computer. May not be null.
     */
    private SwoglContainer(
        Component visibleComponent, 
        SwoglComponentHandler swoglComponentHandler, 
        PickingRayComputer pickingRayComputer)
    {
        if (visibleComponent == null)
        {
            throw new NullPointerException("visibleComponent may not be null");
        }
        if (swoglComponentHandler == null)
        {
            throw new NullPointerException("swoglComponentHandler may not be null");
        }
        if (pickingRayComputer == null)
        {
            throw new NullPointerException("pickingRayComputer may not be null");
        }
        this.visibleComponent = visibleComponent;
        this.swoglComponentHandler = swoglComponentHandler;
        this.pickingRayComputer = pickingRayComputer;
        
        swoglComponents = new CopyOnWriteArrayList<SwoglComponent>();
        popups = new IdentityHashMap<SwoglComponent, Void>();
        dirtyRectangleManager =  new DirtyRectangleManager<Component>();
        
        // Put the panel containing the SwoglComponents into a
        // scroll pane
        containerPanel = new ContainerPanel();
        final JScrollPane containerScrollPane = 
            new JScrollPane(containerPanel);

        // Hide the containerPanel behind the visible component.
        final JLayeredPane layeredPane = new JLayeredPane();
        
        // For debugging:
        if (DEBUG_SHOW_CONTENTS) layeredPane.setLayout(new GridLayout(2,1));
        
        layeredPane.add(containerScrollPane, Integer.valueOf(0));
        layeredPane.add(visibleComponent, Integer.valueOf(5000));

        // Ensure that the visible component and the scrollPane
        // always have the same size as this container. 
        layeredPane.addComponentListener(new ComponentAdapter()
        {
        	@Override
            public void componentResized(ComponentEvent e)
            {
        	    Rectangle bounds = layeredPane.getBounds();
                containerScrollPane.setBounds(bounds);
                SwoglContainer.this.visibleComponent.setBounds(bounds);
            }
        });

        // Adding the layered pane as the only component in the
        // main component
        mainComponent = new JPanel(new GridLayout(1,1));
        mainComponent.add(layeredPane);
        
        SwoglCore.registerSwoglContainer(this);
    }
    
    
    
//    public void setVisibleComponent(Component component)
//    {
//        if (this.visibleComponent != null)
//        {
//            throw new IllegalStateException(
//                "SwoglContainer already has a visible component");
//        }
//        this.visibleComponent = component;
//        layeredPane.add(visibleComponent, Integer.valueOf(5000));
//    }
    
//    /**
//     * Returns the visible component of this SwoglContainer, 
//     * that was given in the constructor.
//     * 
//     * @return The visible component of this SwoglContainer
//     */
//    Component getVisibleComponent()
//    {
//        return visibleComponent;
//    }
    
    /**
     * Returns the main component of this SwoglContainer. This component
     * will only show the visibleComponent (although it will contain 
     * other, hidden components)
     * 
     * @return The main component of this SwoglContainer
     */
    public JComponent getMainComponent()
    {
        return mainComponent;
    }
    
    /**
     * Interface for all classes that may process mouse events. Instances of 
     * classes implementing this interface will be used inside the 
     * SwoglContainer, and receive all mouse events from the AWT event queue. 
     * The implementations of this interface are responsible for properly 
     * forwarding the mouse events either to the SwoglComponents, or the 
     * listeners that are attached to the main component of the 
     * SwoglContainer.  
     */
    private interface MouseEventProcessor
    {
        /**
         * Process the given mouse event. This may modify the given 
         * event and/or dispatch it to other listeners. 
         * 
         * @param mouseEvent The mouse event to process.
         */
        void process(MouseEvent mouseEvent);
    }
    
    /**
     * Default implementation of a MouseEventProcessor
     */
    private class DefaultMouseEventProcessor 
        implements MouseEventProcessor
    {
        /**
         * The condition for dispatching a MouseEvent to the listeners
         */
        private final Condition<MouseEvent> listenerDispatchingCondition;
        
        /**
         * The condition for dispatching a MouseEvent to the SwoglComponents
         */
        private final Condition<MouseEvent> swoglComponentDispatchingCondition;
        
        /**
         * Creates a new DefaultMouseEventProcessor
         * 
         * @param listenerDispatchingCondition The listener dispatching 
         * condition
         * @param swoglComponentDispatchingCondition The SwoglComponent 
         * dispatching condition
         */
        DefaultMouseEventProcessor(
            Condition<MouseEvent> listenerDispatchingCondition,
            Condition<MouseEvent> swoglComponentDispatchingCondition)
        {
            this.listenerDispatchingCondition = 
                listenerDispatchingCondition;
            this.swoglComponentDispatchingCondition = 
                swoglComponentDispatchingCondition;
        }
        
        @Override
        public void process(MouseEvent mouseEvent)
        {
            boolean wasDispatchedToListeners = false;
            boolean wasDispatchedToSwoglComponent = false;
            if (listenerDispatchingCondition.is(mouseEvent))
            {
                doDispatchToListeners(mouseEvent);
                wasDispatchedToListeners = true;
            }
            if (swoglComponentDispatchingCondition.is(mouseEvent))
            {
                wasDispatchedToSwoglComponent = 
                    doDispatchToSwoglComponents(mouseEvent);
            }

            // When a mouse wheel event happens on a component, it will be
            // propagated up the containment hierarchy, and a copy of the 
            // event will dispatched to the first component that wants to 
            // process mouse wheel events. When the event was already 
            // dispatched to the listeners, and NOT to any SwoglComponent,
            // then it should not be dispatched again, and thus will be
            // consumed here.
            if (!wasDispatchedToSwoglComponent &&
                wasDispatchedToListeners && 
                mouseEvent.getID() == MouseEvent.MOUSE_WHEEL)
            {
                mouseEvent.consume();
            }
        }
    }
    
    /**
     * Set the conditions for the dispatching of MouseEvents to either the
     * listeners that are attached to the main component of this 
     * SwoglContainer, or to the SwoglComponents inside this SwoglContainer.
     * If either of the conditions is 'null', then the negation of the other
     * condition will be used in this place. If both conditions are null,
     * then the conditions are set to the 'true' condition, accepting all
     * events.
     * 
     * @param listenerDispatchingCondition The condition that checks whether
     * an event should be dispatched to the listeners that are attached to
     * the main component.
     * @param swoglComponentDispatchingCondition The condition that checks
     * whether an event should be dispatched to the SwoglComponents
     */
    public void setMouseEventDispatchingConditions(
        Condition<MouseEvent> listenerDispatchingCondition,
        Condition<MouseEvent> swoglComponentDispatchingCondition)
    {
        if (listenerDispatchingCondition == null &&
            swoglComponentDispatchingCondition == null)
        {
            this.mouseEventProcessor = new DefaultMouseEventProcessor(
                Conditions.<MouseEvent>trueCondition(), 
                Conditions.<MouseEvent>trueCondition());
        }
        else if (listenerDispatchingCondition == null)
        {
            this.mouseEventProcessor = new DefaultMouseEventProcessor(
                Conditions.not(swoglComponentDispatchingCondition), 
                swoglComponentDispatchingCondition);
        }
        else if (swoglComponentDispatchingCondition == null)
        {
            this.mouseEventProcessor = new DefaultMouseEventProcessor(
                listenerDispatchingCondition, 
                Conditions.not(listenerDispatchingCondition));
        }
        else
        {
            this.mouseEventProcessor = new DefaultMouseEventProcessor(
                listenerDispatchingCondition, 
                swoglComponentDispatchingCondition);
        }
    }
    
    /**
     * Handles all mouse events that happen on the visible component. 
     * 
     * @param mouseEvent The event to handle
     */
    void handleMouseEvent(MouseEvent mouseEvent)
    {
        if (DEBUG_EVENTS) System.out.println("");
        debugEvent("handleMouseEvent ", mouseEvent);
        //System.out.println("handleMouseEvent "+mouseEvent);
        
        mouseEventProcessor.process(mouseEvent);
        debugEvent("After handling ", mouseEvent);
    }
    
    /**
     * Dispatches the given event to all event listeners that are
     * registered for the main component 
     * 
     * @param mouseEvent The event to dispatch
     */
    private void doDispatchToListeners(MouseEvent mouseEvent)
    {
        debugEvent("Dispatch to Listeners ", mouseEvent);

        Component originalSource = mouseEvent.getComponent();
        convertMouseEvent(mouseEvent, mainComponent);
        EventUtils.dispatchToListeners(mainComponent, mouseEvent);
        convertMouseEvent(mouseEvent, originalSource);
    }
    
    
    
    /**
     * Tries to dispatch the given event to the SwoglComponents. This 
     * method will compute the {@link PickingInfo} for the mouse position. 
     * If the PickingInfo is non-null, then a SwoglComponent has been hit. 
     * The MouseEvent will be transformed to originate from the window 
     * position of the hit component, and may afterwards be handled by the 
     * usual event dispatching mechanism. If a SwoglComponent was hit,
     * then <code>true</code> is returned (and in this case, the event 
     * will be modified in a way so that it should not be used by other
     * listeners any more). Otherwise, <code>false</code> is returned.
     * 
     * @param mouseEvent The mouse event
     * @return Whether a SwoglComponent has been hit and the event
     * has been transformed to the target component.
     */
    private boolean doDispatchToSwoglComponents(MouseEvent mouseEvent)
    {
        debugEvent("Dispatch to SwoglComponents", mouseEvent);

        Component originalSource = mouseEvent.getComponent();
        convertMouseEvent(mouseEvent, mainComponent);
        
        // Compute the PickingInfo, to find out if (and where) a 
        // SwoglComponent has been hit.
        PickingInfo pickingInfo = 
            computePickingInfo(mouseEvent.getPoint());
        
        if (pickingInfo != null)
        {
            //System.out.println("PickingInfo "+pickingInfo);

            JComponent swoglComponentContents = pickingInfo.getComponent();
            Point inSwoglComponent = pickingInfo.getPoint();

            // Ensure that the respective point is visible - otherwise,
            // the events will not reach the desired target component.
            // Check if the currently visible rectangle contains the
            // point. If it does not contain the point, adjust the 
            // visible rectangle so that the point is in its center.
            // This way, the panel does not continuously scroll during
            // a mouse movement, but only adjusts the scrolling 
            // position when necessary.
            Point visible = new Point();
            visible.x = swoglComponentContents.getX() + inSwoglComponent.x;
            visible.y = swoglComponentContents.getY() + inSwoglComponent.y;
            Rectangle visibleRectangle = containerPanel.getVisibleRect();
            if (!visibleRectangle.contains(visible))
            {
                visibleRectangle.x = visible.x - visibleRectangle.width / 2;
                visibleRectangle.y = visible.y - visibleRectangle.height / 2;
                containerPanel.scrollRectToVisible(visibleRectangle);
            }

            // Cause the event to be handled as it if originated from
            // the window that contains the SwoglComponent. This will
            // cause the event to reach the underlying Swing component
            // since it will not be handled by the visible component.
            Window window = 
                SwingUtilities.getWindowAncestor(mainComponent);
            //System.out.println("Window "+window);
            Point newPoint = SwingUtilities.convertPoint(
                swoglComponentContents, inSwoglComponent, window);
            mouseEvent.translatePoint(
                newPoint.x - mouseEvent.getX(), 
                newPoint.y - mouseEvent.getY());
            mouseEvent.setSource(window);
            
            return true;
        }

        convertMouseEvent(mouseEvent, originalSource);
        return false;
    }
    
    
    /**
     * Computes the picking {@link Ray} that results from a picking at
     * the given point. The direction of the returned ray is normalized.
     * 
     * @param pointInSwoglContainer The point inside this SwoglContainer
     * @return The picking ray
     */
    public Ray computePickingRay(Point pointInSwoglContainer)
    {
        int x = pointInSwoglContainer.x;
        int y = pointInSwoglContainer.y;
        Ray ray = pickingRayComputer.computePickingRay(x,y);
        return ray;
    }
    
    /**
     * Compute the {@link IntersectionInfo IntersectionInfos} for the given 
     * point. A picking ray will be computed and intersected with all 
     * SwoglComponents. For each intersection, an IntersectionInfo will be 
     * created. An (possibly empty) list of all IntersectionInfos 
     * (in unspecified order) will be returned. 
     * 
     * @param pointInSwoglContainer The point inside this SwoglContainer
     * @return The resulting IntersectionInfos
     */
    private List<IntersectionInfo> computeIntersectionInfosInternal(
        Point pointInSwoglContainer)
    {
        // Compute the picking ray
        int x = pointInSwoglContainer.x;
        int y = pointInSwoglContainer.y;
        Ray ray = pickingRayComputer.computePickingRay(x,y);
        if (DEBUG_PICKING) System.out.printf ("Ray %4d %4d    %s\n", x, y, String.valueOf(ray));

        // Find the intersections of the ray with SwoglComponents  
        Point3f p0 = new Point3f();
        Point3f p1 = new Point3f();
        Point3f p2 = new Point3f();
        Point3f result = new Point3f();
        Point3f bary = new Point3f();
        List<IntersectionInfo> intersectionInfos = 
            new ArrayList<IntersectionInfo>();
        for (SwoglComponent swoglComponent : swoglComponents)
        {
            // Convert the picking ray to the local coordinate
            // system of the SwoglComponent
            Matrix4f transform = swoglComponent.getTransform();
            try
            {
                transform.invert();
            }
            catch (SingularMatrixException e)
            {
                continue;
            }
            Point3f origin = ray.getOrigin();
            Vector3f direction = ray.getDirection();
            transform.transform(origin);
            transform.transform(direction);
            
            // NO normalization here, otherwise the distance
            // will be wrong if the transform contains a
            // scaling component != 1.
            //direction.normalize(); 
            
            // Intersect the picking ray with the triangles
            // of the geometry
            Geometry geometry = swoglComponent.getGeometry();
            for (int i=0; i<geometry.getNumTriangles(); i++)
            {
                GeometryUtils.getVertices(geometry, i, p0, p1, p2);
                
                boolean intersect = Intersection.RayTriangle.intersect(
                    origin, direction, p0, p1, p2, result);
                if (intersect)
                {
                    // Compute the barycentric coordinates of the intersection
                    bary.x = (1-result.x-result.y);
                    bary.y = result.x;
                    bary.z = result.y;

                    IntersectionInfo intersectionInfo = 
                        new IntersectionInfo(
                            swoglComponent, result.z, i, bary);
                    intersectionInfos.add(intersectionInfo);
                }
            }
        }
        return intersectionInfos;
    }
    
    
    
    /**
     * Compute the {@link IntersectionInfo IntersectionInfos} for the given 
     * point. A picking ray will be computed and intersected with all 
     * SwoglComponents. For each intersection, an IntersectionInfo will be 
     * created. An unmodifiable (possibly empty) list of all 
     * IntersectionInfos will be returned. The list will contain the 
     * IntersectionInfos ordered by their distance along the picking ray,
     * with the closest intersection first. 
     * 
     * @param pointInSwoglContainer The point inside this SwoglContainer
     * @return The resulting IntersectionInfos
     */
    public List<IntersectionInfo> computeIntersectionInfos(
        Point pointInSwoglContainer)
    {
        List<IntersectionInfo> intersectionInfos = 
            computeIntersectionInfosInternal(pointInSwoglContainer);
        if (intersectionInfos.size() > 1)
        {
            //for (IntersectionInfo i : intersectionInfos)
            //{
            //    System.out.println("Distance "+i.getDistance()+" for "+
            //        i.createPickingInfo().getComponent().getName());
            //}
            Collections.sort(intersectionInfos, INTERSECTION_INFO_COMPARATOR);
        }
        return Collections.unmodifiableList(intersectionInfos);
        
    }
    
    
    /**
     * Compute the closest {@link IntersectionInfo} for the given point. 
     * A picking ray will be computed and intersected with all 
     * SwoglComponents. If the ray intersects at least one SwoglComponent,
     * then the IntersectionInfo for the closest intersection will be
     * returned. Otherwise, <code>null</code> will be returned.  
     * 
     * @param pointInSwoglContainer The point inside this SwoglContainer
     * @return The closest IntersectionInfo, or null
     */
    public IntersectionInfo computeIntersectionInfo(Point pointInSwoglContainer)
    {
        List<IntersectionInfo> intersectionInfos = 
            computeIntersectionInfos(pointInSwoglContainer);
        IntersectionInfo closestIntersectionInfo = null;
        float minDistance = Float.POSITIVE_INFINITY;
        for (IntersectionInfo intersectionInfo : intersectionInfos)
        {
            if (intersectionInfo.getDistance() < minDistance)
            {
                closestIntersectionInfo = intersectionInfo;
                minDistance = intersectionInfo.getDistance();
            }
        }
        if (DEBUG_PICKING) System.out.println("IntersectionInfo "+closestIntersectionInfo);
        return closestIntersectionInfo;
    }
    
    /**
     * Compute the PickingInfo for the given point. A picking ray will be 
     * computed and intersected with all SwoglComponents. If at least one
     * SwoglComponent is hit by the ray, then the PickingInfo describing
     * the closest hit point will be returned.
     * 
     * @param pointInSwoglContainer The point inside this SwoglContainer
     * @return The resulting PickingInfo, or null
     */
    public PickingInfo computePickingInfo(Point pointInSwoglContainer)
    {
        IntersectionInfo intersectionInfo = 
            computeIntersectionInfo(pointInSwoglContainer);
        PickingInfo pickingInfo = null;
        if (intersectionInfo != null)
        {
            pickingInfo = intersectionInfo.createPickingInfo();
        }
        if (DEBUG_PICKING) System.out.println("PickingInfo      "+pickingInfo);
        return pickingInfo;
    }
    
    
    
    
    
    

    
    /**
     * Returns whether the given component is contained in 
     * the visible component of this SwoglContainer (or 
     * if it is the visible component itself)
     * 
     * @param component The component to check
     * @return Whether the given component is contained
     * in the visible component or is the visible component
     */
    boolean containsAsVisibleComponent(Component component)
    {
        if (component != null) 
        {
            return SwingUtilities.isDescendingFrom(
                component, visibleComponent);
        }
        return false;
    }


    /**
     * Returns the SwoglComponent that contains the given component,
     * or null if no such component exists
     * 
     * @param c The component whose containing SwoglComponent should
     * be found
     * @return The SwoglComponent that contains the given component,
     * or null if no such SwoglComponent exists
     */
    SwoglComponent findSwoglComponentContaining(Component c)
    {
        for (SwoglComponent swoglComponent : swoglComponents)
        {
            if (swoglComponent.getComponent() == c)
            {
                return swoglComponent;
            }
            if (swoglComponent.getComponent().isAncestorOf(c))
            {
                return swoglComponent;
            }
        }
        return null;
    }
    

    /**
     * Add the specified dirty rectangle to the given component. 
     * The final dirty rectangle will be repainted when 
     * {@link SwoglContainer#paintDirtyRectangles()}
     * is called.
     * 
     * @param component The component to add the dirty rectangle to
     * @param x The x coordinate of the dirty rectangle.
     * @param y The y coordinate of the dirty rectangle.
     * @param w The width of the dirty rectangle.
     * @param h The height of the dirty rectangle.
     */
    void addDirtyRectangle(Component component, int x, int y, int w, int h)
    {
        dirtyRectangleManager.addDirtyRectangle(component, x, y, w, h);
    }
    
    
    /**
     * Will paint all dirty rectangles. That is, the dirty rectangles of all 
     * components will be painted into the Texture, and the dirty 
     * rectangles will be cleared.
     */
    void paintDirtyRectangles()
    {
        for (SwoglComponent swoglComponent : swoglComponents)
        {
            paintDirtyRectangle(swoglComponent);
        }
        dirtyRectangleManager.clearDirtyRectangles();
    }

    /**
     * The dirty rectangle of the given component will be painted 
     * into the Texture, and the respective area of the 
     * Texture will be marked as dirty, so that it may be
     * updated before the next rendering pass.
     * 
     * @param component The component to refresh
     */
    private void paintDirtyRectangle(SwoglComponent swoglComponent)
    {
        JComponent component = swoglComponent.getComponent();
        Rectangle rectangle = dirtyRectangleManager.getDirtyRectangle(component);
        if (rectangle != null)
        {
            Texture texture = swoglComponent.getTexture();
            BufferedImage bufferedImage = texture.getImage();
            int x = rectangle.x;
            int y = rectangle.y;
            int w = rectangle.width;
            int h = rectangle.height;
            
            //System.out.println("Painting "+x+" "+y+" "+w+" "+h+"      of "+component);

            Graphics2D g = bufferedImage.createGraphics();
            g.setClip(x, y, w, h);
            
            // Clear the alpha component of the target rectangle, to 
            // properly paint semi-transparent components
            g.setComposite(AlphaComposite.Clear);
            g.setColor(new Color(0,0,0,0));
            g.fill(rectangle);
            g.setComposite(AlphaComposite.SrcOver);
            
            component.paint(g);
            g.dispose();
            //System.out.println("Painting "+x+" "+y+" "+w+" "+h+" DONE of "+component);
            texture.addDirtyRectangle(rectangle);
        }
    }


    
    
    
    
    /**
     * Set the LayoutManager3D that should be used by this 
     * SwoglContainer to lay out its SwoglComponents
     * 
     * @param newLayoutManager3D The LayoutManager3D to use
     */
    public synchronized void setLayout3D(LayoutManager3D newLayoutManager3D)
    {
        //System.out.println("setLayout3D "+layoutManager3D);
        
        if (layoutManager3D != newLayoutManager3D)
        {
            if (layoutManager3D != null)
            {
                layoutManager3D.setSwoglContainer(null);
            }
            layoutManager3D = newLayoutManager3D;
            if (layoutManager3D != null)
            {
                layoutManager3D.setSwoglContainer(this);
                layoutManager3D.doLayout3D();
                repaint();
            }
        }
    }
    
    /**
     * Returns the current LayoutManager3D
     * 
     * @return The current LayoutManager3D
     */ 
    public LayoutManager3D getLayout3D()
    {
        return layoutManager3D;
    }
    
    /**
     * Will cause the LayoutManager3D of this SwoglContainer
     * to be called in order to update the layout of the
     * SwoglComponents. Will do nothing if there currently
     * is no layout manager set.<br />
     * <br />
     * This method may be called from any thread.
     */
    public synchronized void validate3D()
    {
        if (layoutManager3D == null)
        {
            return;
        }
        if (SwingUtilities.isEventDispatchThread())
        {
            layoutManager3D.doLayout3D();
            repaint();
        }
        else
        {
            SwingUtilities.invokeLater(new Runnable()
            {
                @Override
                public void run()
                {
                    layoutManager3D.doLayout3D();
                    repaint();
                }
            });
        }
    }

    /**
     * Adds the given SwoglComponent to this SwoglContainer
     * 
     * @param swoglComponent The SwoglComponent to add
     */
    public void add(SwoglComponent swoglComponent)
    {
        swoglComponentHandler.handleSwoglComponent(swoglComponent);
        swoglComponents.add(swoglComponent);
        
        final JComponent component = swoglComponent.getComponent();
        addDirtyRectangle(component, 0,0, 
            component.getWidth(), component.getHeight());
        containerPanel.add(component);
        mainComponent.validate();
        
        if (layoutManager3D != null)
        {
            layoutManager3D.doLayout3D();
        }
        repaint();

        // Under some conditions components are not repainted when they are 
        // added to their parent (e.g. JLabels inside JPanels with GridLayout) 
        // TODO: This looks like bogus. Find a different solution for this.
        final ComponentListener componentListener = new ComponentAdapter()
        {
            @Override
            public void componentResized(ComponentEvent e)
            {
                component.repaint();
                component.removeComponentListener(this);
            }
        }; 
        component.addComponentListener(componentListener);
        
    }
    
    /**
     * Add the given SwoglComponent as a popup. Popups are not
     * maintainable by the client, but only added and removed
     * internally by the SwoglPopup class. 
     * 
     * @param swoglComponent The SwoglComponent to add
     */
    void addPopup(SwoglComponent swoglComponent)
    {
        //System.out.println("Added popup "+swoglComponent);
        
        popups.put(swoglComponent, null);
        add(swoglComponent);
    }
    

    /**
     * Removes the given SwoglComponent from this SwoglContainer
     * 
     * @param swoglComponent The SwoglComponent to remove
     */
    public void remove(SwoglComponent swoglComponent)
    {
        //System.out.println("Removed "+swoglComponent);
        
        swoglComponentHandler.releaseSwoglComponent(swoglComponent);
        popups.remove(swoglComponent);
        swoglComponents.remove(swoglComponent);
        containerPanel.remove(swoglComponent.getComponent());
        if (layoutManager3D != null)
        {
            layoutManager3D.doLayout3D();
        }
        mainComponent.validate();
        mainComponent.repaint();
    }
    
    /**
     * Removes the SwoglComponent with the given index
     * 
     * @param index The index of the SwoglComponent to remove
     */
    public void remove(int index)
    {
        remove(getSwoglComponent(index));
    }
    
    /**
     * Removes all SwoglComponents 
     */
    public void removeAll()
    {
        int n = getSwoglComponentCount();
        for (int i=0; i<n; i++)
        {
            SwoglComponent swoglComponent = getSwoglComponent(0);
            swoglComponentHandler.releaseSwoglComponent(swoglComponent);
            swoglComponents.remove(swoglComponent);
            containerPanel.remove(swoglComponent.getComponent());
        }
        if (layoutManager3D != null)
        {
            layoutManager3D.doLayout3D();
        }
        mainComponent.validate();
        mainComponent.repaint();
    }
    
    
    
    /**
     * Returns the number of SwoglComponents that are currently
     * present in this SwoglContainer
     * 
     * @return The number of SwoglComponents in this SwoglContainer
     */
    public int getSwoglComponentCount()
    {
        int count = 0;
        for (SwoglComponent swoglComponent : swoglComponents)
        {
            if (!popups.containsKey(swoglComponent))
            {
                count++;
            }
        }
        return count;
    }

    /**
     * Returns the SwoglComponent with the given index
     * 
     * @param index The index of the SwoglComponent
     * @return The SwoglComponent with the given index
     */
    public SwoglComponent getSwoglComponent(int index)
    {
        int idx = -1;
        for (SwoglComponent swoglComponent : swoglComponents)
        {
            if (!popups.containsKey(swoglComponent))
            {
                idx++;
            }
            if (idx == index)
            {
                return swoglComponent;
            }
        }
        throw new IndexOutOfBoundsException("Index "+index+" out of bounds");
    }
    
    
    
    
    /**
     * Causes a repaint of the visible component of this SwoglContainer.
     */
    public void repaint()
    {
        //super.repaint();
        //System.out.println("Repaint "+visibleComponent);
        visibleComponent.repaint();
    }

    
    /**
     * Utility function that converts the given mouse events so that it seems 
     * to originate from the given target component
     * 
     * @param mouseEvent The mouse event
     * @param targetComponent The target component
     */
    private static void convertMouseEvent(
        MouseEvent mouseEvent, Component targetComponent)
    {
        Component source = mouseEvent.getComponent();
        if (source == targetComponent)
        {
            return;
        }
        
        debugEvent("Converting from ", mouseEvent);
        
        Point point = mouseEvent.getPoint();
        Point converted = SwingUtilities.convertPoint(
            source, point, targetComponent);
        mouseEvent.translatePoint(
            converted.x - point.x, 
            converted.y - point.y);
        mouseEvent.setSource(targetComponent);
        
        debugEvent("to ", mouseEvent);
    }
    

    /**
     * Prints a debugging message about the given event
     * @param info The info string
     * @param mouseEvent The event
     */
    private static void debugEvent(String info, MouseEvent mouseEvent)
    {
        if (DEBUG_EVENTS) 
        {
            String pad = String.format("%"+(30-info.length())+"s", " ");
            System.out.println(info+pad+EventUtils.createInfoString(mouseEvent));
        }
    }
    
    
//    private boolean focusOwnerWillReceiveMouseEvents()
//    {
//        Component focusOwner = 
//            KeyboardFocusManager.
//                getCurrentKeyboardFocusManager().getFocusOwner();
//        return willReceiveMouseEventsRec(focusOwner);
//    }
//    
//    private boolean willReceiveMouseEventsRec(Component component)
//    {
//        if (component == null)
//        {
//            return false;
//        }
//        MouseWheelListener[] mouseWheelListeners = 
//            component.getMouseWheelListeners();
//        if (mouseWheelListeners.length > 0)
//        {
//            return true;
//        }
//        Component parent = component.getParent();
//        return willReceiveMouseEventsRec(parent);
//    }
    
}
