/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.EventQueue;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.WeakHashMap;

import javax.swing.JComponent;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.RepaintManager;
import javax.swing.SwingUtilities;

/**
 * The core class for Swogl, performing all static initializations.
 */
class SwoglCore
{
    /**
     * A WeakHashMap containing the SwoglContainers that have been
     * created so far, and not yet been garbage collected
     */
    private static WeakHashMap<SwoglContainer, Object> swoglContainers = 
        new WeakHashMap<SwoglContainer, Object>();

    /**
     * Whether the initialization has already been performed
     */
    private static volatile boolean initialized = false;
    
    /**
     * Static initialization
     */
    static
    {
        init();
    }
    
    /**
     * Execute the static initialization
     */
    private static void init()
    {
        if (!initialized)
        {
            // Setting this property to 'false' will prevent the popup
            // location from being adjusted to fit into the screen.
            // This is only required when a popup is about to be 
            // opened that will not fit into the screen. Unfortunately,
            // this does not prevent the popup to be adjusted so that
            // it pops up "upwards". 
            System.setProperty(
                "javax.swing.adjustPopupLocationToFit", "false");
            
            initMouseEventQueue();
            initRepaintManager();
            initPopupFactory();
            initialized = true;
        }
    }

    /**
     * Register the given SwoglContainer
     *
     * @param swoglContainer The SwoglContainer to register
     */
    static synchronized void registerSwoglContainer(
        SwoglContainer swoglContainer)
    {
        WeakHashMap<SwoglContainer, Object> newSwoglContainers = 
            new WeakHashMap<SwoglContainer, Object>(swoglContainers); 
        newSwoglContainers.put(swoglContainer, new Object());
        swoglContainers = newSwoglContainers;
    }

    /**
     * Initializes the EventQueue which catches ALL AWT events and checks
     * how to handle them if they originated from a SwoglContainer.<br />
     * <br />
     * If the event is a MouseEvent, it will be passed to the 
     * handleMouseEvent method of the SwoglContainer under the mouse.
     * This will transform the mouse (and possibly the contents of the
     * SwoglContainer) so that the right component is hit when the
     * Event is dispatched normally afterwards. 
     */
    private static void initMouseEventQueue()
    {
        // This function should be called implicitly on the EDT,
        // because SwoglContainers should (like any other Swing
        // Component) only be created on the EDT. However, when
        // this function is called on the EDT, there may be a
        // short period of time where TWO Event Dispatch Threads
        // exist, because the EventQueue#push method will cause
        // a new EDT to be created, and the other one to be shut
        // down lazily. In this short time, the two EDTs may 
        // concurrently access the rendering resources, which 
        // may cause errors. To avoid this, the new event queue
        // is pushed on to the event queue in the event dispatch
        // thread. Hopefully, this makes sense. In any case, the
        // threading behavior of EventQueue#push is not specified 
        // and will change in Java 7 (build 102). Also see
        // //http://bugs.sun.com/view_bug.do?bug_id=6424157 
        if (SwingUtilities.isEventDispatchThread())
        {
            SwingUtilities.invokeLater(new Runnable()
            {
                @Override
                public void run()
                {
                    doInitMouseEventQueue();
                }
            });
        }
        else
        {
            doInitMouseEventQueue();
        }
    }

    /**
     * Implementation of {@link #initMouseEventQueue()}
     */
    private static void doInitMouseEventQueue()
    {
        EventQueue eventQueue = 
            Toolkit.getDefaultToolkit().getSystemEventQueue();
        eventQueue.push(new EventQueue()
        {
            @Override
            public void dispatchEvent(AWTEvent awtEvent)
            {
                //System.err.println("Dispatching on queue: "+EventUtils.createInfoString(awtEvent)+" Event: "+awtEvent);
                if (awtEvent instanceof MouseEvent)
                {
                    MouseEvent mouseEvent = (MouseEvent)awtEvent;
                    //System.out.println("Dispatching mouse event: "+EventUtils.createInfoString(awtEvent));
                    SwoglContainer swoglContainer = 
                        findSwoglContainerFor(mouseEvent);
                    if (swoglContainer != null)
                    {
                        //System.out.println("Handle mouse event on "+swoglContainer);
                        swoglContainer.handleMouseEvent(mouseEvent);
                        
                        if (mouseEvent.isConsumed())
                        {
                            //System.out.println("Mouse event was consumed "+EventUtils.createInfoString(awtEvent));
                            return;
                        }
                    }
                    //System.out.println("Normally handling mouse event "+EventUtils.createInfoString(awtEvent));
                }
                super.dispatchEvent(awtEvent);
            }
        });
    }
    
    /**
     * Returns the SwoglContainer that contains the component which contains
     * the source component of the mouse event, or null if the MouseEvent did 
     * not happen on such a component.
     *
     * @param mouseEvent The MouseEvent
     * @return The SwoglContainer as described above.
     */
    private static SwoglContainer findSwoglContainerFor(
        MouseEvent mouseEvent)
    {
        return findSwoglContainerFor(
            mouseEvent.getComponent(),
            mouseEvent.getX(),
            mouseEvent.getY());
    }

    /**
     * Returns the SwoglContainer that contains the component which is
     * the deepest component at the given position of the given 
     * component as its visible component or a descendant of the visible
     * component, or null if there is no such SwoglContainer.
     *
     * @param sourceComponent The component
     * @param x The x-position
     * @param y The y-position
     * @return The SwoglContainer as described above.
     */
    static SwoglContainer findSwoglContainerFor(
        Component sourceComponent, int x, int y)
    {
        // Note: In many cases, this mechanism could be simplified, by 
        // checking whether the sourceComponent is identical 
        // (<code>==</code>) to the visible component of the SwoglContainer. 
        // However, for lightweight visible components or for the case that 
        // the SwoglContainer is in "debug" mode, different constellations 
        // of "containment" are possible, and these are covered here.
        Component component = 
            SwingUtilities.getDeepestComponentAt(sourceComponent, x, y);

        //System.out.println("At "+x+" "+y+" found\n"+component);
        
        for (SwoglContainer swoglContainer : swoglContainers.keySet())
        {
            if (swoglContainer.containsAsVisibleComponent(component))
            {
                //System.out.println("At "+x+" "+y+" for "+component);
                //System.out.println("found "+swoglContainer);

                return swoglContainer;
            }
        }
        return null;
    }

    /**
     * Will initialize the RepaintManager that causes the dirty regions of the
     * SwoglComponents to be repainted into the {@link PaintTarget} which 
     * they are associated with via the {@link ComponentManager}.
     */
    private static void initRepaintManager()
    {
        RepaintManager.setCurrentManager(new RepaintManager()
        {
            /**
             * The SwoglContainers that contain components that
             * have been refreshed, and thus may need a repaint
             * on their own.
             */
            private Set<SwoglContainer> pendingRepaints = 
                new LinkedHashSet<SwoglContainer>();
            
            /**
             * Temporarily stores rectangle data for convenience 
             */
            private Rectangle rectangle = new Rectangle();
            
            /**
             * Whether this RepaintManager is currently 
             * painting dirty regions.
             */
            private boolean isPaintingDirtyRegions = false;

            /**
             * It will be checked if the given component is contained in a 
             * component that is located in a SwoglComponent. If it is 
             * contained in a SwoglComponent, the dirty rectangle will be 
             * transformed into the coordinate system of this component and
             * used to mark the component as 'dirty'. When paintDirtyRegions 
             * is called, this region will be refreshed. This will cause
             * the corresponding dirty rectangle of the Texture to
             * be set, which in turn will be updated before the next
             * rendering happens.
             */
            @Override
            public void addDirtyRegion(
                JComponent c, int x, int y, int w, int h)
            {
                super.addDirtyRegion(c, x, y, w, h);

                //System.out.println("addDirtyRegion " + x + " " + y + " " + w + " " + h + " on " + c);
                if (isPaintingDirtyRegions)
                {
                    //System.out.println("Ignoring");
                    return;
                }
                if (!isMainComponent(c))
                {
                    //System.out.println("addDirtyRegion " + x + " " + y + " " + w + " " + h + " on " + c);
                    
                    SwoglContainer swoglContainer = 
                        findSwoglContainerOwning(c);
                    
                    //System.out.println("In "+swoglContainer);
                    if (swoglContainer != null)
                    {
                        SwoglComponent swoglComponent = 
                            swoglContainer.findSwoglComponentContaining(c);
                        //System.out.println("comp "+swoglComponent);
                        if (swoglComponent != null)
                        {
                            JComponent component = swoglComponent.getComponent();
                        	rectangle.x = x;
                            rectangle.y = y;
                            rectangle.width = w;
                            rectangle.height = h; 
                            Rectangle finalRectangle = 
                                SwingUtilities.convertRectangle(
                                    c, rectangle, component);
                            //System.out.println("addDirtyRegion final "+finalRectangle+" on " + c);
                            swoglContainer.addDirtyRectangle(
                                component, 
                                finalRectangle.x, finalRectangle.y, 
                                finalRectangle.width, finalRectangle.height);

                            pendingRepaints.add(swoglContainer);
                        }
                    }
                }
            }

            /**
             * Will cause the dirty regions of all SwoglContainers 
             * which contain dirty regions to be refreshed by calling
             * {@link SwoglContainer#paintDirtyRectangles()}
             */
            @Override
            public void paintDirtyRegions()
            {
                isPaintingDirtyRegions = true;
                super.paintDirtyRegions();
                //System.out.println("SwoglCore paintDirtyRegions()");
                for (SwoglContainer swoglContainer : pendingRepaints)
                {
                    swoglContainer.paintDirtyRectangles();
                    //System.out.println("Trigger pending repaint");
                    swoglContainer.repaint();
                    //System.out.println("Trigger pending repaint done");
                }
                pendingRepaints.clear();
                isPaintingDirtyRegions = false;
            }
            
        });
    }

    /**
     * Initializes the PopupFactory, which creates SwoglPopups instead
     * of usual Popups when the "owner" of the Popup is a descendant
     * of a SwoglContainer
     */
    private static void initPopupFactory()
    {
        PopupFactory.setSharedInstance(new PopupFactory()
        {
            @Override
            public Popup getPopup(
                Component owner, Component contents, int x, int y)
            {
                if (contents instanceof JComponent)
                {
                    JComponent contentsComponent = (JComponent)contents;
                    
                    SwoglContainer swoglContainer = 
                        findSwoglContainerOwning(owner);
    
                    //System.out.println("Popup for SwoglContainer "+swoglContainer);
                    //System.out.println("    at "+x+" "+y);
                    //System.out.println("    owner    "+owner);
                    //System.out.println("    contents "+contents);
    
                    if (swoglContainer != null)
                    {
                        return new SwoglPopup(
                            owner, contentsComponent, x,y, swoglContainer);
                    }
                }
                return super.getPopup(owner, contents, x, y);
            }

        });
    }


    /**
     * Returns the SwoglContainer whose main component is an ancestor of the 
     * given component, or null if no such SwoglContainer exists
     *
     * @param c The component whose ancestor should be returned
     */
    private static SwoglContainer findSwoglContainerOwning(Component c)
    {
        for (SwoglContainer swoglContainer : swoglContainers.keySet())
        {
            if (swoglContainer.getMainComponent().isAncestorOf(c))
            {
                return swoglContainer;
            }
        }
        return null;
    }
    
    /**
     * Returns whether the given component is the main component
     * of a SwoglContainer
     * 
     * @param c The component
     * @return Whether the component is a main component
     */
    private static boolean isMainComponent(Component c)
    {
        for (SwoglContainer swoglContainer : swoglContainers.keySet())
        {
            if (swoglContainer.getMainComponent() == c)
            {
                return true;
            }
        }
        return false;
    }

    

}
