/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;

import javax.swing.JComponent;
import javax.swing.Popup;
import javax.swing.SwingUtilities;
import javax.vecmath.Matrix4f;

import de.javagl.swogl.geometry.Box;
import de.javagl.swogl.geometry.Geometry;
import de.javagl.swogl.geometry.GeometryUtils;

/**
 * This class is a Popup that may be used to show temporary, "popping-up" 
 * components inside a {@link SwoglContainer}. Instances of this class are 
 * returned by the shared PopupFactory instance that is established in 
 * the {@link SwoglCore}, if the owner of the Popup is a descendant of a 
 * SwoglContainer
 */
class SwoglPopup extends Popup
{
    /**
     * The owner of the Popup that is created, e.g. a JComboBox
     */
    private Component owner;
    
    /**
     * The contents of the Popup
     */
    private JComponent contents;
    
    /**
     * The SwoglContainer that contains the owner
     */
    private SwoglContainer swoglContainer;
    
    /**
     * The SwoglComponent that displays the popup contents
     */
    private SwoglComponent swoglPopupComponent;
    
    /**
     * The x coordinate on the screen(!) where the Popup should occur
     */
    private int x;
    
    /**
     * The y coordinate on the screen(!) where the Popup should occur
     */
    private int y;
    
    /**
     * Creates a new SwoglPopup, whose show/hide method will cause
     * a SwoglComponent with the given contents to be placed in the
     * given SwoglContainer
     * 
     * @param owner The owner of the Popup
     * @param contents The contents of the Popup
     * @param x The x-coordinate of the popup on the screen(!)
     * @param y The y-coordinate of the popup on the screen(!)
     * @param swoglContainer The SwoglContainer to which the 
     * SwoglComponent displaying the contents will be added
     */
    public SwoglPopup(Component owner, JComponent contents, int x, int y, 
                      SwoglContainer swoglContainer)
    {
        this.x = x;
        this.y = y;
        this.owner = owner;
        this.contents = contents;
        this.swoglContainer = swoglContainer;
    }
    
    /**
     * Will show this SwoglPopup. A SwoglComponent will be created,
     * containing the contents of this Popup. This SwoglComponent
     * will be added to the SwoglContainer
     */
    @Override
    public void show()
    {
        //System.out.println("Showing "+this);
        //new Exception().printStackTrace();
        //System.out.println("owner    "+owner);
        //System.out.println("contents "+contents);

        // Compute the location of the popup inside the owner.
        // (Note that the coordinates (x,y) which are passed
        // in the constructor are SCREEN coordinates). This
        // has to be computed here, because otherwise the
        // location may be computed wrong when a popup is 
        // closed due to a click that opens another popup
        // and might cause the ScrollPane of the SwoglContainer
        // to scroll. 
        final Point locationInOwner = new Point(x,y);
        SwingUtilities.convertPointFromScreen(locationInOwner, owner);
        
        SwingUtilities.invokeLater(new Runnable()
        {
            @Override
            public void run()
            {
                // Create the SwoglComponent that will display 
                // the popup contents
                Dimension contentSize = contents.getPreferredSize();
                final SwoglComponent swoglPopupComponentDelegate = 
                    SwoglComponents.create(contents);

                contents.setBounds(
                    new Rectangle(0,0,contentSize.width,contentSize.height));

                // Find the component that contains the owner of the popup
                SwoglComponent parentSwoglComponent = 
                    swoglContainer.findSwoglComponentContaining(owner);
                
                // Compute the bounds of the popup owner inside the SwoglComponent
                JComponent parentComponent = 
                    parentSwoglComponent.getComponent();
                Rectangle ownerBounds = 
                    SwingUtilities.convertRectangle(
                        owner.getParent(), owner.getBounds(), parentComponent);

                final SwoglComponent parent = 
                    swoglContainer.findSwoglComponentContaining(owner);
                Box boundingBox = 
                    GeometryUtils.computeBoundingBox(parent.getGeometry());
                
                // Compute the translation of the popup-SwoglComponent
                float dx = 
                    - boundingBox.getSizeX() / 2.0f +    // Upper left corner of owner       
                    ownerBounds.x + locationInOwner.x +  // Location referring to owner          
                    contentSize.width / 2.0f;            // Center of content

                float dy = 
                    - boundingBox.getSizeY() / 2.0f +    // Upper left corner of owner
                    ownerBounds.y + locationInOwner.y +  // Location referring to owner
                    contentSize.height / 2.0f;           // Center of content

                final Matrix4f positioningTransform = 
                    MatrixUtils.translate(dx, -dy, boundingBox.getMaxZ()+1);
                
                /*/
                System.out.println("owner "+owner);
                System.out.println("ownerBounds "+ownerBounds);
                System.out.println("contents "+contents);
                System.out.println("contentSize "+contentSize);
                System.out.println("x "+x);
                System.out.println("y "+y);
                System.out.println("locationInOwner "+locationInOwner);
                System.out.println("dx "+dx);
                System.out.println("dy "+dy);
                //*/
                
                boolean useDelegate = true;
                if (!useDelegate)
                {
                    swoglPopupComponent = swoglPopupComponentDelegate;
                    Matrix4f transform = parent.getTransform();
                    transform.mul(positioningTransform);
                    swoglPopupComponent.setTransform(transform);
                }
                else
                {
                    swoglPopupComponent = createSwoglPopupComponent(
                        parent, swoglPopupComponentDelegate, 
                        positioningTransform);
                }

                //System.out.println("Adding swoglPopupComponent "+swoglPopupComponent);
                swoglContainer.addPopup(swoglPopupComponent);
            }
        });
    }
    
    /**
     * Create a SwoglComponent that delegates all calls to the given
     * delegate, except for the getTransform call, which obtains the
     * transform of the parent, multiplies it with the given 
     * positioningTransform, and returns the result.
     * 
     * @param parent The parent
     * @param delegate The delegate
     * @param positioningTransform The transform
     * @return The SwoglComponent
     */
    private static SwoglComponent createSwoglPopupComponent(
        final SwoglComponent parent,
        final SwoglComponent delegate, 
        final Matrix4f positioningTransform)
    {
        return new SwoglComponent()
        {
            @Override
            public void setTransform(Matrix4f matrix)
            {
            }
            
            @Override
            public Matrix4f getTransform()
            {
                Matrix4f parentTransform = parent.getTransform();
                parentTransform.mul(positioningTransform);
                return parentTransform;
            }
            
            @Override
            public Texture getTexture()
            {
                return delegate.getTexture();
            }
            
            @Override
            public Material getMaterial()
            {
                return delegate.getMaterial();
            }
            
            @Override
            public Geometry getGeometry()
            {
                return delegate.getGeometry();
            }
            
            @Override
            public JComponent getComponent()
            {
                return delegate.getComponent();
            }
        };
        
    }
    
    
    /**
     * Will hide this SwoglPopup, by removing the corresponding 
     * SwoglComponent from the SwoglContainer
     */
    @Override
    public void hide()
    {
        //System.out.println("Hiding");
        //System.out.println("owner    "+owner);
        //System.out.println("contents "+contents);

        SwingUtilities.invokeLater(new Runnable()
        {
            @Override
            public void run()
            {
                //System.out.println("Removing swoglPopupComponent "+swoglPopupComponent);
                swoglContainer.remove(swoglPopupComponent);
            }
        });
    }
}
