/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.geometry;

import java.util.ArrayList;
import java.util.List;

import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Tuple2f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * Abstract base implementation of a {@link Geometry}
 */
public class AbstractGeometry implements Geometry
{
    /**
     * The indices of this Geometry
     */
    private final List<Integer> vertexIndices;

    /**
     * The vertices of this Geometry
     */
    private final List<Point3f> vertices;
    
    /**
     * The normals of this Geometry
     */
    private final List<Vector3f> normals;
    
    /**
     * The texture coordinates of this Geometry
     */
    private final List<TexCoord2f> texCoords;
    
    /**
     * The flag indicating whether the vertices
     * have been modified
     */
    private boolean verticesModified;

    /**
     * The flag indicating whether the texture coordinates
     * have been modified
     */
    private boolean texCoordsModified;

    /**
     * Creates a new AbytractGeometry
     */
    protected AbstractGeometry()
    {
        vertexIndices = new ArrayList<Integer>();
        vertices = new ArrayList<Point3f>();
        normals = new ArrayList<Vector3f>();
        texCoords = new ArrayList<TexCoord2f>();
    }
    
    /**
     * Add a copy of the given vertex to this geometry.
     * 
     * @param vertex The vertex
     */
    protected void addVertex(Tuple3f vertex)
    {
        vertices.add(new Point3f(vertex));
    }
    
    /**
     * Add copy of the given texture coordinate to this geometry
     *
     * @param texCoord The texture coordinate
     */
    protected void addTexCoord(TexCoord2f texCoord)
    {
        texCoords.add(new TexCoord2f(texCoord));
    }

    /**
     * Adds copy of the given normal to this geometry
     *
     * @param normal The normal
     */
    protected void addNormal(Tuple3f normal)
    {
        normals.add(new Vector3f(normal));
    }
    
    /**
     * Add the triangle consisting of the given indices to
     * this geometry. 
     * 
     * @param i0 The first index
     * @param i1 The second index
     * @param i2 The third index
     */
    protected void addTriangle(int i0, int i1, int i2)
    {
        vertexIndices.add(i0);
        vertexIndices.add(i1);
        vertexIndices.add(i2);
    }

    @Override
    public void setVerticesModified(boolean modified)
    {
        this.verticesModified = modified;
    }

    @Override
    public boolean areVerticesModified()
    {
        return verticesModified;
    }

    @Override
    public void setTexCoordsModified(boolean modified)
    {
        this.texCoordsModified = modified;
    }

    @Override
    public boolean areTexCoordsModified()
    {
        return texCoordsModified;
    }

    @Override
    public int getNumTriangles()
    {
        return vertexIndices.size() / 3;
    }

    @Override
    public int getVertexIndex(int index)
    {
        return vertexIndices.get(index);
    }

    @Override
    public int getNumVertices()
    {
        return vertices.size();
    }

    @Override
    public void getVertex(int index, Tuple3f vertex)
    {
        vertex.set(vertices.get(index));
    }

    @Override
    public void setVertex(int index, Tuple3f vertex)
    {
        vertices.get(index).set(vertex);
    }

    @Override
    public void getNormal(int index, Tuple3f normal)
    {
        normal.set(normals.get(index));
    }

    @Override
    public void setNormal(int index, Tuple3f normal)
    {
        normals.get(index).set(normal);
    }

    @Override
    public void getTexCoord(int index, Tuple2f texCoord)
    {
        texCoord.set(texCoords.get(index));
    }

    @Override
    public void setTexCoord(int index, Tuple2f texCoord)
    {
        texCoords.get(index).set(texCoord);
    }

}