/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.geometry;

import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;

/**
 * A simple implementation of a {@link Box}
 */
public final class DefaultBox implements Box
{
    /**
     * The minimum x value of this box
     */
    private float minX = Float.POSITIVE_INFINITY;

    /**
     * The minimum y value of this box
     */
    private float minY = Float.POSITIVE_INFINITY;

    /**
     * The minimum z value of this box
     */
    private float minZ = Float.POSITIVE_INFINITY;

    /**
     * The maximum x value of this box
     */
    private float maxX = Float.NEGATIVE_INFINITY;

    /**
     * The maximum y value of this box
     */
    private float maxY = Float.NEGATIVE_INFINITY;

    /**
     * The maximum z value of this box
     */
    private float maxZ = Float.NEGATIVE_INFINITY;

    /**
     * Creates a new box
     */
    public DefaultBox()
    {
    }

    /**
     * Creates a new box with the given size
     * 
     * @param minX The minimum x value of this box
     * @param minY The minimum y value of this box
     * @param minZ The minimum z value of this box
     * @param maxX The maximum x value of this box
     * @param maxY The maximum y value of this box
     * @param maxZ The maximum z value of this box
     */
    public DefaultBox(
        float minX, float minY, float minZ, 
        float maxX, float maxY, float maxZ)
    {
        this.minX = minX;
        this.minY = minY;
        this.minZ = minZ;

        this.maxX = maxX;
        this.maxY = maxY;
        this.maxZ = maxZ;
    }
    
    /**
     * Creates a new box as a copy of the given one
     * 
     * @param other The box that serves as input for this
     */
    public DefaultBox(Box other)
    {
        combine(other);
    }

    @Override
    public float getMinX()
    {
        return minX;
    }

    /**
     * Set the minimum x value of this box
     * 
     * @param minX The minimum x value of this box
     */
    public void setMinX(float minX)
    {
        this.minX = minX;
    }

    @Override
    public float getMinY()
    {
        return minY;
    }

    /**
     * Set the minimum y value of this box
     * 
     * @param minY The minimum y value of this box
     */
    public void setMinY(float minY)
    {
        this.minY = minY;
    }


    @Override
    public float getMinZ()
    {
        return minZ;
    }

    /**
     * Set the minimum z value of this box
     * 
     * @param minZ The minimum z value of this box
     */
    public void setMinZ(float minZ)
    {
        this.minZ = minZ;
    }


    @Override
    public float getMaxX()
    {
        return maxX;
    }

    /**
     * Set the maximum x value of this box
     * 
     * @param maxX The maximum x value of this box
     */
    public void setMaxX(float maxX)
    {
        this.maxX = maxX;
    }


    @Override
    public float getMaxY()
    {
        return maxY;
    }

    /**
     * Set the maximum y value of this box
     * 
     * @param maxY The maximum y value of this box
     */
    public void setMaxY(float maxY)
    {
        this.maxY = maxY;
    }


    @Override
    public float getMaxZ()
    {
        return maxZ;
    }

    /**
     * Set the maximum z value of this box
     * 
     * @param maxZ The maximum z value of this box
     */
    public void setMaxZ(float maxZ)
    {
        this.maxZ = maxZ;
    }


    @Override
    public void getMin(Tuple3f min)
    {
        min.x = minX;
        min.y = minY;
        min.z = minZ;
    }

    /**
     * Set the minimum point of this BoundinBox
     * 
     * @param min The minimum point of this BoundinBox
     */
    public void setMin(Point3f min)
    {
        minX = min.x;
        minY = min.y;
        minZ = min.z;
    }

    @Override
    public void getMax(Tuple3f max)
    {
        max.x = maxX;
        max.y = maxY;
        max.z = maxZ;
    }

    /**
     * Set the maximum point of this BoundinBox
     * 
     * @param max The maximum point of this BoundinBox
     */
    public void setMax(Point3f max)
    {
        maxX = max.x;
        maxY = max.y;
        maxZ = max.z;
    }


    @Override
    public float getSizeX()
    {
        return maxX-minX;
    }

    @Override
    public float getSizeY()
    {
        return maxY-minY;
    }
    
    @Override
    public float getSizeZ()
    {
        return maxZ-minZ;
    }

    @Override
    public void getCenter(Tuple3f center)
    {
        center.x = (minX + maxX) * 0.5f;
        center.y = (minY + maxY) * 0.5f;
        center.z = (minZ + maxZ) * 0.5f;
    }

    /**
     * Combines this box with the given point
     * 
     * @param point The point this box should be combined with
     */
    public void combine(Point3f point)
    {
        combine(point.x, point.y, point.z);
    }

    /**
     * Combines this box with another box
     * 
     * @param other The box this box should be combined with
     */
    public void combine(Box other)
    {
        combine(other.getMinX(), other.getMinY(), other.getMinZ());
        combine(other.getMaxX(), other.getMaxY(), other.getMaxZ());
    }
    
    /**
     * Combines this box with the given point
     * 
     * @param x The x coordinate of the point
     * @param y The y coordinate of the point
     * @param z The z coordinate of the point
     */
    public void combine(float x, float y, float z)
    {
        minX = Math.min(x, minX);
        minY = Math.min(y, minY);
        minZ = Math.min(z, minZ);

        maxX = Math.max(x, maxX);
        maxY = Math.max(y, maxY);
        maxZ = Math.max(z, maxZ);
    }

    @Override
    public String toString()
    {
        return "Box[" +
        	"(" + minX + "," + minY + "," + minZ + ")-" +
        	"(" + maxX + "," + maxY + "," + maxZ + ")]";
    }

}
