/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.geometry;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.Arrays;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point2f;
import javax.vecmath.Point3f;
import javax.vecmath.Tuple2f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

/**
 * A class offering utility methods for {@link Geometry Geometries}
 */
public class GeometryUtils
{
    /**
     * A buffer for triangle normals, used in {@link #updateNormals(Geometry)}
     */
    private static Vector3f triangleNormals[] = new Vector3f[0];
    
    /**
     * Writes the indices of the given {@link Geometry} into the given 
     * buffer.<br />
     * <br />
     * After this method returns, the position of the buffer will 
     * be the same as before this method.<br />
     * <br />
     * The buffer must be large enough to hold the data. That means, 
     * for a geometry with <code>n</code> triangles, it must at least
     * have a size of <code>n*3</code>.
     * 
     * @param geometry The {@link Geometry}
     * @param buffer The target buffer
     */
    public static void writeIndicesToBuffer(
        Geometry geometry, IntBuffer buffer)
    {
        for (int i=0; i<geometry.getNumTriangles(); i++)
        {
            for (int v=0; v<3; v++)
            {
                buffer.put(i*3+v, geometry.getVertexIndex(i*3+v));
            }
        }
    }
    
    /**
     * Writes the vertices of the given {@link Geometry} into the given 
     * buffer.<br />
     * <br />
     * After this method returns, the position of the buffer will 
     * be the same as before this method.<br />
     * <br />
     * The buffer must be large enough to hold the data. That means, 
     * for a geometry with <code>n</code> vertices, it must at least
     * have a size of <code>n*3</code>.
     * 
     * @param geometry The geometry
     * @param buffer The target buffer
     */
    public static void writeVerticesToBuffer(
        Geometry geometry, FloatBuffer buffer)
    {
        Tuple3f tuple = new Point3f();
        for (int i=0; i<geometry.getNumTriangles(); i++)
        {
            for (int v=0; v<3; v++)
            {
                int index = geometry.getVertexIndex(i*3+v);
                geometry.getVertex(index, tuple);
                buffer.put(index*3+0, tuple.x);
                buffer.put(index*3+1, tuple.y);
                buffer.put(index*3+2, tuple.z);
            }
        }
    }
    
    /**
     * Writes the normals of the given {@link Geometry} into the given 
     * buffer.<br />
     * <br />
     * After this method returns, the position of the buffer will 
     * be the same as before this method.<br />
     * <br />
     * The buffer must be large enough to hold the data. That means, 
     * for a geometry with <code>n</code> vertices, it must at least
     * have a size of <code>n*3</code>.
     * 
     * @param geometry The {@link Geometry}
     * @param buffer The target buffer
     */
    public static void writeNormalsToBuffer(
        Geometry geometry, FloatBuffer buffer)
    {
        Tuple3f tuple = new Point3f();
        for (int i=0; i<geometry.getNumTriangles(); i++)
        {
            for (int v=0; v<3; v++)
            {
                int index = geometry.getVertexIndex(i*3+v);
                geometry.getNormal(index, tuple);
                buffer.put(index*3+0, tuple.x);
                buffer.put(index*3+1, tuple.y);
                buffer.put(index*3+2, tuple.z);
            }
        }
    }
    
    /**
     * Writes the texture coordinates of the given {@link Geometry} into 
     * the given buffer.<br />
     * <br />
     * After this method returns, the position of the buffer will 
     * be the same as before this method.<br />
     * <br />
     * The buffer must be large enough to hold the data. That means, 
     * for a geometry with <code>n</code> vertices, it must at least
     * have a size of <code>n*2</code>.
     * 
     * @param geometry The {@link Geometry}
     * @param buffer The target buffer
     */
    public static void writeTexCoordsToBuffer(Geometry geometry, FloatBuffer buffer)
    {
        Tuple2f tuple = new Point2f();
        for (int i=0; i<geometry.getNumTriangles(); i++)
        {
            for (int v=0; v<3; v++)
            {
                int index = geometry.getVertexIndex(i*3+v);
                geometry.getTexCoord(index, tuple);
                buffer.put(index*2+0, tuple.x);
                buffer.put(index*2+1, tuple.y);
            }
        }
    }
    
    /**
     * Obtain the position, normal and texture coordinates of the specified
     * vertex from the given {@link Geometry}. 
     * 
     * @param geometry The {@link Geometry}
     * @param triangleIndex The triangle index
     * @param vertexNumber The number of the vertex inside the triangle
     * @param vertex Will store the vertex position
     * @param normal Will store the vertex normal
     * @param texCoord Will store the vertex texture coordinates
     */
    public static void getVertex(Geometry geometry, 
        int triangleIndex, int vertexNumber, 
        Tuple3f vertex, Tuple3f normal, Tuple2f texCoord)
    {
        int index = geometry.getVertexIndex(triangleIndex*3+vertexNumber);
        geometry.getVertex(index, vertex);
        geometry.getNormal(index, normal);
        geometry.getTexCoord(index, texCoord);
    }
    
    
    /**
     * Write the positions of the vertices of the specified triangle
     * of the given {@link Geometry} into the given tuples
     * 
     * @param geometry The {@link Geometry}
     * @param triangleIndex The triangle index
     * @param v0 The first vertex 
     * @param v1 The second vertex
     * @param v2 The third vertex
     */
    public static void getVertices(Geometry geometry, int triangleIndex, 
        Tuple3f v0, Tuple3f v1, Tuple3f v2)
    {
        int offset = triangleIndex*3;
        int i0 = geometry.getVertexIndex(offset);
        int i1 = geometry.getVertexIndex(offset+1);
        int i2 = geometry.getVertexIndex(offset+2);
        geometry.getVertex(i0, v0);
        geometry.getVertex(i1, v1);
        geometry.getVertex(i2, v2);
    }
    
    /**
     * Write the texture coordinates of the vertices of the specified triangle
     * of the given {@link Geometry} into the given tuples
     * 
     * @param geometry The {@link Geometry}
     * @param triangleIndex The triangle index
     * @param t0 The first texture coordinate 
     * @param t1 The second texture coordinate
     * @param t2 The third texture coordinate
     */
    public static void getTexCoords(Geometry geometry, int triangleIndex, 
        Tuple2f t0, Tuple2f t1, Tuple2f t2)
    {
        int offset = triangleIndex*3;
        int i0 = geometry.getVertexIndex(offset);
        int i1 = geometry.getVertexIndex(offset+1);
        int i2 = geometry.getVertexIndex(offset+2);
        geometry.getTexCoord(i0, t0);
        geometry.getTexCoord(i1, t1);
        geometry.getTexCoord(i2, t2);
    }
    
    /**
     * Write the normals of the vertices of the specified triangle
     * of the given {@link Geometry} into the given tuples
     * 
     * @param geometry The {@link Geometry}
     * @param triangleIndex The triangle index
     * @param n0 The first normal 
     * @param n1 The second normal
     * @param n2 The third normal
     */
    public static void getNormals(Geometry geometry, int triangleIndex, 
        Tuple3f n0, Tuple3f n1, Tuple3f n2)
    {
        int offset = triangleIndex*3;
        int i0 = geometry.getVertexIndex(offset);
        int i1 = geometry.getVertexIndex(offset+1);
        int i2 = geometry.getVertexIndex(offset+2);
        geometry.getNormal(i0, n0);
        geometry.getNormal(i1, n1);
        geometry.getNormal(i2, n2);
    }
    
    /**
     * Update the vertex normals of the given {@link Geometry} by computing 
     * the average of the normals of the triangles that are
     * adjacent to the respective vertex. This method is not thread
     * safe, and may only be called by one thread at a time.
     * 
     * @param geometry The {@link Geometry}
     */
    public static void updateNormals(Geometry geometry)
    {
        // Compute the triangle normals
        Vector3f edge0 = new Vector3f();
        Vector3f edge1 = new Vector3f();
        Point3f v0 = new Point3f();
        Point3f v1 = new Point3f();
        Point3f v2 = new Point3f();
        int numTriangles = geometry.getNumTriangles();
        if (triangleNormals.length < numTriangles)        
        {
            int oldLength = triangleNormals.length;
            triangleNormals = Arrays.copyOfRange(
                triangleNormals, 0, numTriangles);
            for (int i=oldLength; i<numTriangles; i++)
            {
                triangleNormals[i] = new Vector3f();
            }
        }
        for (int i = 0; i < numTriangles; i++)
        {
            GeometryUtils.getVertices(geometry, i, v0, v1, v2);
            edge0.sub(v1, v0);
            edge1.sub(v2, v0);
            Vector3f triangleNormal = triangleNormals[i];
            triangleNormal.cross(edge0, edge1);
        }

        // Compute the vertex normals
        int numVertices = geometry.getNumVertices();
        Vector3f vertexNormal = new Vector3f();
        for (int i = 0; i < numVertices; i++)
        {
            geometry.setNormal(i, vertexNormal);
        }
        for (int i = 0; i < numTriangles; i++)
        {
            Vector3f triangleNormal = triangleNormals[i];
            for (int j=0; j<3; j++)
            {
                int index = geometry.getVertexIndex(i*3+j);
                geometry.getNormal(index, vertexNormal);
                vertexNormal.add(triangleNormal);
                geometry.setNormal(index, vertexNormal);
            }
        }
        for (int i = 0; i < numVertices; i++)
        {
            geometry.getNormal(i, vertexNormal);
            vertexNormal.normalize();
            geometry.setNormal(i, vertexNormal);
        }
    }
    
    
    /**
     * Computes the bounding {@link Box} of the given {@link Geometry}
     * 
     * @param geometry The {@link Geometry}
     * @return The bounding {@link Box}
     */
    public static Box computeBoundingBox(Geometry geometry)
    {
        DefaultBox boundingBox = new DefaultBox();
        Point3f vertex = new Point3f();
        for (int i=0; i<geometry.getNumVertices(); i++)
        {
            geometry.getVertex(i, vertex);
            boundingBox.combine(vertex);
        }
        return boundingBox;
    }

    /**
     * Computes the bounding {@link Box} of the given {@link Geometry} 
     * when it is transformed with the given matrix.
     * 
     * @param geometry The {@link Geometry}
     * @param matrix The matrix
     * @return The bounding {@link Box}
     */
    public static Box computeBoundingBox(
        Geometry geometry, Matrix4f matrix)
    {
        DefaultBox boundingBox = new DefaultBox();
        Point3f vertex = new Point3f();
        for (int i=0; i<geometry.getNumVertices(); i++)
        {
            geometry.getVertex(i, vertex);
            matrix.transform(vertex);
            boundingBox.combine(vertex);
        }
        return boundingBox;
    }


    /**
     * Transform the given {@link Geometry} with the given matrix
     * 
     * @param geometry The {@link Geometry}
     * @param matrix The matrix
     */
    public static void transform(Geometry geometry, Matrix4f matrix)
    {
        Point3f point = new Point3f();
        for (int i=0; i<geometry.getNumVertices(); i++)
        {
            geometry.getVertex(i, point);
            matrix.transform(point);
            geometry.setVertex(i, point);
        }
    }
    
    
    /**
     * Private constructor to prevent instantiation
     */
    private GeometryUtils()
    {
    }

    
    
    
    
    
    
    
    
}
