/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.geometry;

import javax.vecmath.Tuple2f;
import javax.vecmath.Tuple3f;

/**
 * A Geometry that consists of a rectangular grid of points.
 */
public interface SimpleGeometry extends Geometry
{
    /**
     * Returns the number of vertices of this Geometry in x-direction
     * 
     * @return The number of vertices of this Geometry in x-direction
     */
    int getNumPointsX();

    /**
     * Returns the number of vertices of this Geometry in y-direction
     * 
     * @return The number of vertices of this Geometry in y-direction
     */
    int getNumPointsY();
    
    /**
     * Store the coordinates of the specified vertex in the given tuple
     * 
     * @param x The x-index of the vertex
     * @param y The y-index of the vertex
     * @param vertex The position of the vertex
     */
    void getVertex(int x, int y, Tuple3f vertex);

    /**
     * Set the position of the specified vertex. 
     * After modifying vertices, {@link #setVerticesModified(boolean)
     * setVerticesModified(true)} should be called, so that the
     * rendering may take the modified data into account. 
     * 
     * @param x The x-index of the vertex
     * @param y The y-index of the vertex
     * @param vertex The position of the vertex
     */
    void setVertex(int x, int y, Tuple3f vertex);

    /**
     * Store the texture coordinates of the specified vertex in the given tuple
     * 
     * @param x The x-index of the vertex
     * @param y The y-index of the vertex
     * @param texCoord The texture coordinates of the vertex
     */
    void getTexCoord(int x, int y, Tuple2f texCoord);

    /**
     * Set the texture coordinates of the specified vertex. 
     * After modifying coordinates, {@link #setTexCoordsModified(boolean)
     * setTexCoordsModified(true)} should be called, so that the
     * rendering may take the modified data into account. 
     * 
     * @param x The x-index of the vertex
     * @param y The y-index of the vertex
     * @param texCoord The texture coordinates of the vertex
     */
    void setTexCoord(int x, int y, Tuple2f texCoord);
    
    /**
     * Store the coordinates of the specified vertex in the given tuple
     * 
     * @param x The x-index of the vertex
     * @param y The y-index of the vertex
     * @param normal The normal of the vertex
     */
    void getNormal(int x, int y, Tuple3f normal);

    /**
     * Set the normal of the specified vertex. 
     * After modifying vertices, {@link #setVerticesModified(boolean)
     * setVerticesModified(true)} should be called, so that the
     * rendering may take the modified data into account. 
     * 
     * @param x The x-index of the vertex
     * @param y The y-index of the vertex
     * @param normal The normal of the vertex
     */
    void setNormal(int x, int y, Tuple3f normal);
    
}
