/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.interaction;

import javax.vecmath.TexCoord2f;

import de.javagl.swogl.geometry.SimpleGeometry;

/**
 * This class transforms texture coordinates to achieve a visual effect that
 * looks like watching the texture through a magnifying glass.
 */
class MagnifyingGlass
{
    /**
     * The size of the magnified texture in pixels, in x-direction
     */
    private int sizeX = 0;

    /**
     * The size of the magnified texture in pixels, in y-direction
     */
    private int sizeY = 0;

    /**
     * The x-position of the magnifying glass on the texture
     */
    private int focusX = 0;

    /**
     * The y-position of the magnifying glass on the texture
     */
    private int focusY = 0;
    
    /**
     * The radius of the magnifying glass
     */
    private float radius = 100;
    
    /**
     * The magnification factor, between 0 and 1
     */
    private float magnification = 0.8f;
    
    /**
     * The relative distance from the center which describes
     * where the "lens effect" distortion of the magnifying
     * glass starts
     */
    private float threshold = 0.6f;
    
    /**
     * The exponent determining the shape of the magnification
     * function, i.e. the distortion at the border of the
     * magnifying glass
     */
    private float falloffExponent = 3.0f;
    
    /**
     * Creates a new MagnifyingGlass for a texture with the given size
     * 
     * @param sizeX The size of the texture, in pixels, in x-direction
     * @param sizeY The size of the texture, in pixels, in y-direction
     */
    public MagnifyingGlass(int sizeX, int sizeY)
    {
        this.sizeX = sizeX;
        this.sizeY = sizeY;
    }
    
    /**
     * Set the magnification factor of this magnifying glass. The
     * given factor will be clamped to the range [0,1]
     * 
     * @param magnification The magnification factor
     */
    public void setMagnification(float magnification)
    {
        this.magnification = magnification;
        this.magnification = Math.min(this.magnification, 1f);
        this.magnification = Math.max(this.magnification, 0f);
    }
    
    /**
     * Returns the magnification factor of this magnifying glass,
     * a value in the range [0,1]
     *  
     * @return The magnification factor of this magnifying glass
     */
    public float getMagnification()
    {
        return magnification;
    }
    
    /**
     * Set the radius of this magnifying glass
     * 
     * @param radius The radius of this magnifying glass
     */
    public void setRadius(float radius)
    {
        this.radius = radius;
    }
    
    /**
     * Return the radius of this magnifying glass
     * 
     * @return The radius of this magnifying glass
     */
    public float getRadius()
    {
        return radius;
    }
    
    /**
     * Set the threshold where the "lens effect" distortion of 
     * the magnifying glass starts. The given value will be
     * relative to the radius, and thus will be clamped to 
     * the range [0,1]
     * 
     * @param threshold The distortion threshold
     */
    public void setThreshold(float threshold)
    {
        this.threshold = threshold;
        this.threshold = Math.min(this.threshold, 1f);
        this.threshold = Math.max(this.threshold, 0f);
    }
    
    /**
     * Set the focus of this magnifying glass, i.e. its position
     * on the texture.
     * 
     * @param focusX The x-position
     * @param focusY The y-position
     */
    public void setFocus(int focusX, int focusY)
    {
        this.focusX = focusX;
        this.focusY = focusY;
    }
    
    
    /**
     * Transform the texture coordinates of the given {@link SimpleGeometry} 
     * so that they range from [0,0] to [1,1], but are distorted according
     * to the current parameters of this magnifying glass.
     * 
     * @param simpleGeometry The {@link SimpleGeometry} whose texture 
     * coordinates should be transformed.
     */
    void transformTexCoords(SimpleGeometry simpleGeometry)
    {
        TexCoord2f texCoord = new TexCoord2f();
        float normalizationX = 1.0f / (simpleGeometry.getNumPointsX() - 1);
        float normalizationY = 1.0f / (simpleGeometry.getNumPointsY() - 1);
        for (int x = 0; x < simpleGeometry.getNumPointsX(); x++)
        {
            for (int y = 0; y < simpleGeometry.getNumPointsY(); y++)
            {
                texCoord.x = x * normalizationX;
                texCoord.y = 1 - y * normalizationY;
                transformTexCoord(texCoord);
                simpleGeometry.setTexCoord(x, y, texCoord);
            }
        }
    }
    
    /**
     * This will transform the texture coordinate according to the current
     * parameters of this magnifying glass, to achieve a lens effect at its
     * current position.
     * 
     * @param texCoord The texture coordinates
     */
    private void transformTexCoord(TexCoord2f texCoord)
    {
        // Compute the absolute position of this texture coordinate
        float currentX = texCoord.x * sizeX;
        float currentY = texCoord.y * sizeY;

        // Compute the distance of this texture point from the
        // focus position of this magnifying glass
        float distX = (currentX - focusX);
        float distY = (currentY - focusY);
        float dist = (float)Math.hypot(distX, distY);

        // If the texture coordinate is under the magnifying
        // glass, distort the texture: The texture will be
        // "pulled outwards" from the focus position to
        // the border of the magnifying glass
        if (dist < radius)
        {

            // Compute the normalized distance of the point,
            // i.e. its distance relative to the size of
            // the texture
            float normDistX = distX / sizeX;
            float normDistY = distY / sizeY;
            float alpha = dist / radius;

            // Compute the distortion at the border of the
            // magnifying glass
            float distortionFactor = magnification;
            if (alpha > threshold)
            {
                float t = (alpha - threshold) / (1.0f - threshold);
                distortionFactor *= (1 - (float)Math.pow(t, falloffExponent));
            }

            // Distort the texture towards the border of
            // the magnifying glass
            texCoord.x -= distortionFactor * normDistX;
            texCoord.y -= distortionFactor * normDistY;
        }
    }
}