/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.interaction;

import java.awt.Point;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.SingularMatrixException;
import javax.vecmath.Vector3f;

import de.javagl.swogl.IntersectionInfo;
import de.javagl.swogl.MatrixUtils;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglContainer;
import de.javagl.swogl.geometry.Ray;


/**
 * This class allows moving the {@link SwoglComponent SwoglComponents} 
 * of a {@link SwoglContainer} in space. It may be used for the cases
 * where the SwoglContainer has no layout manager. <br />
 * <br />
 * Usually this class will be controlled by a MouseListener, 
 * MouseMotionListener and MouseWheelListener by calling the 
 * main public methods accordingly:
 * <ul>
 *   <li> 
 *     {@link #startMovement(Point)}
 *      when the mouse is pressed to start a movement
 *   </li>
 *   <li> 
 *      {@link #doMovement(Point, int)}
 *      when the mouse is dragged or the mouse wheel is rotated
 *      to move a SwoglComponent
 *   </li>
 * </ul>
 * However, different applications of these methods may be implemented
 * by the client.
 */
public class MovingBehavior
{
    /**
     * The currently moved SwoglComponent
     */
    private SwoglComponent currentSwoglComponent;
    
    /**
     * The current mouse position in world coordinates
     */
    private final Point3f currentWorldPosition = new Point3f();
    
    /**
     * The SwoglContainer this behavior operates on
     */
    private SwoglContainer swoglContainer;
    
    /**
     * Creates a new MovingBehavior for the given SwoglContainer.
     * 
     * @param swoglContainer The SwoglContainer this behavior operates on
     * @throws NullPointerException if the given SwoglContainer
     * is <code>null</code>
     */
    public MovingBehavior(SwoglContainer swoglContainer)
    {
        if (swoglContainer == null)
        {
            throw new NullPointerException(
                "The SwoglContainer may not be null");
        }
        this.swoglContainer = swoglContainer;
    }
    
    /**
     * Start the movement of a SwoglComponent. If a SwoglComponent is
     * found inside the SwoglContainer at the given point, then 
     * subsequent calls to {@link #doMovement(Point, int)} will move
     * the SwoglComponent.
     *  
     * @param point The point where the movement should start
     */
    public void startMovement(Point point)
    {
        IntersectionInfo intersectionInfo = 
            swoglContainer.computeIntersectionInfo(point);
        if (intersectionInfo == null)
        {
            currentSwoglComponent = null;            
        }
        else
        {
            currentSwoglComponent = intersectionInfo.getSwoglComponent();
            Matrix4f transform = currentSwoglComponent.getTransform();
            currentWorldPosition.set(
                intersectionInfo.computeIntersectionPosition());
            transform.transform(currentWorldPosition);
        }
    }
    
    /**
     * Move the current SwoglComponent. If no SwoglComponent was hit by
     * the previous call to {@link #startMovement(Point)}, then this 
     * method does not have any effect. Otherwise, this method will
     * move the SwoglComponent so that the position where it was picked
     * will be at the position that is passed in as the argument, or
     * along the viewing Z-axis by the given amount.
     * 
     * @param point The point where the SwoglComponent should move to
     * @param deltaZ The movement delta along the picking ray
     */
    public void doMovement(Point point, int deltaZ)
    {
        if (currentSwoglComponent != null)
        {
            // Compute the picking ray origin and direction
            Ray ray = swoglContainer.computePickingRay(point);
            Point3f rayOrigin = ray.getOrigin();
            Vector3f rayDirection = ray.getDirection();

            // Compute the position on the picking ray that has   
            // the same distance as the current world position
            float distance =
                -(rayOrigin.z - currentWorldPosition.z) / rayDirection.z;
            Point3f newDraggedPoint = new Point3f();
            newDraggedPoint.scaleAdd(
                distance + deltaZ, rayDirection, rayOrigin);

            // Compute the movement delta and the new world position
            Vector3f delta = new Vector3f();
            delta.sub(newDraggedPoint, currentWorldPosition);
            Matrix4f transform = currentSwoglComponent.getTransform();
            Matrix4f inverseTransform = new Matrix4f();
            try
            {
                inverseTransform.invert(transform);
            }
            catch (SingularMatrixException e)
            {
                return;
            }
            inverseTransform.transform(delta);
            currentWorldPosition.set(newDraggedPoint);

            // Apply the transformation to the SwoglComponent
            Matrix4f newTransform = 
                MatrixUtils.mul(transform, MatrixUtils.translate(delta));
            currentSwoglComponent.setTransform(newTransform);
            
            swoglContainer.repaint();
        }
    }

}
