/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.interaction;

import java.awt.Point;

import javax.vecmath.Matrix4f;
import javax.vecmath.Vector4f;

import de.javagl.swogl.IntersectionInfo;
import de.javagl.swogl.MatrixUtils;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglContainer;

/**
 * This class allows rotating the {@link SwoglComponent SwoglComponents} 
 * of a {@link SwoglContainer} in space. It may be used for the cases
 * where the SwoglContainer has no layout manager. <br />
 * <br />
 * Usually this class will be controlled by a MouseListener, 
 * and MouseWheelListener by calling the main public methods 
 * accordingly:
 * <ul>
 *   <li> 
 *     {@link #startRotation(Point)}
 *      when the mouse is pressed to start a rotation
 *   </li>
 *   <li> 
 *      {@link #doRotation(Point)}
 *      when the mouse is dragged to rotate a SwoglComponent
 *   </li>
 * </ul>
 * However, different applications of these methods may be implemented
 * by the client.
 */
public class RotatingBehavior
{
    /**
     * The rotation speed factor
     */
    private float rotationSpeed = 0.5f;
    
    /**
     * The currently rotated SwoglComponent
     */
    private SwoglComponent currentSwoglComponent;

    /**
     * The current mouse position 
     */
    private final Point currentPosition = new Point();
    
    /**
     * The SwoglContainer this behavior operates on
     */
    private final SwoglContainer swoglContainer;
    
    /**
     * Creates a new RotatingBehavior for the given {@link SwoglContainer}.
     * The {@link #setRotationSpeed(float) rotation speed} will be 0.5.
     * 
     * @param swoglContainer The SwoglContainer this behavior operates on
     * @throws NullPointerException if the given SwoglContainer
     * is <code>null</code>
     */
    public RotatingBehavior(SwoglContainer swoglContainer)
    {
        if (swoglContainer == null)
        {
            throw new NullPointerException(
                "The SwoglContainer may not be null");
        }
        this.swoglContainer = swoglContainer;
    }
    
    /**
     * Start the rotation of a {@link SwoglComponent}. If a SwoglComponent 
     * is found inside the SwoglContainer at the given point, then 
     * subsequent calls to {@link #doRotation(Point)} will rotate
     * the SwoglComponent.
     *  
     * @param point The point where the rotation should start
     * @throws NullPointerException if the given point
     * is <code>null</code>
     */
    public void startRotation(Point point)
    {
        if (point == null)
        {
            throw new NullPointerException(
                "The point may not be null");
        }
        currentPosition.setLocation(point);
        IntersectionInfo intersectionInfo = 
            swoglContainer.computeIntersectionInfo(point);
        if (intersectionInfo == null)
        {
            currentSwoglComponent = null;
        }
        else
        {
            currentSwoglComponent = intersectionInfo.getSwoglComponent();
        }
    }
    
    /**
     * Rotate the current {@link SwoglComponent}. If no SwoglComponent was 
     * hit by the previous call to {@link #startRotation(Point)}, then this 
     * method does not have any visible effect. Otherwise, this method will
     * rotate the SwoglComponent depending on the difference between
     * the given point and the previous point.
     * 
     * @param point The current point
     * @throws NullPointerException if the given point
     * is <code>null</code>
     */
    public void doRotation(Point point)
    {
        if (point == null)
        {
            throw new NullPointerException(
                "The point may not be null");
        }
        if (currentSwoglComponent != null)
        {
            int dy = point.x - currentPosition.x;
            int dx = point.y - currentPosition.y;
            float rx = (float)Math.toRadians(dx * rotationSpeed);
            float ry = (float)Math.toRadians(dy * rotationSpeed);
            rotateSwoglComponent(currentSwoglComponent, rx, ry);
            swoglContainer.repaint();
        }
        currentPosition.setLocation(point);
    }
    
    /**
     * Set the rotation speed factor for this behavior. This is the
     * factor that is used for converting a mouse movement (in pixels)
     * into degrees for the rotation.
     * 
     * @param rotationSpeed The rotation speed factor
     */
    public void setRotationSpeed(float rotationSpeed)
    {
        this.rotationSpeed = rotationSpeed;
    }

    /**
     * Rotate the given SwoglComponent around the origin of its local
     * coordinate system, by the given angles.
     * 
     * @param swoglComponent The SwoglComponent to rotate
     * @param angleRadX The angle of rotation around the x-axis, in radians
     * @param angleRadY The angle of rotation around the y-axis, in radians
     */
    private static void rotateSwoglComponent(
        SwoglComponent swoglComponent, float angleRadX, float angleRadY)
    {
        Matrix4f trans = swoglComponent.getTransform();
        Vector4f translation = new Vector4f();
        trans.getColumn(3, translation);
        trans.setColumn(3, 0,0,0,1);
        trans.mul(MatrixUtils.rotateY(angleRadY), trans);
        trans.mul(MatrixUtils.rotateX(angleRadX), trans);
        trans.setColumn(3, translation);
        swoglComponent.setTransform(trans);
    }

}
