/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.interaction;

import java.awt.Point;

import javax.vecmath.Matrix4f;

import de.javagl.swogl.IntersectionInfo;
import de.javagl.swogl.MatrixUtils;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglContainer;

/**
 * This class allows scaling the {@link SwoglComponent SwoglComponents} 
 * of a {@link SwoglContainer} in space. It may be used for the cases
 * where the SwoglContainer has no layout manager. <br />
 * <br />
 * Usually this class will be controlled by a MouseListener 
 * and MouseMotionListener by calling the main public 
 * methods accordingly:
 * <ul>
 *   <li> 
 *     {@link #startScaling(Point)}
 *      when the mouse is pressed to start a scaling
 *   </li>
 *   <li> 
 *      {@link #doScaling(Point)}
 *      when the mouse is dragged to scale a SwoglComponent
 *   </li>
 * </ul>
 * However, different applications of these methods may be implemented
 * by the client.
 */
public class ScalingBehavior
{
    /**
     * Factor for the scaling speed 
     */
    private float scalingSpeed = 1.0f;
    
    /**
     * Whether the scaling should be uniform
     */
    private boolean uniformScaling = false;
    
    /**
     * The current SwoglComponent
     */
    private SwoglComponent currentSwoglComponent;
    
    /**
     * The current mouse position
     */
    private final Point currentPosition = new Point();
    
    /**
     * The SwoglContainer this behavior operates on
     */
    private final SwoglContainer swoglContainer;
    
    /**
     * Creates a new ScalingBehavior for the given {@link SwoglContainer}.
     * The {@link #setScalingSpeed(float) scaling speed} will be 1.0 and
     * the scaling will be {@link #setUniformScaling(boolean) non-uniform}.
     * 
     * @param swoglContainer The SwoglContainer this behavior operates on
     * @throws NullPointerException if the given SwoglContainer
     * is <code>null</code>
     */
    public ScalingBehavior(SwoglContainer swoglContainer)
    {
        if (swoglContainer == null)
        {
            throw new NullPointerException(
                "The SwoglContainer may not be null");
        }
        this.swoglContainer = swoglContainer;
    }

    /**
     * Start scaling the current SwoglComponent. If a SwoglComponent is
     * found inside the SwoglContainer at the given point, then 
     * subsequent calls to {@link #doScaling(Point)} will scale
     * the SwoglComponent.
     *  
     * @param point The point where the scaling should start
     * @throws NullPointerException if the given point
     * is <code>null</code>
     */
    public void startScaling(Point point)
    {
        if (point == null)
        {
            throw new NullPointerException(
                "The point may not be null");
        }
        currentPosition.setLocation(point);
        IntersectionInfo intersectionInfo = 
            swoglContainer.computeIntersectionInfo(point);
        if (intersectionInfo == null)
        {
            currentSwoglComponent = null;
        }
        else
        {
            currentSwoglComponent = intersectionInfo.getSwoglComponent();
        }
    }
    
    /**
     * Scale the current SwoglComponent. If no SwoglComponent was hit by
     * the previous call to {@link #startScaling(Point)}, then this 
     * method does not have any effect. Otherwise, this method will
     * scale the SwoglComponent depending on the difference between
     * the given point and the previous point.
     * 
     * @param point The current point
     * @throws NullPointerException if the given point
     * is <code>null</code>
     */
    public void doScaling(Point point)
    {
        if (point == null)
        {
            throw new NullPointerException(
                "The point may not be null");
        }
        if (currentSwoglComponent != null)
        {
            int dx = point.x - currentPosition.x;
            int dy = point.y - currentPosition.y;
            float sx = 1.0f + dx * scalingSpeed * 0.01f;
            float sy = 1.0f - dy * scalingSpeed * 0.01f;
            
            if (uniformScaling)
            {
                float s = Math.max(sx, sy);
                Matrix4f m = currentSwoglComponent.getTransform();
                m.mul(MatrixUtils.scale(s, s, s));
                currentSwoglComponent.setTransform(m);
            }
            else
            {
                Matrix4f m = currentSwoglComponent.getTransform();
                m.mul(MatrixUtils.scale(sx, sy, 1));
                currentSwoglComponent.setTransform(m);
            }
            
            
            currentPosition.setLocation(point);
            swoglContainer.repaint();
        }
    }

    /**
     * Set the scaling speed factor. This factor will be used
     * to convert mouse movement (in pixels) into scaling 
     * factors (in percent)
     * 
     * @param scalingSpeed The scaling speed factor
     */
    public void setScalingSpeed(float scalingSpeed)
    {
        this.scalingSpeed = scalingSpeed;
    }

    /**
     * Set whether the scaling should be uniform. When the scaling
     * is uniform, the actual scaling will be the maximum of the
     * scalings implied by the movement in x- and y-direction.
     * 
     * @param uniformScaling Whether the scaling should be uniform
     */
    public void setUniformScaling(boolean uniformScaling)
    {
        this.uniformScaling = uniformScaling;
    }
}
