/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.layout;

import javax.swing.SwingUtilities;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import de.javagl.swogl.MatrixUtils;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglContainer;
import de.javagl.swogl.geometry.Box;
import de.javagl.swogl.geometry.Geometry;
import de.javagl.swogl.geometry.GeometryUtils;
import de.javagl.swogl.scene.Camera;
import de.javagl.swogl.scene.CameraListener;

/**
 * This is an implementation of the {@link LayoutManager3D} interface
 * that lays out all SwoglComponents so that they face the viewer.
 * The computation is based on a {@link Camera} that provides the
 * viewing information. This class also offers a static utility
 * method for computing the billboarding transform, which may 
 * be used in other implementations of layout managers.
 */
public class BillboardLayout3D implements LayoutManager3D, CameraListener
{
    /**
     * The Camera which the components will face to
     */
    private final Camera camera;
    
    /**
     * The SwoglContainer this layout operates on 
     */
    private SwoglContainer swoglContainer;
    
    
    /**
     * Creates a new BillboardLayout3D that lets all SwoglComponents
     * face the given camera. This layout will be attached as a 
     * CameraListener to the given Camera.
     *  
     * @param camera The camera
     */
    public BillboardLayout3D(Camera camera)
    {
        this.camera = camera;
        camera.addCameraListener(this);
    }
    
    @Override
    public void cameraChanged(Camera camera)
    {
        updateLayout();
    }

    @Override
    public void setSwoglContainer(SwoglContainer swoglContainer)
    {
        this.swoglContainer = swoglContainer;
    }
    
    private void updateLayout()
    {
        SwingUtilities.invokeLater(new Runnable()
        {
            @Override
            public void run()
            {
                swoglContainer.validate3D();
            }
        });
    }
    

    @Override
    public void doLayout3D()
    {
        int componentCount = swoglContainer.getSwoglComponentCount();
        for (int i = 0; i < componentCount; i++)
        {
            SwoglComponent swoglComponent = 
                swoglContainer.getSwoglComponent(i);
            Matrix4f newTransform = 
                computeBillboardTransform(swoglComponent, camera);
            swoglComponent.setTransform(newTransform);
        }
        
    }
    
    
    /**
     * Utility method that computes the transform that must be applied to
     * the given {@link SwoglComponent} so that it faces the given 
     * {@link Camera}.
     * 
     * @param swoglComponent The SwoglComponent
     * @param camera The Camera
     * @return The billboarding transform
     */
    public static Matrix4f computeBillboardTransform(
        SwoglComponent swoglComponent, Camera camera)
    {
        // Compute the center of the component
        Geometry geometry = swoglComponent.getGeometry();
        Box boundingBox = GeometryUtils.computeBoundingBox(geometry);
        Point3f center = new Point3f();
        boundingBox.getCenter(center);

        // Compute the rotation matrix as described on NeHe
        Vector3f look = new Vector3f();
        look.sub(camera.getEyePoint(), center);
        look.normalize();
        Vector3f up = camera.getUpVector();
        Vector3f right = new Vector3f();
        right.cross(up, look);
        up.cross(look, right); // Looks right :-)
        Matrix4f rotation = new Matrix4f();
        rotation.setIdentity();
        rotation.m00 = right.x;
        rotation.m10 = right.y;
        rotation.m20 = right.z;
        rotation.m01 = up.x;
        rotation.m11 = up.y;
        rotation.m21 = up.z;
        rotation.m02 = look.x;
        rotation.m12 = look.y;
        rotation.m22 = look.z;
        
        // Compute the final transform of the SwoglComponent
        Matrix4f currentTransform = swoglComponent.getTransform();
        currentTransform.transform(center);
        
        Matrix4f newTransform = MatrixUtils.mul(
            MatrixUtils.translate(center),
            rotation
        );
        return newTransform;
    }

    
}
