/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.Raster;
import java.nio.IntBuffer;

/**
 * Utility methods for images.
 */
class ImageUtils
{
    
    /**
     * Returns the data of the given image as an int[] array.
     * The image must have the type BufferedImage.TYPE_INT_ARGB,
     * BufferedImage.TYPE_INT_BGR or BufferedImage.TYPE_INT_RGB. 
     * For other types, an IllegalArgumentException will be thrown.
     * 
     * @param image The image
     * @return The data of the image.
     * @throws IllegalArgumentException If the image does not have
     * a DataBufferInt.
     */
    private static int[] getDataArray(BufferedImage image)
    {
        Raster raster = image.getRaster();
        DataBuffer dataBuffer = raster.getDataBuffer();
        if (!(dataBuffer instanceof DataBufferInt))
        {
            throw new IllegalArgumentException(
                "Image does not have a DataBufferInt");
        }
        DataBufferInt dataBufferInt = (DataBufferInt)dataBuffer;
        int data[] = dataBufferInt.getData();
        return data;
    }
    
    /**
     * Write the data from the given BufferedImage into the given IntBuffer.
     * The position of the given buffer will be advanced accordingly.
     * This method assumes that the BufferedImage contains an DataBufferInt.
     * So the image must have the type BufferedImage.TYPE_INT_ARGB,
     * BufferedImage.TYPE_INT_BGR or BufferedImage.TYPE_INT_RGB. For other
     * types, an IllegalArgumentException will be thrown.
     *  
     * @param image The image
     * @param buffer The target buffer
     * @throws IllegalArgumentException If the image does not have
     * a DataBufferInt.
     */
    public static void writeImageToBuffer(
        BufferedImage image, IntBuffer buffer)
    {
        buffer.put(getDataArray(image));
    }
    
    /**
     * Only for internal tests
     * @param image Do not use
     * @param buffer Do not use
     */
    public static void writeImageToBufferARGBtoABGR(
        BufferedImage image, IntBuffer buffer)
    {
        writeImageToBuffer(image, 2, 1, 0, 3, buffer);
    }
    
    private static void writeImageToBuffer(
        BufferedImage image, int s0, int s1, int s2, int s3, IntBuffer buffer)
    {
        int data[] = getDataArray(image);
        int ss0 = s0 << 3;
        int ss1 = s1 << 3;
        int ss2 = s2 << 3;
        int ss3 = s3 << 3;
        for (int i=0; i<data.length; i++)
        {
            int s = data[i];
            int c0 = (s >>> ss0) & 0xFF;
            int c1 = (s >>> ss1) & 0xFF;
            int c2 = (s >>> ss2) & 0xFF;
            int c3 = (s >>> ss3) & 0xFF;
            int d = c0 | (c1 <<  8) | (c2 << 16) | (c3 << 24);
            buffer.put(d);
        }
    }
    
    
    
    /**
     * Private constructor to prevent instantiation
     */
    private ImageUtils()
    {
    }
    
}
