/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import static org.lwjgl.opengl.GL11.GL_MODELVIEW;
import static org.lwjgl.opengl.GL11.GL_MODELVIEW_MATRIX;
import static org.lwjgl.opengl.GL11.GL_PROJECTION;
import static org.lwjgl.opengl.GL11.GL_PROJECTION_MATRIX;
import static org.lwjgl.opengl.GL11.GL_VIEWPORT;
import static org.lwjgl.opengl.GL11.glGetFloat;
import static org.lwjgl.opengl.GL11.glGetInteger;
import static org.lwjgl.opengl.GL11.glMatrixMode;

import java.awt.Rectangle;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.vecmath.Matrix4f;

import de.javagl.swogl.BufferUtils;
import de.javagl.swogl.MatrixUtils;
import de.javagl.swogl.PickingRayComputer;
import de.javagl.swogl.PickingUtils;
import de.javagl.swogl.geometry.Ray;

/**
 * Implementation of a {@link PickingRayComputer} that obtains the
 * information required for picking from the viewport and modelview 
 * and projection matrix of LWJGL <br />
 * <br />
 * Note that, in order to make this PickingRayComputer work properly, 
 * the {@link #update()} method has to be called after the GL modelview 
 * matrix has been set up to include the view transformation.
 * 
 */
public class PickingRayComputerLWJGL implements PickingRayComputer
{
    /**
     * A buffer holding a 4x4 float matrix
     */
    private final FloatBuffer matrixBuffer;
    
    /**
     * A buffer holding a viewport
     */
    private final IntBuffer viewportBuffer;
    
    /**
     * The view matrix that was read during the last update
     */
    private final Matrix4f viewMatrix;
    
    /**
     * The projection matrix that was read during the last update
     */
    private final Matrix4f projectionMatrix;
    
    /**
     * The viewport that was read during the last update.
     */
    private final Rectangle viewport;
    
    /**
     * Creates a new PickingRayComputerLWJGL 
     */
    public PickingRayComputerLWJGL()
    {
        matrixBuffer = BufferUtils.createFloatBuffer(16);
        viewportBuffer = BufferUtils.createIntBuffer(16);
        viewMatrix = new Matrix4f();
        projectionMatrix = new Matrix4f();
        viewport = new Rectangle();
    }
    
    /**
     * Update this picking ray computer based on the state that is 
     * obtained from the current GL. This method should be called
     * during each rendering pass, on the rendering thread.
     */
    public void update()
    {
        glGetInteger(GL_VIEWPORT, viewportBuffer);
        viewport.x = viewportBuffer.get(0);
        viewport.y = viewportBuffer.get(1);
        viewport.width = viewportBuffer.get(2);
        viewport.height = viewportBuffer.get(3);
        
        glMatrixMode(GL_PROJECTION);
        glGetFloat(GL_PROJECTION_MATRIX, matrixBuffer);
        MatrixUtils.writeBufferToMatrix(matrixBuffer, projectionMatrix);
        matrixBuffer.rewind();

        glMatrixMode(GL_MODELVIEW);
        glGetFloat(GL_MODELVIEW_MATRIX, matrixBuffer);
        MatrixUtils.writeBufferToMatrix(matrixBuffer, viewMatrix);
        matrixBuffer.rewind();

        //System.out.println("Viewport "+viewport);
        //System.out.println("Cam \n"+cameraMatrix);
        //System.out.println("Pro \n"+projectionMatrix);
    }
    
    @Override
    public Ray computePickingRay(int x, int y)
    {
        return PickingUtils.computePickingRay(
            viewMatrix, projectionMatrix, viewport, x, y);
    }

}
