/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import java.awt.Rectangle;
import java.nio.IntBuffer;

import javax.media.opengl.GL3;
import javax.vecmath.Matrix4f;

import de.javagl.swogl.BufferUtils;
import de.javagl.swogl.PickingRayComputer;
import de.javagl.swogl.PickingUtils;
import de.javagl.swogl.geometry.Ray;

/**
 * Implementation of a {@link PickingRayComputer} that obtains the
 * information required for picking from the viewport and modelview 
 * and projection matrix of JOGL. <br />
 * <br />
 * This implementation requires the current GL object to work.
 * This GL object should be set by calling {@link #setGL(GL3)}
 * in the <code>GLEventListener#init(GLDrawable)<code> method 
 * of JOGL. 
 * <br />
 * Note that, in order to make this PickingRayComputer work properly, 
 * the {@link #update(Matrix4f, Matrix4f)} method has to be called 
 * with the current camera- and projection matrix during each
 * rendering pass.
 */
public class PickingRayComputerShaderJOGL implements PickingRayComputer
{
    /**
     * The current GL
     */
    private GL3 gl;
    
    /**
     * A buffer holding a viewport
     */
    private final IntBuffer viewportBuffer;
    
    /**
     * The camera (modelview) matrix that was read during the last update
     */
    private final Matrix4f cameraMatrix;
    
    /**
     * The projection matrix that was read during the last update
     */
    private final Matrix4f projectionMatrix;
    
    /**
     * The viewport that was read during the last update.
     */
    private final Rectangle viewport;
    
    /**
     * Creates a new PickingRayComputerShaderJOGL
     */
    public PickingRayComputerShaderJOGL()
    {
        viewportBuffer = BufferUtils.createIntBuffer(4);
        cameraMatrix = new Matrix4f();
        projectionMatrix = new Matrix4f();
        viewport = new Rectangle();
    }
    
    /**
     * Set the current GL
     *  
     * @param gl The GL
     */
    public void setGL(GL3 gl)
    {
        this.gl = gl;
    }
    
    /**
     * Update this picking ray computer with the given camera- and 
     * projection matrix. This method should be called
     * during each rendering pass, on the rendering thread.
     * 
     * @param cameraMatrix The camera matrix (i.e. the inverse
     * of the view matrix)
     * @param projectionMatrix The projection matrix
     */
    public void update(Matrix4f cameraMatrix, Matrix4f projectionMatrix)
    {
        gl.glGetIntegerv(GL3.GL_VIEWPORT, viewportBuffer);
        viewport.x = viewportBuffer.get(0);
        viewport.y = viewportBuffer.get(1);
        viewport.width = viewportBuffer.get(2);
        viewport.height = viewportBuffer.get(3);
        
        this.cameraMatrix.set(cameraMatrix);
        this.projectionMatrix.set(projectionMatrix);
    }
    
    @Override
    public Ray computePickingRay(int x, int y)
    {
        return PickingUtils.computePickingRay(
            cameraMatrix, projectionMatrix, viewport, x, y);
    }

}
