/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import java.nio.FloatBuffer;

import javax.vecmath.Matrix4f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Tuple4f;

import de.javagl.swogl.BufferUtils;
import de.javagl.swogl.MatrixUtils;

/**
 * Utility methods for rendering. <br />
 * <br />
 * Note that the methods in this class are NOT thread safe. They 
 * will always return the same buffer, filled with the given 
 * values. The buffer should thus not be stored by clients, but 
 * may only be passed to functions which only once read the 
 * contents and do not use the buffer afterwards. 
 */
class RenderingUtils
{
    /**
     * Direct buffer to temporarily store a matrix
     */
    private static final FloatBuffer tempMatrixBuffer = 
        BufferUtils.createFloatBuffer(16);
    
    /**
     * Direct buffer to temporarily store a 4-tuple
     */
    private static final FloatBuffer tempTupleBuffer3 =
        BufferUtils.createFloatBuffer(3);
    
    /**
     * Direct buffer to temporarily store a 3-tuple
     */
    private static final FloatBuffer tempTupleBuffer4 =
        BufferUtils.createFloatBuffer(4);
    
    /**
     * Direct buffer to temporarily store a float value
     */
    private static final FloatBuffer tempFloatBuffer =
        BufferUtils.createFloatBuffer(1);
    
    
    
    /**
     * Returns a direct buffer with the contents of the given matrix.
     * 
     * @param matrix The matrix
     * @return The buffer
     */
    static FloatBuffer bufferWith(Matrix4f matrix)
    {
        MatrixUtils.writeMatrixToBuffer(matrix, tempMatrixBuffer);
        tempMatrixBuffer.rewind();
        return tempMatrixBuffer;
    }

    /**
     * Returns a direct buffer with the contents of the given tuple.
     * 
     * @param tuple The tuple
     * @return The buffer
     */
    static FloatBuffer bufferWith(Tuple4f tuple)
    {
        tempTupleBuffer4.put(0, tuple.x);
        tempTupleBuffer4.put(1, tuple.y);
        tempTupleBuffer4.put(2, tuple.z);
        tempTupleBuffer4.put(3, tuple.w);
        tempTupleBuffer4.rewind();
        return tempTupleBuffer4;
    }

    /**
     * Returns a direct buffer with the given values.
     * 
     * @param x The x value
     * @param y The y value
     * @param z The z value
     * @param w The w value
     * @return The buffer
     */
    static FloatBuffer bufferWith(float x, float y, float z, float w)
    {
        tempTupleBuffer4.put(0, x);
        tempTupleBuffer4.put(1, y);
        tempTupleBuffer4.put(2, z);
        tempTupleBuffer4.put(3, w);
        tempTupleBuffer4.rewind();
        return tempTupleBuffer4;
    }
    
    /**
     * Returns a direct buffer with the contents of the given tuple.
     * 
     * @param tuple The tuple
     * @return The buffer
     */
    static FloatBuffer bufferWith(Tuple3f tuple)
    {
        tempTupleBuffer3.put(0, tuple.x);
        tempTupleBuffer3.put(1, tuple.y);
        tempTupleBuffer3.put(2, tuple.z);
        tempTupleBuffer3.rewind();
        return tempTupleBuffer3;
    }

    /**
     * Returns a direct buffer with the contents of the given tuple
     * and the given fourth value. 
     * 
     * @param tuple The tuple
     * @param w The fourth component to put into the buffer
     * @return The buffer
     */
    static FloatBuffer bufferWith(Tuple3f tuple, float w)
    {
        tempTupleBuffer4.put(0, tuple.x);
        tempTupleBuffer4.put(1, tuple.y);
        tempTupleBuffer4.put(2, tuple.z);
        tempTupleBuffer4.put(3, w);
        tempTupleBuffer4.rewind();
        return tempTupleBuffer4;
    }

    /**
     * Returns a direct buffer with the given value.
     * 
     * @param value The tuple
     * @return The buffer
     */
    static FloatBuffer bufferWith(float value)
    {
        tempFloatBuffer.put(0, value);
        tempFloatBuffer.rewind();
        return tempFloatBuffer;
    }

    /**
     * Private constructor to prevent instantiation
     */
    private RenderingUtils()
    {
    }

}
