/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;


import static javax.media.opengl.GL.GL_COLOR_BUFFER_BIT;
import static javax.media.opengl.GL.GL_DEPTH_BUFFER_BIT;
import static javax.media.opengl.GL.GL_DEPTH_TEST;
import static javax.media.opengl.GL.GL_TRUE;
import static javax.media.opengl.GL2.GL_LIGHT_MODEL_AMBIENT;
import static javax.media.opengl.GL2.GL_LIGHT_MODEL_TWO_SIDE;
import static javax.media.opengl.GL2.GL_MAX_LIGHTS;
import static javax.media.opengl.GL2.GL_AMBIENT;
import static javax.media.opengl.GL2.GL_CONSTANT_ATTENUATION;
import static javax.media.opengl.GL2.GL_DIFFUSE;
import static javax.media.opengl.GL2.GL_LIGHT0;
import static javax.media.opengl.GL2.GL_LIGHTING;
import static javax.media.opengl.GL2.GL_LINEAR_ATTENUATION;
import static javax.media.opengl.GL2.GL_POSITION;
import static javax.media.opengl.GL2.GL_QUADRATIC_ATTENUATION;
import static javax.media.opengl.GL2.GL_SMOOTH;
import static javax.media.opengl.GL2.GL_SPECULAR;
import static javax.media.opengl.GL2.GL_SPOT_CUTOFF;
import static javax.media.opengl.GL2.GL_SPOT_DIRECTION;
import static javax.media.opengl.GL2.GL_SPOT_EXPONENT;
import static javax.media.opengl.GL2.GL_MODELVIEW;
import static javax.media.opengl.GL2.GL_PROJECTION;

import java.awt.Component;
import java.awt.Rectangle;
import java.util.List;

import javax.media.opengl.GL2;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLProfile;
import javax.media.opengl.awt.GLCanvas;
import javax.media.opengl.awt.GLJPanel;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

import de.javagl.swogl.DefaultPickingRayComputer;
import de.javagl.swogl.PickingRayComputer;
import de.javagl.swogl.SwoglApplication;
import de.javagl.swogl.SwoglComponentHandler;
import de.javagl.swogl.SwoglRenderer;
import de.javagl.swogl.scene.Camera;
import de.javagl.swogl.scene.CameraListener;
import de.javagl.swogl.scene.CameraUtils;
import de.javagl.swogl.scene.DefaultCamera;
import de.javagl.swogl.scene.DefaultLightSetup;
import de.javagl.swogl.scene.DefaultView;
import de.javagl.swogl.scene.Light;
import de.javagl.swogl.scene.Light.Type;
import de.javagl.swogl.scene.LightSetup;
import de.javagl.swogl.scene.MutableCamera;
import de.javagl.swogl.scene.View;
import de.javagl.swogl.scene.ViewUtils;

/**
 * Implementation of a {@link SwoglApplication} using JOGL
 */
public class SwoglApplicationJOGL implements SwoglApplication
{
    /**
     * The current GL
     */
    private GL2 gl;

    /**
     * The {@link SwoglRenderer} used in this application
     */
    private final SwoglRendererJOGL swoglRenderer;
    
    /**
     * The {@link View} that is used in this application
     */
    private final DefaultView view;
    
    /**
     * The {@link Camera} that is used in this application
     */
    private final MutableCamera camera;
    
    /**
     * The {@link LightSetup} that is used in this application
     */
    private final LightSetup lightSetup;
    
    /**
     * The {@link PickingRayComputer} that is used in this application
     */
    private final PickingRayComputer pickingRayComputer;

    /**
     * The rendering component of this application
     */
    private Component canvas;
    
    /**
     * Whether a GLJPanel should be used instead of a GLCanvas
     * (experimental)
     */
    private final boolean useGLJPanel;    
    
    /**
     * The GLEventListener that will be attached to the rendering component
     */
    private GLEventListener glEventListener = new GLEventListener()
    {
        @Override
        public void init(GLAutoDrawable drawable)
        {
            GL2 gl = drawable.getGL().getGL2();
            SwoglApplicationJOGL.this.gl = gl;
            swoglRenderer.setGL(gl);
            gl.setSwapInterval(0);
        }
        
        @Override
        public void display(GLAutoDrawable drawable) 
        {
            render();
        }
        
        @Override
        public void reshape(
            GLAutoDrawable drawable, int x, int y, int width, int height)
        {
            setupView();
        }

        @Override
        public void dispose(GLAutoDrawable arg0)
        {
        }
    };

    
    /**
     * Creates a new SwoglApplicationJOGL
     */
    public SwoglApplicationJOGL()
    {
        this(false);
    }
    
    /**
     * Creates a new SwoglApplicationJOGL.
     * 
     * @param useGLJPanel Whether a GLJPanel should be used
     * instead of a GLCanvas (experimental)
     */
    public SwoglApplicationJOGL(boolean useGLJPanel)
    {
        this.view = new DefaultView();
        this.camera = new DefaultCamera();
        this.camera.setEyePoint(new Point3f(0,0,500));
        this.lightSetup = new DefaultLightSetup();
        this.pickingRayComputer = 
            new DefaultPickingRayComputer(getCamera(), getView());
        
        this.swoglRenderer = new SwoglRendererJOGL();
        this.useGLJPanel = useGLJPanel;
        initialize();
    }
    
    /**
     * Internal initialization
     */
    private void initialize()
    {
        GLProfile profile = GLProfile.get(GLProfile.GL2);
        GLCapabilities capabilities = new GLCapabilities(profile);
        capabilities.setNumSamples(2);
        capabilities.setSampleBuffers(true);
        
        if (useGLJPanel)
        {
            GLJPanel c = new GLJPanel(capabilities);
            c.addGLEventListener(glEventListener);
            canvas = c;
        }
        else
        {
            GLCanvas c = new GLCanvas(capabilities);
            c.addGLEventListener(glEventListener);
            canvas = c;
        }
        
        camera.addCameraListener(new CameraListener()
        {
            @Override
            public void cameraChanged(Camera camera)
            {
                canvas.repaint();
            }
        });
    }
    
    @Override
    public View getView()
    {
        return view;
    }
    
    @Override
    public MutableCamera getCamera()
    {
        return camera;
    }
    
    @Override 
    public LightSetup getLightSetup()
    {
        return lightSetup;
    }
    
    @Override 
    public PickingRayComputer getPickingRayComputer()
    {
        return pickingRayComputer;
    }
    
    @Override
    public Component getRenderComponent()
    {
        return canvas;
    }
    
    @Override
    public SwoglComponentHandler getSwoglComponentHandler()
    {
        return swoglRenderer;
    }
    
    /**
     * Returns the SwoglRenderer that is used in this application
     * 
     * @return The SwoglRenderer that is used in this application
     */
    protected SwoglRendererJOGL getSwoglRenderer()
    {
        return swoglRenderer;
    }
    
    /**
     * Set up the view according to the current size of the 
     * rendering component
     */
    protected void setupView()
    {
        int width = getRenderComponent().getWidth();
        int height = getRenderComponent().getHeight();
        float aspect = (float)width/height;
        view.setAspect(aspect);
        view.setViewport(new Rectangle(0,0,width,height));
        gl.glViewport(0, 0, width, height);
    }
    
    /**
     * Renders the current scene. This will call {@link #prepareRender()}
     * and {@link #applyView()}, and then let the SwoglRenderer render
     * all SwoglComponents.
     */
    protected void render() 
    {
        prepareRender();
        applyView();
        
        gl.glPushMatrix();
        Matrix4f viewMatrix = CameraUtils.computeViewMatrix(camera);
        gl.glMultMatrixf(RenderingUtils.bufferWith(viewMatrix));
        
        applyLightSetup();

        swoglRenderer.renderAllSwoglComponents();
        
        gl.glPopMatrix();
    }
    
    /**
     * Prepares the rendering pass
     */
    protected void prepareRender()
    {
        gl.glClearColor(0,0,0,0);
        gl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        gl.glEnable(GL_DEPTH_TEST);
        gl.glShadeModel (GL_SMOOTH);
    }

    /**
     * Apply the {@link LightSetup} of this application
     */
    protected void applyLightSetup()
    {
        List<Light> lights = lightSetup.getLights();
        if (lights.size() > 0)
        {
            gl.glEnable(GL_LIGHTING);
            gl.glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
            gl.glLightModelfv(GL_LIGHT_MODEL_AMBIENT, 
                RenderingUtils.bufferWith(lightSetup.getAmbient()));
        }
        else
        {
            gl.glDisable(GL_LIGHTING);
        }
        
        int maxLightsArray[] = { 0 };
        gl.glGetIntegerv(GL_MAX_LIGHTS, maxLightsArray, 0);
        for (int i=0; i<maxLightsArray[0]; i++)
        {
            int id = GL_LIGHT0+i;
            gl.glDisable(id);
        }
        
        for (int i=0; i<lights.size(); i++)
        {
            Light light = lights.get(i);
            int id = GL_LIGHT0+i;
            gl.glEnable(id);

            gl.glLightfv(id, GL_AMBIENT, 
                RenderingUtils.bufferWith(light.getAmbientColor()));
            gl.glLightfv(id, GL_DIFFUSE, 
                RenderingUtils.bufferWith(light.getDiffuseColor()));
            gl.glLightfv(id, GL_SPECULAR, 
                RenderingUtils.bufferWith(light.getSpecularColor()));

            gl.glLightf(id, GL_CONSTANT_ATTENUATION, 
                light.getConstantAttenuation());
            gl.glLightf(id, GL_LINEAR_ATTENUATION, 
                light.getLinearAttenuation());
            gl.glLightf(id, GL_QUADRATIC_ATTENUATION, 
                light.getQuadraticAttenuation());
            
            // Whether OpenGL interprets the GL_POSITION of a light as a 
            // position or a direction depends on the w-component of the 
            // given tuple. If the w-component of the tuple is 0, then 
            // the GL_POSITION is actually interpreted as the direction 
            // of a directional light.
            // So when the light is DIRECTIONAL, then the direction will
            // be passed in as GL_POSITION, with the w-component being 0.
            // If the light is not DIRECTIONAL, then the position will be
            // passed in as GL_POSITION, but with the w-component being 1,
            // so that it is really interpreted as the position of a 
            // point- or spot light. 
            if (light.getType() == Type.DIRECTIONAL)
            {
                gl.glLightfv(id, GL_POSITION, 
                    RenderingUtils.bufferWith(light.getDirection(), 0));
                gl.glLightf(id, GL_SPOT_CUTOFF, 180);
            }
            else if (light.getType() == Type.POINT)
            {
                gl.glLightfv(id, GL_POSITION, 
                    RenderingUtils.bufferWith(light.getPosition(), 1));
                gl.glLightf(id, GL_SPOT_CUTOFF, 180);
            }
            else if (light.getType() == Type.SPOT)
            {
                gl.glLightfv(id, GL_POSITION, 
                    RenderingUtils.bufferWith(light.getPosition(), 1));
                gl.glLightfv(id, GL_SPOT_DIRECTION, 
                    RenderingUtils.bufferWith(light.getDirection(), 0));
                gl.glLightf(id, GL_SPOT_CUTOFF, light.getSpotCutoff());
                gl.glLightf(id, GL_SPOT_EXPONENT, light.getSpotExponent());
            }
            
        }
    }
    
    
    /**
     * Apply the current view state to GL.
     */
    protected void applyView()
    {
        gl.glMatrixMode(GL_PROJECTION);
        Matrix4f projectionMatrix = 
            ViewUtils.computeProjectionMatrix(camera.getFovY(), view);
        gl.glLoadMatrixf(RenderingUtils.bufferWith(projectionMatrix));
        gl.glMatrixMode(GL_MODELVIEW);        
    }
    
}
