/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;
import static org.lwjgl.opengl.GL11.GL_AMBIENT;
import static org.lwjgl.opengl.GL11.GL_COLOR_BUFFER_BIT;
import static org.lwjgl.opengl.GL11.GL_CONSTANT_ATTENUATION;
import static org.lwjgl.opengl.GL11.GL_DEPTH_BUFFER_BIT;
import static org.lwjgl.opengl.GL11.GL_DEPTH_TEST;
import static org.lwjgl.opengl.GL11.GL_DIFFUSE;
import static org.lwjgl.opengl.GL11.GL_LIGHT0;
import static org.lwjgl.opengl.GL11.GL_LIGHTING;
import static org.lwjgl.opengl.GL11.GL_LIGHT_MODEL_TWO_SIDE;
import static org.lwjgl.opengl.GL11.GL_LINEAR_ATTENUATION;
import static org.lwjgl.opengl.GL11.GL_MAX_LIGHTS;
import static org.lwjgl.opengl.GL11.GL_MODELVIEW;
import static org.lwjgl.opengl.GL11.GL_POSITION;
import static org.lwjgl.opengl.GL11.GL_PROJECTION;
import static org.lwjgl.opengl.GL11.GL_QUADRATIC_ATTENUATION;
import static org.lwjgl.opengl.GL11.GL_SMOOTH;
import static org.lwjgl.opengl.GL11.GL_SPECULAR;
import static org.lwjgl.opengl.GL11.GL_SPOT_CUTOFF;
import static org.lwjgl.opengl.GL11.GL_SPOT_DIRECTION;
import static org.lwjgl.opengl.GL11.GL_SPOT_EXPONENT;
import static org.lwjgl.opengl.GL11.GL_TRUE;
import static org.lwjgl.opengl.GL11.glClear;
import static org.lwjgl.opengl.GL11.glClearColor;
import static org.lwjgl.opengl.GL11.glDisable;
import static org.lwjgl.opengl.GL11.glEnable;
import static org.lwjgl.opengl.GL11.glGetInteger;
import static org.lwjgl.opengl.GL11.glLight;
import static org.lwjgl.opengl.GL11.glLightModeli;
import static org.lwjgl.opengl.GL11.glLightf;
import static org.lwjgl.opengl.GL11.glLoadMatrix;
import static org.lwjgl.opengl.GL11.glMatrixMode;
import static org.lwjgl.opengl.GL11.glMultMatrix;
import static org.lwjgl.opengl.GL11.glPopMatrix;
import static org.lwjgl.opengl.GL11.glPushMatrix;
import static org.lwjgl.opengl.GL11.glShadeModel;
import static org.lwjgl.opengl.GL11.glViewport;

import java.awt.Canvas;
import java.awt.Component;
import java.awt.Rectangle;
import java.util.List;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

import org.lwjgl.LWJGLException;
import org.lwjgl.opengl.AWTGLCanvas;

import de.javagl.swogl.DefaultPickingRayComputer;
import de.javagl.swogl.PickingRayComputer;
import de.javagl.swogl.SwoglApplication;
import de.javagl.swogl.SwoglComponentHandler;
import de.javagl.swogl.SwoglRenderer;
import de.javagl.swogl.scene.Camera;
import de.javagl.swogl.scene.CameraListener;
import de.javagl.swogl.scene.CameraUtils;
import de.javagl.swogl.scene.DefaultCamera;
import de.javagl.swogl.scene.DefaultLightSetup;
import de.javagl.swogl.scene.DefaultView;
import de.javagl.swogl.scene.Light;
import de.javagl.swogl.scene.Light.Type;
import de.javagl.swogl.scene.LightSetup;
import de.javagl.swogl.scene.MutableCamera;
import de.javagl.swogl.scene.View;
import de.javagl.swogl.scene.ViewUtils;

/**
 * Implementation of a {@link SwoglApplication} using LWJGL
 */
public class SwoglApplicationLWJGL implements SwoglApplication
{
    /**
     * The {@link SwoglRenderer} used in this application
     */
    private final SwoglRendererLWJGL swoglRenderer;
    
    /**
     * The {@link View} that is used in this application
     */
    private final DefaultView view;
    
    /**
     * The {@link Camera} that is used in this application
     */
    private final MutableCamera camera;
    
    /**
     * The {@link LightSetup} that is used in this application
     */
    private final LightSetup lightSetup;
    
    /**
     * The {@link PickingRayComputer} that is used in this application
     */
    private final PickingRayComputer pickingRayComputer;

    /**
     * The AWTGLCanvas, i.e. the rendering component of this application
     */
    private Canvas canvas;
    
    /**
     * Creates a new SimpleRendererLWJGL
     */
    public SwoglApplicationLWJGL()
    {
        this.view = new DefaultView();
        this.camera = new DefaultCamera();
        this.camera.setEyePoint(new Point3f(0,0,500));
        this.lightSetup = new DefaultLightSetup();
        this.pickingRayComputer = 
            new DefaultPickingRayComputer(getCamera(), getView());
        
        this.swoglRenderer = new SwoglRendererLWJGL();
        initialize();
    }
    
    /**
     * Internal initialization
     */
    public void initialize()
    {
        try
        {
            canvas = new AWTGLCanvas()
            {
                /**
                 * Serial UID
                 */
                private static final long serialVersionUID = 
                    5564898904227409392L;

                @Override
                public void initGL()
                {
                    swoglRenderer.contextChanged();
                }
                
                @Override
                public void paintGL()
                {
                    render();
                    try
                    {
                        swapBuffers();
                    }
                    catch (LWJGLException e)
                    {
                        throw new RuntimeException(
                            "Could not swap buffers", e);
                    }
                }
            };
            
        }
        catch (LWJGLException e)
        {
            throw new RuntimeException(
                "Could not create canvas", e);
        }
        
        camera.addCameraListener(new CameraListener()
        {
            @Override
            public void cameraChanged(Camera camera)
            {
                canvas.repaint();
            }
        });
    }
    
    @Override
    public View getView()
    {
        return view;
    }
    
    @Override
    public MutableCamera getCamera()
    {
        return camera;
    }
    
    @Override 
    public LightSetup getLightSetup()
    {
        return lightSetup;
    }
    
    @Override 
    public PickingRayComputer getPickingRayComputer()
    {
        return pickingRayComputer;
    }
    
    @Override
    public Component getRenderComponent()
    {
        return canvas;
    }
    
    @Override
    public SwoglComponentHandler getSwoglComponentHandler()
    {
        return swoglRenderer;
    }
    
    /**
     * Returns the SwoglRenderer that is used in this application
     * 
     * @return The SwoglRenderer that is used in this application
     */
    protected SwoglRenderer getSwoglRenderer()
    {
        return swoglRenderer;
    }
    
    /**
     * Set up the view according to the current size of the 
     * rendering component
     */
    protected void setupView()
    {
        int width = getRenderComponent().getWidth();
        int height = getRenderComponent().getHeight();
        float aspect = (float)width/height;
        view.setAspect(aspect);
        view.setViewport(new Rectangle(0,0,width,height));
        glViewport(0, 0, width, height);
    }
    
    
    
    /**
     * Renders the current scene. This will call {@link #prepareRender()}
     * and {@link #applyView()}, and then let the SwoglRenderer render
     * all SwoglComponents.
     */
    protected void render() 
    {
        prepareRender();
        setupView();
        applyView();
        
        glPushMatrix();
        Matrix4f viewMatrix = CameraUtils.computeViewMatrix(camera);
        glMultMatrix(RenderingUtils.bufferWith(viewMatrix));
        
        applyLightSetup();
        
        swoglRenderer.renderAllSwoglComponents();
        glPopMatrix();
    }
    
    /**
     * Prepares the rendering pass
     */
    protected void prepareRender()
    {
        glClearColor(0,0,0,0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glEnable(GL_DEPTH_TEST);
        glShadeModel(GL_SMOOTH);
    }
    
    
    /**
     * Apply the {@link LightSetup} of this application
     */
    protected void applyLightSetup()
    {
        List<Light> lights = lightSetup.getLights();
        if (lights.size() > 0)
        {
            glEnable(GL_LIGHTING);
            glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
        }
        else
        {
            glDisable(GL_LIGHTING);
        }
        
        int maxLights = glGetInteger(GL_MAX_LIGHTS);
        for (int i=0; i<maxLights; i++)
        {
            int id = GL_LIGHT0+i;
            glDisable(id);
        }
        
        for (int i=0; i<lights.size(); i++)
        {
            Light light = lights.get(i);
            int id = GL_LIGHT0+i;
            glEnable(id);
            
            glLight(id, GL_AMBIENT, 
                RenderingUtils.bufferWith(light.getAmbientColor()));
            glLight(id, GL_DIFFUSE, 
                RenderingUtils.bufferWith(light.getDiffuseColor()));
            glLight(id, GL_SPECULAR, 
                RenderingUtils.bufferWith(light.getSpecularColor()));

            glLightf(id, GL_CONSTANT_ATTENUATION, 
                light.getConstantAttenuation());
            glLightf(id, GL_LINEAR_ATTENUATION, 
                light.getLinearAttenuation());
            glLightf(id, GL_QUADRATIC_ATTENUATION, 
                light.getQuadraticAttenuation());
            
            // Whether OpenGL interprets the GL_POSITION of a light as a 
            // position or a direction depends on the w-component of the 
            // given tuple. If the w-component of the tuple is 0, then 
            // the GL_POSITION is actually interpreted as the direction 
            // of a directional light.
            // So when the light is DIRECTIONAL, then the direction will
            // be passed in as GL_POSITION, with the w-component being 0.
            // If the light is not DIRECTIONAL, then the position will be
            // passed in as GL_POSITION, but with the w-component being 1,
            // so that it is really interpreted as the position of a 
            // point- or spot light. 
            if (light.getType() == Type.DIRECTIONAL)
            {
                glLight(id, GL_POSITION, 
                    RenderingUtils.bufferWith(light.getDirection(), 0));
                glLightf(id, GL_SPOT_CUTOFF, 180);
            }
            else if (light.getType() == Type.POINT)
            {
                glLight(id, GL_POSITION, 
                    RenderingUtils.bufferWith(light.getPosition(), 1));
                glLightf(id, GL_SPOT_CUTOFF, 180);
            }
            else if (light.getType() == Type.SPOT)
            {
                glLight(id, GL_POSITION, 
                    RenderingUtils.bufferWith(light.getPosition(), 1));
                glLight(id, GL_SPOT_DIRECTION, 
                    RenderingUtils.bufferWith(light.getDirection(), 0));
                glLightf(id, GL_SPOT_CUTOFF, light.getSpotCutoff());
                glLightf(id, GL_SPOT_EXPONENT, light.getSpotExponent());
            }
            
        }
    }
    
    
    /**
     * Apply the current view state to GL.
     */
    protected void applyView()
    {
        glMatrixMode(GL_PROJECTION);
        Matrix4f projectionMatrix = 
            ViewUtils.computeProjectionMatrix(camera.getFovY(), view);
        glLoadMatrix(RenderingUtils.bufferWith(projectionMatrix));
        glMatrixMode(GL_MODELVIEW);        
    }
    
}
