/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;


import java.awt.Component;
import java.awt.Rectangle;

import javax.media.opengl.GL3;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLProfile;
import javax.media.opengl.awt.GLCanvas;
import javax.media.opengl.awt.GLJPanel;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

import de.javagl.swogl.DefaultPickingRayComputer;
import de.javagl.swogl.PickingRayComputer;
import de.javagl.swogl.SwoglApplication;
import de.javagl.swogl.SwoglComponentHandler;
import de.javagl.swogl.SwoglRenderer;
import de.javagl.swogl.scene.Camera;
import de.javagl.swogl.scene.CameraListener;
import de.javagl.swogl.scene.CameraUtils;
import de.javagl.swogl.scene.DefaultCamera;
import de.javagl.swogl.scene.DefaultLightSetup;
import de.javagl.swogl.scene.DefaultView;
import de.javagl.swogl.scene.LightSetup;
import de.javagl.swogl.scene.MutableCamera;
import de.javagl.swogl.scene.View;
import de.javagl.swogl.scene.ViewUtils;

/**
 * Implementation of a {@link SwoglApplication} using JOGL
 * with OpenGL version 3
 */
public class SwoglApplicationShaderJOGL implements SwoglApplication
{
    /**
     * The current GL
     */
    private GL3 gl;

    /**
     * The {@link SwoglRenderer} used in this application
     */
    private final SwoglRendererShaderJOGL swoglRenderer;
    
    /**
     * The {@link View} that is used in this application
     */
    private final DefaultView view;
    
    /**
     * The {@link Camera} that is used in this application
     */
    private final MutableCamera camera;
    
    /**
     * The {@link LightSetup} that is used in this application
     */
    private final LightSetup lightSetup;
    
    /**
     * The {@link PickingRayComputer} that is used in this application
     */
    private final PickingRayComputer pickingRayComputer;

    /**
     * The rendering component of this application
     */
    private Component canvas;
    
    /**
     * Whether a GLJPanel should be used instead of a GLCanvas
     * (experimental)
     */
    private final boolean useGLJPanel;    
    
    /**
     * The GLEventListener that will be attached to the rendering component
     */
    private GLEventListener glEventListener = new GLEventListener()
    {
        @Override
        public void init(GLAutoDrawable drawable)
        {
            GL3 gl = drawable.getGL().getGL3();
            //gl = new DebugGL3(gl);
            SwoglApplicationShaderJOGL.this.gl = gl;
            swoglRenderer.setGL(gl);
            gl.setSwapInterval(0);
        }
        
        @Override
        public void display(GLAutoDrawable drawable) 
        {
            render();
        }
        
        @Override
        public void reshape(
            GLAutoDrawable drawable, int x, int y, int width, int height)
        {
            setupView();
        }

        @Override
        public void dispose(GLAutoDrawable arg0)
        {
        }
    };

    
    /**
     * Creates a new SwoglApplicationShaderJOGL
     */
    public SwoglApplicationShaderJOGL()
    {
        this(false);
    }
    
    /**
     * Creates a new SwoglApplicationShaderJOGL.
     * 
     * @param useGLJPanel Whether a GLJPanel should be used
     * instead of a GLCanvas (experimental)
     */
    public SwoglApplicationShaderJOGL(boolean useGLJPanel)
    {
        this.view = new DefaultView();
        this.camera = new DefaultCamera();
        this.camera.setEyePoint(new Point3f(0,0,500));
        this.lightSetup = new DefaultLightSetup();
        this.pickingRayComputer = 
            new DefaultPickingRayComputer(getCamera(), getView());
        
        this.swoglRenderer = new SwoglRendererShaderJOGL();
        this.useGLJPanel = useGLJPanel;
        initialize();
    }
    
    /**
     * Internal initialization
     */
    private void initialize()
    {
        GLProfile profile = GLProfile.get(GLProfile.GL3);
        GLCapabilities capabilities = new GLCapabilities(profile);
        capabilities.setNumSamples(2);
        capabilities.setSampleBuffers(true);
        
        if (useGLJPanel)
        {
            GLJPanel c = new GLJPanel(capabilities);
            c.addGLEventListener(glEventListener);
            canvas = c;
        }
        else
        {
            GLCanvas c = new GLCanvas(capabilities);
            c.addGLEventListener(glEventListener);
            canvas = c;
        }
        
        camera.addCameraListener(new CameraListener()
        {
            @Override
            public void cameraChanged(Camera camera)
            {
                canvas.repaint();
            }
        });
    }
    
    @Override
    public View getView()
    {
        return view;
    }
    
    @Override
    public MutableCamera getCamera()
    {
        return camera;
    }
    
    @Override 
    public LightSetup getLightSetup()
    {
        return lightSetup;
    }
    
    @Override 
    public PickingRayComputer getPickingRayComputer()
    {
        return pickingRayComputer;
    }
    
    @Override
    public Component getRenderComponent()
    {
        return canvas;
    }
    
    @Override
    public SwoglComponentHandler getSwoglComponentHandler()
    {
        return swoglRenderer;
    }
    
    /**
     * Returns the SwoglRenderer that is used in this application
     * 
     * @return The SwoglRenderer that is used in this application
     */
    protected SwoglRendererShaderJOGL getSwoglRenderer()
    {
        return swoglRenderer;
    }
    
    /**
     * Set up the view according to the current size of the 
     * rendering component
     */
    protected void setupView()
    {
        int width = getRenderComponent().getWidth();
        int height = getRenderComponent().getHeight();
        float aspect = (float)width/height;
        view.setAspect(aspect);
        view.setViewport(new Rectangle(0,0,width,height));
        gl.glViewport(0, 0, width, height);
    }
    
    /**
     * Renders the current scene. This will call {@link #prepareRender()}
     * and pass the light setup and required matrices to the renderer,
     * and then let the SwoglRenderer render all SwoglComponents.
     */
    protected void render() 
    {
        prepareRender();

        swoglRenderer.setLightSetup(getLightSetup());
        
        Matrix4f projectionMatrix = 
            ViewUtils.computeProjectionMatrix(camera.getFovY(), view);
        swoglRenderer.setProjectionMatrix(projectionMatrix);
        
        Matrix4f viewMatrix = CameraUtils.computeViewMatrix(camera);
        swoglRenderer.setViewMatrix(viewMatrix);
        
        swoglRenderer.renderAllSwoglComponents();
    }
    
    /**
     * Prepares the rendering pass
     */
    protected void prepareRender()
    {
        gl.glClearColor(0,0,0,0);
        gl.glClear(GL3.GL_COLOR_BUFFER_BIT | GL3.GL_DEPTH_BUFFER_BIT);
        gl.glEnable(GL3.GL_DEPTH_TEST);
    }

    
    
}
