/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import static org.lwjgl.opengl.GL11.GL_COLOR_BUFFER_BIT;
import static org.lwjgl.opengl.GL11.GL_DEPTH_BUFFER_BIT;
import static org.lwjgl.opengl.GL11.GL_DEPTH_TEST;
import static org.lwjgl.opengl.GL11.glClear;
import static org.lwjgl.opengl.GL11.glClearColor;
import static org.lwjgl.opengl.GL11.glEnable;
import static org.lwjgl.opengl.GL11.glViewport;

import java.awt.Component;
import java.awt.Rectangle;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

import org.lwjgl.LWJGLException;
import org.lwjgl.opengl.AWTGLCanvas;

import de.javagl.swogl.DefaultPickingRayComputer;
import de.javagl.swogl.PickingRayComputer;
import de.javagl.swogl.SwoglApplication;
import de.javagl.swogl.SwoglComponentHandler;
import de.javagl.swogl.SwoglRenderer;
import de.javagl.swogl.scene.Camera;
import de.javagl.swogl.scene.CameraListener;
import de.javagl.swogl.scene.CameraUtils;
import de.javagl.swogl.scene.DefaultCamera;
import de.javagl.swogl.scene.DefaultLightSetup;
import de.javagl.swogl.scene.DefaultView;
import de.javagl.swogl.scene.LightSetup;
import de.javagl.swogl.scene.MutableCamera;
import de.javagl.swogl.scene.View;
import de.javagl.swogl.scene.ViewUtils;

/**
 * Implementation of a {@link SwoglApplication} using LWJGL
 * with OpenGL version 3
 */
public class SwoglApplicationShaderLWJGL implements SwoglApplication
{
    /**
     * The {@link SwoglRenderer} used in this application
     */
    private final SwoglRendererShaderLWJGL swoglRenderer;
    
    /**
     * The {@link View} that is used in this application
     */
    private final DefaultView view;
    
    /**
     * The {@link Camera} that is used in this application
     */
    private final MutableCamera camera;
    
    /**
     * The {@link LightSetup} that is used in this application
     */
    private final LightSetup lightSetup;
    
    /**
     * The {@link PickingRayComputer} that is used in this application
     */
    private final PickingRayComputer pickingRayComputer;

    /**
     * The rendering component of this application
     */
    private Component canvas;
    
    /**
     * Creates a new SwoglApplicationShaderLWJGL
     */
    public SwoglApplicationShaderLWJGL()
    {
        this.view = new DefaultView();
        this.camera = new DefaultCamera();
        this.camera.setEyePoint(new Point3f(0,0,500));
        this.lightSetup = new DefaultLightSetup();
        this.pickingRayComputer = 
            new DefaultPickingRayComputer(getCamera(), getView());
        
        this.swoglRenderer = new SwoglRendererShaderLWJGL();
        initialize();
    }
    
    /**
     * Internal initialization
     */
    private void initialize()
    {
        try
        {
            canvas = new AWTGLCanvas()
            {
                /**
                 * Serial UID
                 */
                private static final long serialVersionUID = 
                    5564898904227409392L;

                @Override
                public void initGL()
                {
                    swoglRenderer.contextChanged();
                }
                
                @Override
                public void paintGL()
                {
                    render();
                    try
                    {
                        swapBuffers();
                    }
                    catch (LWJGLException e)
                    {
                        throw new RuntimeException(
                            "Could not swap buffers", e);
                    }
                }
            };
            
        }
        catch (LWJGLException e)
        {
            throw new RuntimeException(
                "Could not create canvas", e);
        }
        
        camera.addCameraListener(new CameraListener()
        {
            @Override
            public void cameraChanged(Camera camera)
            {
                canvas.repaint();
            }
        });
    }
    
    @Override
    public View getView()
    {
        return view;
    }
    
    @Override
    public MutableCamera getCamera()
    {
        return camera;
    }
    
    @Override 
    public LightSetup getLightSetup()
    {
        return lightSetup;
    }
    
    @Override 
    public PickingRayComputer getPickingRayComputer()
    {
        return pickingRayComputer;
    }
    
    @Override
    public Component getRenderComponent()
    {
        return canvas;
    }
    
    @Override
    public SwoglComponentHandler getSwoglComponentHandler()
    {
        return swoglRenderer;
    }
    
    /**
     * Returns the SwoglRenderer that is used in this application
     * 
     * @return The SwoglRenderer that is used in this application
     */
    protected SwoglRendererShaderLWJGL getSwoglRenderer()
    {
        return swoglRenderer;
    }
    
    /**
     * Set up the view according to the current size of the 
     * rendering component
     */
    protected void setupView()
    {
        int width = getRenderComponent().getWidth();
        int height = getRenderComponent().getHeight();
        float aspect = (float)width/height;
        view.setAspect(aspect);
        view.setViewport(new Rectangle(0,0,width,height));
        glViewport(0, 0, width, height);
    }
    
    /**
     * Renders the current scene. This will call {@link #prepareRender()}
     * and pass the light setup and required matrices to the renderer,
     * and then let the SwoglRenderer render all SwoglComponents.
     */
    protected void render() 
    {
        prepareRender();
        setupView();

        swoglRenderer.setLightSetup(getLightSetup());
        
        Matrix4f projectionMatrix = 
            ViewUtils.computeProjectionMatrix(camera.getFovY(), view);
        swoglRenderer.setProjectionMatrix(projectionMatrix);
        
        Matrix4f viewMatrix = CameraUtils.computeViewMatrix(camera);
        
        //System.out.println("viewMatrix of "+camera+" is\n"+viewMatrix);
        
        swoglRenderer.setViewMatrix(viewMatrix);
        
        swoglRenderer.renderAllSwoglComponents();
    }
    
    /**
     * Prepares the rendering pass
     */
    protected void prepareRender()
    {
        glClearColor(0,0,0,0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glEnable(GL_DEPTH_TEST);
    }

    
    
}
