/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import static javax.media.opengl.GL2.GL_BLEND;
import static javax.media.opengl.GL2.GL_NORMALIZE;
import static javax.media.opengl.GL2.GL_CULL_FACE;
import static javax.media.opengl.GL2.GL_FRONT_AND_BACK;
import static javax.media.opengl.GL2.GL_LINEAR;
import static javax.media.opengl.GL2.GL_LINES;
import static javax.media.opengl.GL2.GL_ONE_MINUS_SRC_ALPHA;
import static javax.media.opengl.GL2.GL_RGBA;
import static javax.media.opengl.GL2.GL_SRC_ALPHA;
import static javax.media.opengl.GL2.GL_TEXTURE_2D;
import static javax.media.opengl.GL2.GL_TEXTURE_MAG_FILTER;
import static javax.media.opengl.GL2.GL_TEXTURE_MIN_FILTER;
import static javax.media.opengl.GL2.GL_TEXTURE_WRAP_S;
import static javax.media.opengl.GL2.GL_TEXTURE_WRAP_T;
import static javax.media.opengl.GL2.GL_TRIANGLES;
import static javax.media.opengl.GL2.GL_UNPACK_ALIGNMENT;
import static javax.media.opengl.GL2.GL_UNSIGNED_BYTE;
import static javax.media.opengl.GL2.GL_CLAMP;
import static javax.media.opengl.GL2.GL_ENABLE_BIT;
import static javax.media.opengl.GL2.GL_LIGHT_MODEL_COLOR_CONTROL;
import static javax.media.opengl.GL2.GL_SEPARATE_SPECULAR_COLOR;
import static javax.media.opengl.GL2.GL_TEXTURE_BIT;
import static javax.media.opengl.GL2.GL_BGRA;
import static javax.media.opengl.GL2.GL_AMBIENT;
import static javax.media.opengl.GL2.GL_DIFFUSE;
import static javax.media.opengl.GL2.GL_EMISSION;
import static javax.media.opengl.GL2.GL_LIGHTING;
import static javax.media.opengl.GL2.GL_SHININESS;
import static javax.media.opengl.GL2.GL_SPECULAR;
import static javax.media.opengl.GL2.GL_MODELVIEW;

import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.IntBuffer;
import java.util.HashMap;
import java.util.Map;

import javax.media.opengl.GL2;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Tuple2f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import de.javagl.swogl.BufferUtils;
import de.javagl.swogl.Material;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglRenderer;
import de.javagl.swogl.Texture;
import de.javagl.swogl.geometry.Geometry;
import de.javagl.swogl.geometry.GeometryUtils;

/**
 * Implementation of a {@link SwoglRenderer} that uses JOGL.
 * <br />
 * This implementation requires the current GL object to work.
 * This GL object should be set by calling {@link #setGL(GL2)}
 * in the <code>GLEventListener#init(GLDrawable)<code> method 
 * of JOGL. 
 */
public class SwoglRendererJOGL 
    extends AbstractSwoglRenderer 
    implements SwoglRenderer
{
    /**
     *  The current GL
     */
    private GL2 gl;
    
    /**
     * The map from Textures to GL texture IDs
     */
    private final Map<Texture, Integer> textureIDs;
    
    /**
     * A buffer which are used for the update
     */
    private IntBuffer textureBuffer;
    
    /**
     * Creates a new SwoglRendererJOGL
     */
    public SwoglRendererJOGL()
    {
        textureIDs = new HashMap<Texture, Integer>();
    }
    
    /**
     * Set the current GL. When the current GL changes due to this call,
     * this will cause all existing textures to be re-initialized.
     * 
     * @param gl The current GL
     */
    public void setGL(GL2 gl)
    {
        boolean contextChanged = false;
        if (this.gl != null && this.gl != gl)
        {
            // When the GL changes, everything has to be re-initialized
            contextChanged = true;
        }
        this.gl = gl;
        if (contextChanged)
        {
            contextChanged();
            contextChanged = false;
        }
    }
    
    /**
     * Should be called when the GL context changed. This will re-initialize
     * all GL objects that may have become invalid due to a changed context.
     */
    void contextChanged()
    {
        Map<Texture, Integer> oldTextureIDs = 
            new HashMap<Texture, Integer>(textureIDs);
        textureIDs.clear();
        for (Texture texture : oldTextureIDs.keySet())
        {
            initTexture(texture);
        }
    }
    

    @Override
    protected void handleSwoglComponentInternal(SwoglComponent swoglComponent)
    {
        Texture texture = swoglComponent.getTexture();
        initTexture(texture);
    }
    
    private void initTexture(Texture texture)
    {
        int textureIDArray[] = {0};
        gl.glGenTextures(1, textureIDArray, 0);
        int textureID = textureIDArray[0];
        textureIDs.put(texture, textureID);
        
        gl.glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        
        gl.glBindTexture(GL_TEXTURE_2D, textureID);
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        
        updateTextureInternal(texture);
    }
    
    
    @Override
    protected void updateTextureInternal(Texture texture)
    {
        BufferedImage image = texture.getImage();
        int w = image.getWidth();
        int h = image.getHeight();

        if (textureBuffer == null || textureBuffer.capacity() < w*h)
        {
            ByteBuffer buffer = BufferUtils.createByteBuffer(w*h*4);
            textureBuffer = 
                buffer.order(ByteOrder.LITTLE_ENDIAN).asIntBuffer();
        }
        ImageUtils.writeImageToBuffer(image, textureBuffer);
        textureBuffer.rewind();

        int textureID = textureIDs.get(texture);
        gl.glBindTexture(GL_TEXTURE_2D, textureID);
        gl.glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w,
            h, 0, GL_BGRA, GL_UNSIGNED_BYTE, textureBuffer);
    }
    
    

    
    @Override
    protected void releaseSwoglComponentInternal(
        SwoglComponent swoglComponent)
    {
        Texture texture = swoglComponent.getTexture();
        releaseTexture(texture);
    }

    private void releaseTexture(Texture texture)
    {
        int textureID = textureIDs.get(texture);
        gl.glDeleteTextures(1, new int[]{textureID}, 0);
        textureIDs.remove(texture);
    }
    

    @Override
    protected void updateVerticesInternal(Geometry geometry)
    {
        GeometryUtils.updateNormals(geometry);
    }
    
    @Override
    protected void renderSwoglComponentInternal(
        SwoglComponent swoglComponent)
    {
        Geometry geometry = swoglComponent.getGeometry();
        Texture texture = swoglComponent.getTexture();

        gl.glMatrixMode(GL_MODELVIEW);        
        gl.glPushMatrix();
        
        //paintCoordinateSystem();
        
        Matrix4f transformMatrix = swoglComponent.getTransform();
        gl.glMultMatrixf(RenderingUtils.bufferWith(transformMatrix));

        //paintNormals(geometry);
        
        gl.glPushAttrib(
            GL_ENABLE_BIT |
            GL_TEXTURE_BIT); 
        
        gl.glDisable(GL_CULL_FACE);
        gl.glEnable(GL_BLEND);
        gl.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        gl.glEnable(GL_TEXTURE_2D);
        gl.glEnable(GL_NORMALIZE);
        
        int textureID = textureIDs.get(texture);
        gl.glBindTexture(GL_TEXTURE_2D, textureID);
        
        applyMaterial(swoglComponent.getMaterial());
        
        Tuple3f vertex = new Point3f();
        Tuple3f normal = new Vector3f();
        Tuple2f texCoords = new TexCoord2f();
        
        gl.glBegin(GL_TRIANGLES);
        for (int i=0; i<geometry.getNumTriangles(); i++)
        {
            for (int j=0; j<3; j++)
            {
                GeometryUtils.getVertex(
                    geometry, i, j,  vertex, normal, texCoords);

                gl.glTexCoord2f(texCoords.x, texCoords.y);
                gl.glNormal3f(normal.x, normal.y, normal.z);
                gl.glVertex3f(vertex.x, vertex.y, vertex.z);
            }
        }
        gl.glEnd();
        gl.glPopMatrix();
        
        gl.glPopAttrib();
    }
    
    /**
     * Apply the given {@link Material} 
     * 
     * @param material The {@link Material}
     */
    private void applyMaterial(Material material)
    {
        gl.glLightModeli(
            GL_LIGHT_MODEL_COLOR_CONTROL, 
            GL_SEPARATE_SPECULAR_COLOR);
        gl.glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, 
            RenderingUtils.bufferWith(material.getAmbientColor()));
        gl.glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, 
            RenderingUtils.bufferWith(material.getDiffuseColor()));
        gl.glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, 
            RenderingUtils.bufferWith(material.getSpecularColor()));
        gl.glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, 
            RenderingUtils.bufferWith(material.getEmissionColor()));
        gl.glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, 
            RenderingUtils.bufferWith(material.getShininess()));
    }
    
    
    
    /**
     * Paints a simple coordinate system for debugging
     */
    @SuppressWarnings("unused")
    private void paintCoordinateSystem()
    {
        gl.glDisable(GL_LIGHTING);
        gl.glBegin(GL_LINES);
        gl.glColor3f(1,0,0);
        gl.glVertex3f(   0,   0,   0);
        gl.glVertex3f(1000,   0,   0);
        gl.glColor3f(0,1,0);
        gl.glVertex3f(   0,   0,   0);
        gl.glVertex3f(   0,1000,   0);
        gl.glColor3f(0,0,1);
        gl.glVertex3f(   0,   0,   0);
        gl.glVertex3f(   0,   0,1000);
        gl.glEnd();
        gl.glEnable(GL_LIGHTING);
    }

    
    /**
     * Paints a simple coordinate system for debugging
     */
    @SuppressWarnings("unused")
    private void paintNormals(Geometry geometry)
    {
        gl.glDisable(GL_LIGHTING);
        gl.glBegin(GL_LINES);
        gl.glColor3f(1,0,0);
        Point3f v = new Point3f();
        Vector3f n = new Vector3f();
        for (int i=0; i<geometry.getNumVertices(); i++)
        {
            geometry.getVertex(i, v);
            geometry.getNormal(i, n);
            n.scale(100);
            gl.glVertex3f(v.x, v.y, v.z);
            gl.glVertex3f(v.x+n.x, v.y+n.y, v.z+n.z);
        }
        gl.glEnd();
        gl.glEnable(GL_LIGHTING);
    }

    
    

}
