/*
 * Swogl - Swing meets OpenGL
 * 
 * Copyright 2007-2012 Marco Hutter - http://swogl.javagl.de
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package de.javagl.swogl.rendering;

import static javax.media.opengl.GL.GL_ARRAY_BUFFER;
import static javax.media.opengl.GL.GL_DYNAMIC_DRAW;
import static javax.media.opengl.GL.GL_ELEMENT_ARRAY_BUFFER;
import static javax.media.opengl.GL.GL_FLOAT;
import static javax.media.opengl.GL.GL_LINEAR;
import static javax.media.opengl.GL.GL_REPEAT;
import static javax.media.opengl.GL.GL_RGBA;
import static javax.media.opengl.GL.GL_STATIC_DRAW;
import static javax.media.opengl.GL.GL_TEXTURE_2D;
import static javax.media.opengl.GL.GL_TEXTURE_MAG_FILTER;
import static javax.media.opengl.GL.GL_TEXTURE_MIN_FILTER;
import static javax.media.opengl.GL.GL_TEXTURE_WRAP_S;
import static javax.media.opengl.GL.GL_TEXTURE_WRAP_T;
import static javax.media.opengl.GL.GL_TRIANGLES;
import static javax.media.opengl.GL.GL_UNSIGNED_BYTE;
import static javax.media.opengl.GL.GL_UNSIGNED_INT;
import static javax.media.opengl.GL2.GL_FRAGMENT_SHADER;
import static javax.media.opengl.GL2.GL_STREAM_DRAW;
import static javax.media.opengl.GL2.GL_VERTEX_SHADER;
import static javax.media.opengl.GL2.GL_BGRA;
import static javax.media.opengl.GL2.GL_PIXEL_UNPACK_BUFFER;
import static javax.media.opengl.GL2.GL_READ_WRITE;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Map;

import javax.media.opengl.GL3;
import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Tuple4f;
import javax.vecmath.Vector3f;

import de.javagl.swogl.BufferUtils;
import de.javagl.swogl.Material;
import de.javagl.swogl.MatrixUtils;
import de.javagl.swogl.SwoglApplication;
import de.javagl.swogl.SwoglComponent;
import de.javagl.swogl.SwoglRenderer;
import de.javagl.swogl.Texture;
import de.javagl.swogl.geometry.Geometry;
import de.javagl.swogl.geometry.GeometryUtils;
import de.javagl.swogl.scene.Light;
import de.javagl.swogl.scene.Light.Type;
import de.javagl.swogl.scene.LightSetup;

/**
 * Implementation of a {@link SwoglRenderer} that uses JOGL.
 * <br />
 * This implementation requires the current GL object to work.
 * This GL object should be set by calling {@link #setGL(GL3)}
 * in the <code>GLEventListener#init(GLDrawable)<code> method 
 * of JOGL. If this renderer was created by a 
 * {@link SwoglApplication}, then it will call this
 * method accordingly.
 */
public class SwoglRendererShaderJOGL 
    extends AbstractSwoglRenderer 
    implements SwoglRenderer
{
    /**
     * The identifier for directional lights used in the shader
     */
    private static final int SHADER_LIGHT_TYPE_DIRECTIONAL = 0;

    /**
     * The identifier for point lights used in the shader
     */
    private static final int SHADER_LIGHT_TYPE_POINT = 1;

    /**
     * The identifier for spot lights used in the shader
     */
    private static final int SHADER_LIGHT_TYPE_SPOT = 2;
    
    /**
     * A debugging flag indicating whether the access to non-existent
     * uniforms should cause a log message to be printed
     */
    private static final boolean LOG_LOCATIONS = true;
    
    /**
     * A debugging flag indicating whether log output of
     * the shader compilation should be printed
     */
    private static final boolean LOG_COMPILATION = false;
    
    /**
     *  The current GL
     */
    private GL3 gl;
    
    /**
     * A buffer which is used for the initialization of 
     * the geometry indices
     */
    private IntBuffer geometryIndicesBuffer;
    
    /**
     * A buffer which is used for the initialization and
     * update of geometry data (vertices, normals...)
     */
    private FloatBuffer geometryDataBuffer;

    /**
     * The current view matrix
     */
    private final Matrix4f viewMatrix;
    
    /**
     * The current projection matrix
     */
    private final Matrix4f projectionMatrix;

    /**
     * The {@link LightSetup} of the {@link SwoglApplication}
     */
    private LightSetup lightSetup;
    
    /**
     * The OpenGL shader program ID
     */
    private int programID;
    
    
    /**
     * The map from {@link Texture} objects to the corresponding
     * {@link TextureGL} objects.
     */
    private final Map<Texture, TextureGL> textures;
    
    /**
     * The map from {@link Geometry} objects to the corresponding
     * {@link GeometryGL} objects
     */
    private final Map<Geometry, GeometryGL> geometries;

    
    
    /**
     * The OpenGL data for a {@link Texture}
     */
    private static class TextureGL
    {
        /**
         * The texture ID
         */
        private final int textureID;
        
        /**
         * The pixel buffer object 
         */
        private final int pbo;
        
        /**
         * Creates a new TextureGL with the given values
         * 
         * @param textureID The texture ID
         * @param pbo The pixel buffer object 
         */
        private TextureGL(int textureID, int pbo)
        {
            this.textureID = textureID;
            this.pbo = pbo;
        }
    }
    
    /**
     * The OpenGL data for a {@link Geometry}
     */
    private static class GeometryGL
    {
        /**
         * The vertex array object
         */
        private final int vao;
        
        /**
         * The indices vertex buffer object
         */
        private final int indicesVBO;
        
        /**
         * The vertex positions vertex buffer object
         */
        private final int verticesVBO;
        
        /**
         * The normals vertex buffer object
         */
        private final int normalsVBO;
        
        /**
         * The texture coordinates vertex buffer object
         */
        private final int texCoordsVBO;

        /**
         * Creates a new GeometryGL with the given values 
         * @param vao The vertex array object
         * @param indicesVBO The indices vertex buffer object
         * @param verticesVBO The vertex positions vertex buffer object
         * @param normalsVBO The normals vertex buffer object
         * @param texCoordsVBO The texture coordinates vertex buffer object
         */
        private GeometryGL(
            int vao, int indicesVBO, int verticesVBO, 
            int normalsVBO, int texCoordsVBO)
        {
            this.vao = vao;
            this.indicesVBO = indicesVBO;
            this.verticesVBO = verticesVBO;
            this.normalsVBO = normalsVBO;
            this.texCoordsVBO = texCoordsVBO;
        }
    }
    
    
    /**
     * Creates a new SwoglRendererShaderJOGL
     */
    public SwoglRendererShaderJOGL()
    {
        textures = new IdentityHashMap<Texture, TextureGL>();    
        geometries = new IdentityHashMap<Geometry, GeometryGL>();    
        
        projectionMatrix = new Matrix4f();
        projectionMatrix.setIdentity();
        
        viewMatrix = new Matrix4f();
        viewMatrix.setIdentity();
    }
    
    /**
     * Set the current GL. When the current GL changes due to this call,
     * this will cause all existing textures to be re-initialized.
     * 
     * @param gl The current GL
     */
    public void setGL(GL3 gl)
    {
        boolean contextInitialized = false;
        if (this.gl == null && gl != null)
        {
            contextInitialized = true;
        }
        boolean contextChanged = false;
        if (this.gl != null && this.gl != gl)
        {
            // When the GL changes, everything has to be re-initialized
            contextChanged = true;
        }
        this.gl = gl;
        if (contextInitialized)
        {
            initProgram();
        }
        if (contextChanged)
        {
            contextChanged();
            contextChanged = false;
        }
    }
    
    /**
     * Should be called when the GL context changed. This will re-initialize
     * all GL objects that may have become invalid due to a changed context.
     */
    void contextChanged()
    {
        initProgram();
        Map<Texture, TextureGL> oldTextures = 
            new HashMap<Texture, TextureGL>(textures);
        textures.clear();
        for (Texture texture : oldTextures.keySet())
        {
            initTexture(texture);
        }
        Map<Geometry, GeometryGL> oldGeometries = 
            new HashMap<Geometry, GeometryGL>(geometries);
        geometries.clear();
        for (Geometry geometry : oldGeometries.keySet())
        {
            initGeometry(geometry);
        }
    }
    
    
    /**
     * Initialize the program, consisting of vertex- and fragment shader
     */
    protected void initProgram()
    {
        if (programID != 0)
        {
            gl.glDeleteProgram(programID);
        }
        programID  = gl.glCreateProgram();
        
        String vertexShaderCode = readStream("/shaders/swoglShader.vs");
        int vertexShaderID = gl.glCreateShader(GL_VERTEX_SHADER);
        gl.glShaderSource(
            vertexShaderID, 1, new String[]{vertexShaderCode}, null);
        gl.glCompileShader(vertexShaderID);
        if (LOG_COMPILATION) printShaderLogInfo(vertexShaderID);
        gl.glAttachShader(programID, vertexShaderID);
        gl.glDeleteShader(vertexShaderID);

        String fragmentShaderCode = readStream("/shaders/swoglShader.fs");
        int fragmentShaderID = gl.glCreateShader(GL_FRAGMENT_SHADER);
        gl.glShaderSource(
            fragmentShaderID, 1, new String[]{fragmentShaderCode}, null);
        gl.glCompileShader(fragmentShaderID);     
        if (LOG_COMPILATION) printShaderLogInfo(fragmentShaderID);
        gl.glAttachShader(programID, fragmentShaderID);
        gl.glDeleteShader(fragmentShaderID);

        gl.glLinkProgram(programID);
        gl.glValidateProgram(programID);
        if (LOG_COMPILATION) printProgramLogInfo(programID);
    }
    

    @Override
    protected void handleSwoglComponentInternal(
        SwoglComponent swoglComponent)
    {
        Texture texture = swoglComponent.getTexture();
        initTexture(texture);
        
        Geometry geometry = swoglComponent.getGeometry();
        initGeometry(geometry);
    }
    

    /**
     * Initialize the GL representation of the given {@link Texture}
     * 
     * @param texture The {@link Texture}
     */
    private void initTexture(Texture texture)
    {
        // Create the texture and its PBO
        BufferedImage image = texture.getImage();
        int w = image.getWidth();
        int h = image.getHeight();
        
        int textureIDArray[] = {0};
        gl.glGenTextures(1, textureIDArray, 0);
        int textureID = textureIDArray[0];
        
        int pboArray[] = {0};
        gl.glGenBuffers(1, pboArray, 0);
        int pbo = pboArray[0];
        
        // Initialize the texture parameters
        int elements = 4;
        gl.glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo);
        gl.glBufferData(GL_PIXEL_UNPACK_BUFFER, 
            w * h * elements, null, GL_STREAM_DRAW);
        gl.glBindTexture(GL_TEXTURE_2D, textureID);
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); 
        gl.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT); 

        gl.glBindTexture(GL_TEXTURE_2D, 0);
        gl.glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
        
        // Store the GL representation of the texture 
        TextureGL textureGL = new TextureGL(textureID, pbo);
        textures.put(texture, textureGL);
        
        updateTextureInternal(texture);
    }

    
    /**
     * Initialize the GL data for the given {@link Geometry}
     * 
     * @param geometry The {@link Geometry}
     */
    private void initGeometry(Geometry geometry)
    {
        final int intSize = 4;
        final int floatSize = 4;
        
        // Create and bind the VAO
        int vaoArray[] = {0};
        gl.glGenVertexArrays(1, vaoArray, 0);
        int vao = vaoArray[0];
        gl.glBindVertexArray(vao);
        
        // Create the indices VBO
        int vboArray[] = {0};
        gl.glGenBuffers(1, vboArray, 0);
        int indicesVBO = vboArray[0];

        // Fill the indices VBO
        int numIndices = geometry.getNumTriangles() * 3;
        geometryIndicesBuffer = 
            BufferUtils.ensureBuffer(geometryIndicesBuffer, numIndices);
        GeometryUtils.writeIndicesToBuffer(geometry, geometryIndicesBuffer);
        gl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indicesVBO);
        gl.glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
            numIndices * intSize, geometryIndicesBuffer, GL_STATIC_DRAW);
        gl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        

        int numVertices = geometry.getNumVertices();
        geometryDataBuffer = 
            BufferUtils.ensureBuffer(geometryDataBuffer, numVertices*3);
        
        // Create and fill the vertices VBO
        gl.glGenBuffers(1, vboArray, 0);
        int verticesVBO = vboArray[0];
        gl.glBindBuffer(GL_ARRAY_BUFFER, verticesVBO);
        GeometryUtils.writeVerticesToBuffer(geometry, geometryDataBuffer);
        gl.glBufferData(GL_ARRAY_BUFFER, numVertices * 3 * floatSize,
            geometryDataBuffer, GL_DYNAMIC_DRAW);
        gl.glBindBuffer(GL_ARRAY_BUFFER, 0);

        
        // Create and fill the normals VBO
        gl.glGenBuffers(1, vboArray, 0);
        int normalsVBO = vboArray[0];
        gl.glBindBuffer(GL_ARRAY_BUFFER, normalsVBO);
        GeometryUtils.writeNormalsToBuffer(geometry, geometryDataBuffer);
        gl.glBufferData(GL_ARRAY_BUFFER, numVertices * 3 * floatSize,
            geometryDataBuffer, GL_DYNAMIC_DRAW);
        gl.glBindBuffer(GL_ARRAY_BUFFER, 0);

        
        // Create and fill the texCoords VBO
        gl.glGenBuffers(1, vboArray, 0);
        int texCoordsVBO = vboArray[0];
        gl.glBindBuffer(GL_ARRAY_BUFFER, texCoordsVBO);
        GeometryUtils.writeTexCoordsToBuffer(geometry, geometryDataBuffer);
        gl.glBufferData(GL_ARRAY_BUFFER, numVertices * 2 * floatSize,
            geometryDataBuffer, GL_DYNAMIC_DRAW);
        gl.glBindBuffer(GL_ARRAY_BUFFER, 0);

        // Bind the attributes to the input locations
        bindAttribute("inPosition", verticesVBO, 3);
        bindAttribute("inNormal", normalsVBO, 3);
        bindAttribute("inTexCoord", texCoordsVBO, 2);

        // Create and store the GL representation of the geometry
        GeometryGL geometryGL = new GeometryGL(
            vao, indicesVBO, verticesVBO, normalsVBO, texCoordsVBO);
        geometries.put(geometry, geometryGL);
    }
    
    /**
     * Bind the attribute with the given name to the given VBO
     * 
     * @param name The attribute name
     * @param vbo The VBO
     * @param size The element size (in number of float components)
     */
    private void bindAttribute(String name, int vbo, int size)
    {
        int location = gl.glGetAttribLocation(programID, name);
        if (location != -1)
        {
            gl.glBindBuffer(GL_ARRAY_BUFFER, vbo);
            gl.glVertexAttribPointer(location, size, GL_FLOAT, false, 0, 0);
            gl.glEnableVertexAttribArray(location);
        }
        else
        {
            if (LOG_LOCATIONS) System.out.println("No location of "+name);
        }
    }
    
    
    @Override
    protected void updateTextureInternal(Texture texture)
    {
        BufferedImage image = texture.getImage();
        int w = image.getWidth();
        int h = image.getHeight();

        // Obtain the GL representation of the texture
        TextureGL textureGL = textures.get(texture);
        int textureID = textureGL.textureID;
        int pbo = textureGL.pbo;

        // Map the PBO into client space and 
        // fill it with the image data 
        gl.glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo);
        gl.glBindTexture(GL_TEXTURE_2D, textureID);
    
        ByteBuffer mappedPBOBuffer = 
            gl.glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_READ_WRITE);
        IntBuffer intBuffer = mappedPBOBuffer.order(
            ByteOrder.nativeOrder()).asIntBuffer();
        ImageUtils.writeImageToBuffer(image, intBuffer);
        
        gl.glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);

        // Update the texture contents
        int type = GL_UNSIGNED_BYTE;
        int internalFormat = GL_RGBA;
        int format = GL_BGRA;
        gl.glTexImage2D(
            GL_TEXTURE_2D, 0, internalFormat, w, h, 
            0, format, type, 0);
        gl.glBindTexture(GL_TEXTURE_2D, 0);
        gl.glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
    }
    
    

    
    @Override
    protected void releaseSwoglComponentInternal(
        SwoglComponent swoglComponent)
    {
        Texture texture = swoglComponent.getTexture();
        releaseTexture(texture);
        
        Geometry geometry = swoglComponent.getGeometry();
        releaseGeometry(geometry);
    }

    /**
     * Release the GL resources that are associated with 
     * the given {@link Texture}
     * 
     * @param texture The {@link Texture}
     */
    private void releaseTexture(Texture texture)
    {
        TextureGL textureGL = textures.get(texture);
        int textureID = textureGL.textureID;
        int pbo = textureGL.pbo;
        
        gl.glDeleteTextures(1, new int[]{textureID}, 0);
        gl.glDeleteBuffers(1, new int[]{pbo}, 0);
        textures.remove(texture);
    }
    
    /**
     * Release the GL resources that are associated with
     * the given {@link Geometry}
     * 
     * @param geometry The {@link Geometry}
     */
    private void releaseGeometry(Geometry geometry)
    {
        GeometryGL geometryGL = geometries.get(geometry);
        int vao = geometryGL.vao;
        int indicesVBO = geometryGL.indicesVBO;
        int verticesVBO = geometryGL.verticesVBO;
        int normalsVBO = geometryGL.normalsVBO;
        int texCoordsVBO = geometryGL.texCoordsVBO;
        
        gl.glDeleteBuffers(1, new int[]{indicesVBO}, 0);
        gl.glDeleteBuffers(1, new int[]{verticesVBO}, 0);
        gl.glDeleteBuffers(1, new int[]{normalsVBO}, 0);
        gl.glDeleteBuffers(1, new int[]{texCoordsVBO}, 0);
        gl.glDeleteVertexArrays(1, new int[]{vao}, 0);
        
        geometries.remove(geometry);
    }
    

    @Override
    protected void updateVerticesInternal(Geometry geometry)
    {
        ByteBuffer mappedBuffer = null;
        FloatBuffer floatBuffer = null;
        GeometryGL geometryGL = geometries.get(geometry);

        // Map the vertices VBO into client space, and fill it with
        // the vertex data of the geometry 
        gl.glBindBuffer(GL_ARRAY_BUFFER, geometryGL.verticesVBO);
        mappedBuffer = gl.glMapBuffer(GL_ARRAY_BUFFER, GL_READ_WRITE);
        floatBuffer = mappedBuffer.order(
            ByteOrder.nativeOrder()).asFloatBuffer();
        GeometryUtils.writeVerticesToBuffer(geometry, floatBuffer);
        gl.glUnmapBuffer(GL_ARRAY_BUFFER);
        gl.glBindBuffer(GL_ARRAY_BUFFER, 0);
        
        GeometryUtils.updateNormals(geometry);
        
        // Map the normals VBO into client space, and fill it with
        // the normal data of the geometry 
        gl.glBindBuffer(GL_ARRAY_BUFFER, geometryGL.normalsVBO);
        mappedBuffer = gl.glMapBuffer(GL_ARRAY_BUFFER, GL_READ_WRITE);
        floatBuffer = mappedBuffer.order(
            ByteOrder.nativeOrder()).asFloatBuffer();
        GeometryUtils.writeNormalsToBuffer(geometry, floatBuffer);
        gl.glUnmapBuffer(GL_ARRAY_BUFFER);
        gl.glBindBuffer(GL_ARRAY_BUFFER, 0);
        
    }
    

    
    @Override
    protected void renderSwoglComponentInternal(
        SwoglComponent swoglComponent)
    {
        Geometry geometry = swoglComponent.getGeometry();
        Texture texture = swoglComponent.getTexture();
        Material material = swoglComponent.getMaterial();

        // Compute the matrices that should be passed to the shader
        Matrix4f modelMatrix = swoglComponent.getTransform();
        Matrix4f modelviewMatrix = 
            MatrixUtils.mul(viewMatrix, modelMatrix);
        Matrix4f normalMatrix = 
            MatrixUtils.transpose(MatrixUtils.invert(modelviewMatrix));
        
        gl.glUseProgram(programID);

        // Set the matrices for the shader
        setMatrix("viewMatrix", viewMatrix);
        setMatrix("projectionMatrix", projectionMatrix);
        setMatrix("modelMatrix", modelMatrix);
        setMatrix("normalMatrix", normalMatrix);

        // Set the material properties for the shader
        setTuple4f("material.ambient", material.getAmbientColor());
        setTuple4f("material.diffuse", material.getDiffuseColor());
        setTuple4f("material.specular", material.getSpecularColor());
        setTuple4f("material.emission", material.getEmissionColor());
        setFloat("material.shininess", material.getShininess());

        // Set up the lights for the shader
        setTuple4f("globalAmbient", lightSetup.getAmbient());
        int numLights = lightSetup.getLights().size();
        setInt("numLights", numLights);
        
        for (int i=0; i<numLights; i++)
        {
            Light light = lightSetup.getLights().get(i);
            String id = "lights["+i+"]";

            //System.out.println("Light "+i+" is "+light);
            
            setTuple4f(id+".ambient", light.getAmbientColor());
            setTuple4f(id+".diffuse", light.getDiffuseColor());
            setTuple4f(id+".specular", light.getSpecularColor());

            setFloat(id+".constantAttenuation", 
                light.getConstantAttenuation());
            setFloat(id+".linearAttenuation", 
                light.getLinearAttenuation());
            setFloat(id+".quadraticAttenuation", 
                light.getQuadraticAttenuation());
            
            // When the w-component of the position is 0, then the "position" 
            // is actually the direction of a directional light. So
            // when the light is directional and NO spot light, then the 
            // getDirection of this light will be passed to OpenGL which will 
            // interpret it as a direction due to the w-component of the 
            // given 4-tuple being 0.0. If this Light is NOT directional or is 
            // a spot light, then the getPosition will be passed to OpenGL, 
            // which will interpret it as a position (sic) due to the 
            // w-component of the given 4-tuple being 1.0.
            if (light.getType() == Type.DIRECTIONAL)
            {        
                setInt(id+".type", SHADER_LIGHT_TYPE_DIRECTIONAL); 
                
                Vector3f p = light.getDirection();
                viewMatrix.transform(p);
                setTuple4f(id+".position", p, 0.0f);
            }
            else if (light.getType() == Type.POINT)
            {
                setInt(id+".type", SHADER_LIGHT_TYPE_POINT);
                
                Point3f p = light.getPosition();
                viewMatrix.transform(p);
                setTuple4f(id+".position", p, 1.0f);
            }
            else if (light.getType() == Type.SPOT)
            {
                setInt(id+".type", SHADER_LIGHT_TYPE_SPOT);
                
                Point3f p = light.getPosition();
                viewMatrix.transform(p);
                setTuple4f(id+".position", p, 1.0f);
                
                Vector3f v = light.getDirection();
                viewMatrix.transform(v);
                
                setTuple3f(id+".spotDirection", v);
                setFloat(id+".spotExponent", light.getSpotExponent());
                setFloat(id+".spotCutoff", light.getSpotCutoff());
            }
        }
        
        
        // Activate and bind the texture
        TextureGL textureGL = textures.get(texture);
        int textureIndex = 0;
        gl.glActiveTexture(GL3.GL_TEXTURE0+textureIndex);
        gl.glBindTexture(GL_TEXTURE_2D, textureGL.textureID);
        int location = gl.glGetAttribLocation(programID, "texture");
        gl.glUniform1i(location, textureIndex);
        textureIndex++;
        
        // Finally, render. That's what this is all about...
        GeometryGL geometryGL = geometries.get(geometry);
        gl.glUseProgram(programID);
        gl.glBindVertexArray(geometryGL.vao);
        gl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, geometryGL.indicesVBO);
        gl.glDrawElements(GL_TRIANGLES, 
            geometry.getNumTriangles()*3, GL_UNSIGNED_INT, 0);
        gl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        gl.glBindVertexArray(0);
        gl.glBindTexture(GL_TEXTURE_2D, 0);
        gl.glUseProgram(0);
    }
    

    
    /**
     * Set the uniform with the given name to the given value
     * 
     * @param name The name
     * @param matrix The value
     */
    private void setMatrix(String name, Matrix4f matrix)
    {
        int location = gl.glGetUniformLocation(programID, name);
        
        //System.out.println("Matrix "+name+" loc "+location+" is \n"+matrix);
        
        if (location != -1)
        {
            gl.glUniformMatrix4fv(location, 1, false, RenderingUtils.bufferWith(matrix));
        }
        else
        {
            if (LOG_LOCATIONS) System.out.println("No location of "+name);
        }
    }
    
    /**
     * Set the uniform with the given name to the given value
     * 
     * @param name The name
     * @param tuple The value
     */
    private void setTuple3f(String name, Tuple3f tuple)
    {
        int location = gl.glGetUniformLocation(programID, name);
        if (location != -1)
        {
            gl.glUniform3f(location, tuple.x, tuple.y, tuple.z);        
        }
        else
        {
            if (LOG_LOCATIONS) System.out.println("No location of "+name);
        }
    }
    
    /**
     * Set the uniform with the given name to the given value
     * 
     * @param name The name
     * @param tuple The value
     */
    private void setTuple4f(String name, Tuple4f tuple)
    {
        int location = gl.glGetUniformLocation(programID, name);
        if (location != -1)
        {
            gl.glUniform4f(location, tuple.x, tuple.y, tuple.z, tuple.w);        
        }
        else
        {
            if (LOG_LOCATIONS) System.out.println("No location of "+name);
        }
    }
    
    /**
     * Set the uniform with the given name to the given value
     * 
     * @param name The name
     * @param tuple The value
     * @param w The fourth component
     */
    private void setTuple4f(String name, Tuple3f tuple, float w)
    {
        int location = gl.glGetUniformLocation(programID, name);
        if (location != -1)
        {
            gl.glUniform4f(location, tuple.x, tuple.y, tuple.z, w);        
        }
        else
        {
            if (LOG_LOCATIONS) System.out.println("No location of "+name);
        }
    }
    
    /**
     * Set the uniform with the given name to the given value
     * 
     * @param name The name
     * @param value The value
     */
    private void setInt(String name, int value)
    {
        int location = gl.glGetUniformLocation(programID, name);
        if (location != -1)
        {
            gl.glUniform1i(location, value);        
        }
        else
        {
            if (LOG_LOCATIONS) System.out.println("No location of "+name);
        }
    }
    
    /**
     * Set the uniform with the given name to the given value
     * 
     * @param name The name
     * @param value The value
     */
    private void setFloat(String name, float value)
    {
        int location = gl.glGetUniformLocation(programID, name);
        if (location != -1)
        {
            gl.glUniform1f(location, value);        
        }
        else
        {
            if (LOG_LOCATIONS) System.out.println("No location of "+name);
        }
    }
    
    
    /**
     * Set the current projection matrix to have the same contents
     * as the given one
     * 
     * @param projectionMatrix The matrix
     */
    void setProjectionMatrix(Matrix4f projectionMatrix)
    {
        this.projectionMatrix.set(projectionMatrix);
    }

    /**
     * Set the current view matrix to have the same contents
     * as the given one
     * 
     * @param viewMatrix The matrix
     */
    void setViewMatrix(Matrix4f viewMatrix)
    {
        this.viewMatrix.set(viewMatrix);
    }
    
    /**
     * Set the {@link LightSetup} that should be used during
     * rendering
     * 
     * @param lightSetup The {@link LightSetup}
     */
    void setLightSetup(LightSetup lightSetup)
    {
        this.lightSetup = lightSetup;
    }


    
    /**
     * Read the resource with the given name, and return its contents
     * as a String.
     * 
     * @param resourceName The resource name
     * @return The resource contents
     * @throws RuntimeException If the input can not be read
     */
    private static String readStream(String resourceName)
    {
        InputStream stream = null;
        try
        {
            stream = SwoglRendererShaderJOGL.class.getResourceAsStream(resourceName);
            String result = readStream(stream);
            stream.close();
            return result;
        }
        catch (IOException e)
        {
            throw new RuntimeException("Could not read shader", e);
        }
        finally
        {
            try
            {
                stream.close();
            }
            catch (IOException e)
            {
                throw new RuntimeException("Could not close stream", e);
            }
        }
            
    }
    
    /**
     * Read the contents of the given stream and return it as a string.
     * The caller is responsible for closing the stream.
     * 
     * @param stream The stream
     * @return The contents of the stream, as a string
     * @throws IOException If an IO error occurs
     */
    private static String readStream(InputStream stream) throws IOException
    {
        StringBuilder sb = new StringBuilder();
        BufferedReader br = new BufferedReader(new InputStreamReader(stream));
        String line = null;
        while (true)
        {
            line = br.readLine();
            if (line == null)
            {
                break;
            }
            sb.append(line+"\n");
        }
        return sb.toString();
    }
    
    
    /**
     * Debug method that prints the log of the specified shader
     * 
     * @param id The shader
     */
    private void printShaderLogInfo(int id)
    {
        IntBuffer infoLogLength = 
            ByteBuffer.allocateDirect(4).order(
                ByteOrder.nativeOrder()).asIntBuffer();
        gl.glGetShaderiv(id, GL3.GL_INFO_LOG_LENGTH, infoLogLength);

        ByteBuffer infoLog = 
            ByteBuffer.allocateDirect(infoLogLength.get(0)).order(
                ByteOrder.nativeOrder());
        gl.glGetShaderInfoLog(id, infoLogLength.get(0), null, infoLog);

        String infoLogString = 
            Charset.forName("US-ASCII").decode(infoLog).toString();
        System.out.println("Shader Log: " + infoLogString);
    }
    
    /**
     * Debug method that prints the log of the specified program
     * 
     * @param id The program
     */
    private void printProgramLogInfo(int id)
    {
        IntBuffer infoLogLength = 
            ByteBuffer.allocateDirect(4).order(
                ByteOrder.nativeOrder()).asIntBuffer();
        gl.glGetProgramiv(id, GL3.GL_INFO_LOG_LENGTH, infoLogLength);

        ByteBuffer infoLog = 
            ByteBuffer.allocateDirect(infoLogLength.get(0)).order(
                ByteOrder.nativeOrder());
        gl.glGetProgramInfoLog(id, infoLogLength.get(0), null, infoLog);

        String infoLogString = 
            Charset.forName("US-ASCII").decode(infoLog).toString();
        System.out.println("Program Log: " + infoLogString);
    }


}
